/*******************************/
/*   assembler for 6803        */
/*       Main part             */
/*******************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    930818      Ported to Unix on Sun
*
*  0.02    941019   1. files asm03man.c and asm03ma2.c concatenated.
*                   2. tabs replaced with spaces.
*
*  0.03    970314   1. Version printout added.
*
*  0.04    970320   1. The space added last of input is shopped off
*
*  0.05    170124   1. pass variable: FIRST_PASS , LAST_PASS
*
*  0.06    200405   1. return type for adrfield() changed: void => int
*
*/
static char *version = "asm03man 0.06  2020-04-05";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "asmdef.h"


#define TXTSTART  24
#define FCBFDBLINES 4

/* Printout formatting of opcode data part */
#define FORM1     0             /* add no space */
#define FORM2     2             /* add a space before the second byte */

int outform;

int GetParameter(int argc, char *argv[]);
void assemble(char *name, char *targetname);
int setcputype(char *name);

char *manversion()
{

    return version;
}

int main(int argc, char *argv[])
{
   short i, FilesNo;

/* appl_init();  */
   init();
   txtstart = TXTSTART;
   FilesNo = GetParameter(argc, argv);

   actfile = 0;

   assemble(files[0].filesname, "6801/6803");



   fflush(hdl_scr);
/* getchar(); */
/* appl_exit();  */
   exit(errcnt > 0);
}

/*
 * return if there is any address field
 *
 *
 *
 */
int isadrfield()
{
    int ret;

    optype &= TYPEMASK;
    switch (optype) {
    case INH:
    case TRANS:
        ret = 0;
        break;

    default:
        ret = 1;
    }

    return ret;
}

int adrfield()
{
    long val ,ofs;
    short ptrsav;
    short c, adrm;

    c = 0;                        /* default adr mode */
    opclen = 1;
    rflag = 0;
    outform = FORM1;

    switch (optype) {
    case INH:
        break;

    default:
        while(txt[txtptr] == ' ') {
            txtptr++;
        }
        adrm = 0;

        switch(txt[txtptr++]) {   /* check '#', '<' or '>'  */
        case '#':
            adrm = 1;
            break;

        case '<':                 /* forced DIR */
            adrm = 2;
            break;

        case '>':                 /* forced EXT */
            adrm = 3;
            break;

        default:
            --txtptr;
            break;
        }

        exprflag = 0;
        ptrsav = txtptr;          /* check 'nn'  */
        val = expr();
        if (ptrsav != txtptr) {
            adrm |= 4;
        }
        if (txt[txtptr] == ',') {      /* ',X' */
            txtptr++;
            if (toupper(txt[txtptr]) == 'X') {      /* ',X' */
                txtptr++;
                adrm |= 8;
            }
            else {
                error(REGUNREC, txtptr);
            }
        }

        switch(adrm) {
        case 4:             /* DIR, EXT, REL */
            if (optype == REL) {         /* Relative optype */
                c = AREL;
                rflag = 1;
                dta[10] = val >> 8;
                dta[11] = val;
                val &= 0xFFFF;
                ofs = val - prgcnt - 2;
                if (ofs > 127 || ofs < -128) {
                    error(BRANCHOUT, ptrsav);
                    dta[1] = 0;
                }
                else {
                    dta[1] = ofs;
                }
            }
            else {                     /* Other optype */
                if (exprflag & 2) {       /* first pass: sym not def */
                    c = AEXT;           /* use longest possible    */
                                      /* addressing mode         */
                }
                else if (optype != NODIR && (val & 0xFF00) == 0) {
                    c = ADIR;
                }
                else {
                    c = AEXT;
                    if ( lbwrn && (opref & (MNEJUMP>>8))) {
                        val &= 0xFFFF;
                        ofs = val - prgcnt - 2;
                        if (ofs >= -128 && ofs <= 127) {
                            error(SHBRUSED, ptrsav);
                        }
                    }
                }

                if (pass != LAST_PASS) {
                    adrsiztab[adrsizcnt++] = c;
                }
                else {
                    c = adrsiztab[adrsizcnt++];
                }

                if (c == ADIR) {
                    dta[1] = val;
                }
                else {
                    dta[1] = val >> 8;
                    dta[2] = val;
                }
            }
            break;

        case 1:
            error(NOEXPR, ptrsav);
            /* no break */
        case 5:       /* IMM8, IMM16  */
            if (optype == IMM16) {
                c = AIMM16;
                dta[1] = val >> 8;
                dta[2] = val;
            }
            else {
                c = AIMM8;
                dta[1] = val;
                if (val < -128 || val > 255) {
                    error(BIGGERVAL, ptrsav);
                }
            }
            break;

        case 6:       /* forced DIR */
            c = ADIR;
            dta[1] = val;
            break;

        case 7:       /* forced EXT */
            c = AEXT;
            dta[1] = val >> 8;
            dta[2] = val;
            break;

        case 0xC:     /* IND */
            c = AIND;
            dta[1] = val;
            break;

        default:
            c = -1;
            break;
        }
        break;

    case IMMADR:        /* #nn,nn    #nn,nn,X          */
        while(txt[txtptr] == ' ') {
            txtptr++;
        }

        c = AIMMDIR;
        outform = FORM2;

        if (txt[txtptr] == '#') {
            txtptr++;
        }
        ptrsav = txtptr;        /* check 'nn'  */
        opref = READREF;
        val = expr();
        dta[1] = val;
        if (ptrsav == txtptr) {
            error(ADRMODBAD, ptrsav);
        }
        else if (val < -128 || val > 255) {
            error(BIGGERVAL, ptrsav);
        }

        if (txt[txtptr] == ',') {  /* ',nn' */
            txtptr++;
            opref = READREF;
            ptrsav = txtptr;
            val = expr();
            dta[2] = val;
            if (ptrsav == txtptr) {
                error(ADRMODBAD, ptrsav);
            }
            else if (val < 0 || val > 255) {
                error(BIGGERVAL, ptrsav);
            }

            if (txt[txtptr] == ',') {      /* ',X' */
                txtptr++;
                if (toupper(txt[txtptr]) == 'X') {      /* ',X' */
                    txtptr++;
                    c = AIMMIND;
                }
                else {
                    error(REGUNREC, txtptr);
                }
            }

        }
        else {
            error(ADRMODBAD, ptrsav);
        }

        break;
    }

    if (c >= 0) {
        dta[0] = opcode + adrtab[c].addval;
        oplen = adrtab[c].length;
        prgcnt += oplen;

        if ((adrtab[c].mask & adrcnv[optype]) == 0) {
            error(ADRMODBAD, ptrsav);
        }
    }
    else {
        error(ADRMODBAD, ptrsav);
    }

    return 0;
}

int assigndirective_cpu(int opcode, int inptrsav)
{
    int ptrsav;
    int val;

    return 0;
}

/*
 *   return: 0=no directive found, 1=directive found
 */
int directive(int dir)
{

    return 0;
}

/*
 *   return: 0=no opt found, 1=opt found
 */
int diropt2(char *r)
{

    if (!strcmp(r, "Z00")) { /* Z00 : 6800, 6802, 6808*/
        txtptr += 3;
        setcputype("00");
        return 1;
    }
    else if (!strcmp(r, "Z01")) { /* Z01 : 6801, 6803 */
        txtptr += 3;
        setcputype("01");
        return 1;
    }
    else {
        return 0;
    }
}

/*
   
   0         1         2         3         4
   01234567890123456789012345678901234567890
   .      .    .  .   .    .
   LLLLL  AAAA DD DD  FFFF
   LLLLL  AAAA DD DD
   LLLLL  AAAA DD DDDD
   LLLLL  AAAA DD DD DD
   LLLLL..:....:...........str 
   
   */
void formlst(FILE *hdl)
{
    int i, erf, txtp;
    char *ptr;
    int ofm;
    char *p;

    if (prntline == 0) {
        return;
    }

    i = strlen(txt) - 1;    /* shop off the added space at the end */
    if (txt[i] == ' ') {
        txt[i] = 0;
    }

    ptr = txtout;
                /* LLLLL..:....:...........str */
    sprintf(txtout,"%5d                   %s", prntline, txt);

    if (aflag) {       /* AAAA */
        sprintf(txtout+7,"%04X", (int) adr);
        txtout[11] = ' ';
    }

    erf = 0;
    switch(dflag) {
    case ASMPRT:
        sprintf(txtout+12,"%02X",dta[0]);
        txtout[14] = ' ';

        p = txtout + 15;
        ofm = outform;
        for (i = opclen; i < len; i++) {
            if ((ofm & 1)) {
                *p++ = ' ';
            }
            sprintf(p, "%02X", dta[i]);
            p += 2;
            *p = ' ';
            ofm >>= 1;
        }

        if (rflag) {        /* relative op:s */
            sprintf(txtout+TXTSTART-5,"%02X%02X", dta[10], dta[11]);
            txtout[TXTSTART-1] = ' ';
        }
        break;

    case FCBPRT:
        for(i = 0; i < len - 4; i += 4) {
                               /*DD . DD . DD . DD */ 
            sprintf(txtout+12,"%02X %02X %02X %02X", dta[i], dta[i+1],
                                                  dta[i+2], dta[i+3]);
            txtout[23] = ' ';
            fprintf(hdl,"%s\n",txtout);
            sprintf(txtout,"             ");
            for(txtp = 12 ; txtp < 27 ; txtout[txtp++] = 0)
                ;

            if (!erf) {
                formerr(hdl);
                erf = 1;
                if (hdl == hdl_err) {
                    break;
                }
            }
        }

        switch(len-i) {
        case 1:
            sprintf(txtout+12,"%02X", dta[i]);
            txtout[14] = ' ';
            break;

        case 2:
            sprintf(txtout+12,"%02X %02X", dta[i], dta[i+1]);
            txtout[17] = ' ';
            break;

        case 3:
            sprintf(txtout+12,"%02X %02X %02X", dta[i], dta[i+1], dta[i+2]);
            txtout[20] = ' ';
            break;

        case 4:
            sprintf(txtout+12,"%02X %02X %02X %02X", dta[i], dta[i+1],
                                                  dta[i+2], dta[i+3]);
            txtout[23] = ' ';
            break;
        }
        break;

    case FDBPRT:
        for(i = 0; i < len - 4; i += 4) {
                                /*DDDD . DDDD */
            sprintf(txtout+12,"%02X%02X %02X%02X",dta[i], dta[i+1],
                                               dta[i+2], dta[i+3]);
            txtout[21] = ' ';
            fprintf(hdl,"%s\n",txtout);
            sprintf(txtout,"             ");
            for(txtp = 12 ; txtp < 27 ; txtout[txtp++] = 0)
             ;

            if (!erf) {
                formerr(hdl);
                erf = 1;
                if (hdl == hdl_err) {
                    break;
                }
            }
        }

        switch(len - i) {
        case 2:
            sprintf(txtout+12,"%02X%02X", dta[i], dta[i+1]);
            txtout[16] = ' ';
            break;

        case 4:
            sprintf(txtout+12,"%02X%02X %02X%02X", dta[i], dta[i+1],
                                                dta[i+2], dta[i+3]);
            txtout[21] = ' ';
            break;
        }
        break;

    }

    fprintf(hdl, "%s\n", txtout);
    if (!erf) {
        formerr(hdl);
    }
}

/* end */

