/*******************************/
/*   assembler for 6809        */
/*       Main part             */
/*******************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    930818      Ported from Atari to Unix on Sun
*
*  0.02    941019   1. files asm03man.c and asm03ma2.c concatenated.
*                   2. tabs replaced with spaces.
*
*  0.03    970314   1. Version printout added.
*
*  0.04    970320   1. The space added last of input is shopped off
*
*  0.05    120517   1. <stdlib.h> and included to make gcc on linux happier
*                   2. <string.h> and included to make gcc on linux happier
*                   3. some castings added for the same reason...
*
*  0.06    161025   1. FQB print added
*                   2. lbwrn-flag moved to asmcom
*                   3. Test for 'Warning - short branch can ...', was wrong
*
*  0.07    170122   1. Possible to use flag name in ANDCC etc.
*                   2. Common function for registerlists (and flaglists)
*                   3. Bugfix: Calculation of offset for AIM nn,PCR was wrong
*                   4. TXTSTART increased. offset data was overwritten by dest
*                   5. pass variable: FIRST_PASS , MIDDLE_PASS and LAST_PASS
*
*  0.08    171010   1. Dir REG added: get a bitmask from a registerlist
*
*  0.09    180127   1. force of size for branch op added: '>' and '<' used
*
*
*/
static char *version = "asm09man 0.09  2018-01-27";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "asmdef.h"


#define TXTSTART  35
#define FCBFDBLINES 4

/* Printout formatting of opcode data part */
#define FORM1     0             /* add no space */
#define FORM2     2             /* add a space before the second byte */
#define FORM3     6             /* add a space before the second and third bytes */



extern REGNAM pshsreg[];
extern REGNAM pshureg[];
extern REGNAM exgreg[];
extern REGNAM indreg[];
extern REGNAM indofsreg[];
extern REGNAM tfmreg[];
extern REGNAM bandreg[];
extern REGNAM ccflags[];
extern BranchOp braoptab[];
extern TRNSL transtab[];


int dpregval;
int outform;

int GetParameter(int argc, char *argv[]);
char *strdcpy(char *d, char *s, char *e);
char *strndcpy(char *d, char *s, char *e, int n);
int getregval(REGNAM *t, char *s);
int getreglistmask(REGNAM *table, char *s);
void addoplenstat(int oplen, int opclen, int page, int opcode);
void setmaxoplen(int oplen, int opclen, int pages);
void branchstat(int type, int ofs, int prgcnt, int txtpos);
void assemble(char *name, char *targetname);


/*
 * branch offset statistics:
 *
 * For short branch:
 *
 *     - -32
 * -31 - -11
 * -10 - -5
 *  -4 - -1
 *   0
 *   1 - 4
 *   5 - 10
 *  11 - 31
 *  32 -
 *
 * For long branch:
 *      - -257 
 * -256 - -193
 * -192 - -149
 * -148 - -141
 * -140 - -137
 * -136 - -133
 * -132 - -129
 * -128 - -1
 *    0 - 127
 *  128 - 131
 *  132 - 135
 *  136 - 139
 *  140 - 147
 *  148 - 191
 *  192 - 255
 *  256 -
 *
 *
 *
 */






char *manversion()
{

    return version;
}

int main(int argc, char *argv[])
{
    short i;

    init();

    dpregval = 0;
    txtstart = TXTSTART;
    GetParameter(argc, argv);
    setmaxoplen(5, 2, 3);

    actfile = 0;

    assemble(files[0].filesname, "6809");



    fflush(hdl_scr);
    exit(errcnt > 0);
}

/*
 * return if there is any address field
 *
 *
 *
 */
int isadrfield()
{
    int ret;

    optype &= TYPEMASK;
    switch (optype) {
    case INH:
    case TRANS:
        ret = 0;
        break;

    default:
        ret = 1;
    }

    return ret;
}

int findbranchop(int opcode, int page, int optype)
{
    BranchOp *bop;
    int i;
    int i2;

    i = optype == REL ? 0 : 1;
    i2 = optype == REL ? 1 : 0;

    for (bop = braoptab; bop->page[0] >= 0; bop++) {
        if (bop->opcode[i] == opcode && bop->page[i] == page) {
            return (bop->zeroopt << 16) | (bop->page[i2] << 8) | bop->opcode[i2];
        }
    }

    return -1;
}

int adrfield()
{
    long val, val2, val3, ofs;
    long src;
    short ptrsav;
    short c, adrm;
    short pbmode, indirect;
    short ireg, addsiz;
    short ofsreg;
    short pageflag;
    char r[16];
    int i;
    int savedoptype;
    int immval;
    int skipped;

    c = 0;                      /* default adr mode */
    opclen = 1;
    pageflag = 0;
    rflag = 0;
    outform = FORM1;

    if ((optype & PAGEMASK) == PAGE1) {
        pageflag = 1;
        ++opclen;
    }
    else if ((optype & PAGEMASK) == PAGE2) {
        pageflag = 2;
        ++opclen;
    }

    addsiz = 0;                 /* only used for indexed addr mode */
    optype &= TYPEMASK;
    skipped = 0;

    switch (optype) {
    case INH:
        break;

    case TRANS:
        opclen = transtab[opcode].trcount >> 8;
        oplen = transtab[opcode].trcount & 0xff;
        outform = transtab[opcode].troutform;
        dta[0] = transtab[opcode].trdta[0];
        dta[1] = transtab[opcode].trdta[1];
        dta[2] = transtab[opcode].trdta[2];
        dta[3] = transtab[opcode].trdta[3];
        prgcnt += oplen;
        return 0;
        break;

    default:
        while (txt[txtptr] == ' ') {
            txtptr++;
        }

        savedoptype = -1;       /* a hack for AIM etc. */
        if (optype == IMMADR) {
            savedoptype = optype;
            optype = DIR2;
            immval = 0;

            if (txt[txtptr] == '#') {
                txtptr++;
            }
            ptrsav = txtptr;    /* check 'nn'  */
            val = expr();
            if (ptrsav != txtptr) {
                immval = val;
            }
            if (txt[txtptr] == ',') {
                txtptr++;
            }
        }


        adrm = 0;
        pbmode = 0;
        indirect = 0;

/*  /  [  /  A B D E F W  n  /  ,  /  --  -  /  W r  /  ++  +  /  ]  /  */
/*           1 2 3 4 5 6  0     1     2  1      1 0     2   1           */
/*     1                                                                */
/*  indirect  pbmode.b0-2      .b3    .b4-5       ireg  .b6-7           */
/*                                              .b8                     */

        /*   field     pbmode       */
        /*   n      0x00 000000000   */
        /*   n,r    0x08 000001000   */
        /*   A,r    0x09 000001001   */
        /*   B,r    0x0a 000001010   */
        /*   D,r    0x0b 000001011   */
        /*   E,r    0x09 000001100   */
        /*   F,r    0x0a 000001101   */
        /*   W,r    0x0b 000001110   */
        /*   ,r     0x08 000001000   */
        /*   ,--r   0x28 000101000   */
        /*   ,-r    0x18 000011000   */
        /*   ,r++   0x88 010001000   */
        /*   ,r+    0x48 001001000   */
        /*   n,W   0x108 100001000   */
        /*   ,W    0x108 100001000   */
        /*   ,--W  0x128 100101000   */
        /*   ,W++  0x188 110001000   */

        if (txt[txtptr] == '[') {
            txtptr++;
            outform = FORM2;
            indirect = 1;
        }

        switch(txt[txtptr++]) { /* check '#', '<' or '>'  */
        case '#':
            adrm = 1;
            break;

        case '<':
            adrm = 2;
            break;

        case '>':
            adrm = 3;
            break;

        default:
            --txtptr;
            break;
        }

        ptrsav = txtptr;
        i = 0;
        while (isalnum(txt[txtptr]) && i < 4) {
            r[i++] = txt[txtptr++];
        }
        r[i] = 0;
        ofsreg = getregval(indofsreg, r); /*  */
        if (ofsreg > 0 && txt[txtptr] == ',') {
            pbmode = ofsreg;
        }
        else {
            txtptr = ptrsav;    /* backtrack parser */
            exprflag = 0;
            ptrsav = txtptr;    /* check 'nn'  */
            val = expr();
            if (ptrsav != txtptr) {
                adrm |= 4;
            }
        }

        if (txt[txtptr] == ',') {
            txtptr++;
            pbmode |= 8;
            adrm |= 8;

            if (txt[txtptr] == '-') {
                txtptr++;
                if (txt[txtptr] == '-') {
                    txtptr++;
                    pbmode |= 0x20;
                }
                else {
                    pbmode |= 0x10;
                }
            }

            i = 0;
            while (isalnum(txt[txtptr])) {
                r[i++] = txt[txtptr++];
            }
            r[i] = 0;
            ireg = getregval(indreg, r); /*  */

            if (ireg == -1) {
                error(REGUNREC, txtptr - i);
            }
            else if (ireg == 32) { // W
                pbmode |= 0x100;
                ireg = 0;
            }
            if (txt[txtptr] == '+') {
                txtptr++;
                if (txt[txtptr] == '+') {
                    txtptr++;
                    pbmode |= 0x80;
                }
                else {
                    pbmode |= 0x40;
                }
            }
        }

        if (indirect) {
            if (txt[txtptr] != ']') {
                error(ADRMODBAD, txtptr);
            }
        }

        switch(adrm) {
        case 4:                 /* DIR, EXT */
            if (indirect) {
                c = AIND;
                outform = FORM2;
                addsiz = 2;
                dta[1] = 0x9f;
                dta[2] = val >> 8;
                dta[3] = val;
                break;
            }
            if (exprflag & 2) { /* first pass: sym not def */
                c = AEXT;   /* use longest possible    */
                /* addressing mode         */
            }
            else if ((val & 0xFF00) == dpregval) {
                c = ADIR;
            }
            else {
                c = AEXT;
            }

            if (pass != LAST_PASS) {
                adrsiztab[adrsizcnt++] = c;
            }
            else {
                c = adrsiztab[adrsizcnt++];
            }

            if (c == ADIR) {
                dta[1] = val;
            }
            else {
                dta[1] = val >> 8;
                dta[2] = val;
            }
            break;

        case 1:
            error(NOEXPR, ptrsav);
            /* no break */
        case 5:                 /* IMM8, IMM16  */
            if (optype == IMM16) {
                c = AIMM16;
                dta[1] = val >> 8;
                dta[2] = val;
            }
            else if (optype == IMM32) {
                c = AIMM32;
                dta[1] = val >> 24;
                dta[2] = val >> 16;
                dta[3] = val >> 8;
                dta[4] = val;

// special hack for IMM32, since only LDQ use this
// and it has its immediate opcode at page0, but the other at page1
                opcode = 0xcd;
                pageflag = 0;
                --opclen;
            }
            else {
                c = AIMM8;
                dta[1] = val;
                if (val < -128 || val > 255) {
                    error(BIGGERVAL, ptrsav);
                }
            }
            break;

        case 6:                 /* forced DIR */
            c = ADIR;
            dta[1] = val;
            if (indirect) {
                c = -1;
            }
            break;

        case 7:                 /* forced EXT */
            c = AEXT;
            dta[1] = val >> 8;
            dta[2] = val;
            if (indirect) {
                c = -1;
            }
            break;

            /*   field     pbmode       ireg : pb (hex)     */
            /*   A,r    0x09 000001001  0-3    86           */
            /*   B,r    0x0a 000001010  0-3    85           */
            /*   D,r    0x0b 000001011  0-3    8b           */
            /*   E,r    0x0c 000001100  0-3    87           */
            /*   F,r    0x0d 000001101  0-3    8a           */
            /*   W,r    0x0e 000001110  0-3    8e           */
            /*   ,r     0x08 000001000  0-3    84           */
            /*   ,--r   0x28 000101000  0-3    83           */
            /*   ,-r    0x18 000011000  0-3    82           */
            /*   ,r++   0x88 010001000  0-3    81           */
            /*   ,r+    0x48 001001000  0-3    80           */
            /*   ,W    0x108 100001000  -      8f       90  */
            /*   ,--W  0x128 100101000  -      ef       f0  */
            /*   ,W++  0x188 110001000  -      cf       d0  */

        case 0x8:               /* IND with no value*/
            c = AIND;
            outform = FORM2;
            switch (pbmode) {
            case 0x08:          /*   ,*  */
                dta[1] = 0x84;
                break;

            case 0x09:          /*   A,r  */
                dta[1] = 0x86;
                break;
            case 0x0a:          /*   B,r  */
                dta[1] = 0x85;
                break;
            case 0x0b:          /*   D,r  */
                dta[1] = 0x8b;
                break;
            case 0x0c:          /*   E,r  */
                dta[1] = 0x87;
                break;
            case 0x0d:          /*   F,r  */
                dta[1] = 0x8a;
                break;
            case 0x0e:          /*   W,r  */
                dta[1] = 0x8e;
                break;

            case 0x18:         /*   ,-r  */
                dta[1] = 0x82;
                if (indirect)
                 c = -1;
                break;

            case 0x28:          /*   ,--r */
                dta[1] = 0x83;
                break;

            case 0x48:          /*   ,r+  */
                dta[1] = 0x80;
                if (indirect) {
                    c = -1;
                }
                break;

            case 0x88:          /*   ,r++ */
                dta[1] = 0x81;
                break;

            case 0x108:         /*   ,W */
                if (indirect) {
                    dta[1] = 0x90;
                    indirect = 0;
                }
                else {
                    dta[1] = 0x8f;
                }
                break;
            case 0x128:         /*   ,--W */
                if (indirect) {
                    dta[1] = 0xf0;
                    indirect = 0;
                }
                else {
                    dta[1] = 0xef;
                }
                break;
            case 0x188:         /*   ,W++ */
                if (indirect) {
                    dta[1] = 0xd0;
                    indirect = 0;
                }
                else {
                    dta[1] = 0xcf;
                }
                break;

            default:
                c = -1;
                break;
            }

            if (indirect) {
                dta[1] |= 0x10;
            }

            if (ireg < 8) {
                dta[1] |= ireg << 5;
            }
            else {
                c = -1;
            }
            break;

            /*   field     pbmode       ireg : pb (hex)     */
            /*   n,r    0x08 000001000  0-3    00 84 88 89  */
            /*   n,PC   0x08 000001000  9      8c 8d        */
            /*   n,PCR  0x08 000001000  8      8c 8d        */
            /*   <n,r   0x08 000001000  0-3    88           */
            /*   <n,PC  0x08 000001000  9      8c           */
            /*   <n,PCR 0x08 000001000  8      8c           */
            /*   >n,r   0x08 000001000  0-3    89           */
            /*   >n,PC  0x08 000001000  9      8d           */
            /*   >n,PCR 0x08 000001000  8      8d           */
            /*   n,W   0x108 100001000         af       b0  */

        case 0xC:               /* IND  with value*/
        case 0xE:               /* <IND */
        case 0xF:               /* >IND */
            outform = FORM2;
            switch (pbmode) {
            case 0x08:          /*   *,*  */
                if (ireg == 8) { /*  *,PCR  Calculate relative address */
                    rflag = 1;
                    dta[10] = val >> 8;
                    dta[11] = val;
                    val &= 0xFFFF;
                    src = prgcnt + opclen + 2;
                    if (savedoptype >= 0) {   /* continued hack for AIM etc */
                        src++;
                    }
                    ofs = val - src;
                    dta[1] = ofs >> 8;
                    dta[2] = ofs;
                    if ((adrm & 3) == 2) { /* <IND */
                        c = 2;
                    }
                    else if ((adrm & 3) == 3) { /* >IND */
                        c = 3;
                    }
                    else if (exprflag & 2) { /* first pass: sym not def */
                        if (noforwardopt) { /* no forward optimization */
                            c = 3;           /* use longest    */
                                             /* addressing mode         */
                        }
                        else {
                            c = 2;           /* try shortest first   */
                        }
                    }
                    else if (ofs >= -128 && ofs <= 127) { /* 8-bit ofs */
                        c = 2;
                    }
                    else {      /* 16-bit ofs */
                        c = 3;
                    }

                    if (noforwardopt) { /* no forward optimization */
                        if (pass == FIRST_PASS) {
                            adrsiztab[adrsizcnt++] = c; /* remember selected */
                        }
                        else {
                            c = adrsiztab[adrsizcnt++];
                        }
                    }

                    switch (c) {
                    case 2:     /*   8-bit ofs  */
                        addsiz = 1;
                        dta[1] = 0x8c;
                        dta[2] = ofs;
                        if (ofs < -128 || ofs > 127) {
                            error(BIGGERVAL, ptrsav);
                        }
                        break;
                    case 3:     /*   16-bit ofs */
                        addsiz = 2;
                        dta[1] = 0x8d;
                        --ofs;
                        dta[2] = ofs >> 8;
                        dta[3] = ofs;
                        break;
                    }
                    c = AIND;
                }
                else if (ireg == 9) { /*  *,PC */
                    if ((adrm & 3) == 2) { /* <IND */
                        c = 2;
                    }
                    else if ((adrm & 3) == 3) { /* >IND */
                        c = 3;
                    }
                    else if (exprflag & 2) { /* first pass: sym not def */
                        c = 3;               /* use longest possible    */
                                             /* addressing mode         */
                    }
                    else if (val >= -128 && val <= 127) { /* 8-bit ofs */
                        c = 2;
                    }
                    else {      /* 16-bit ofs */
                        c = 3;
                    }

                    if (pass != LAST_PASS) {
                        adrsiztab[adrsizcnt++] = c;
                    }
                    else {
                        c = adrsiztab[adrsizcnt++];
                    }
                    switch (c) {
                    case 2:     /*   8-bit ofs  */
                        addsiz = 1;
                        dta[1] = 0x8c;
                        dta[2] = val;
                        if (val < -128 || val > 127) {
                            error(BIGGERVAL, ptrsav);
                        }
                        break;
                    case 3:     /*   16-bit ofs */
                        addsiz = 2;
                        dta[1] = 0x8d;
                        dta[2] = val >> 8;
                        dta[3] = val;
                        break;
                    }
                    c = AIND;
                }
                else {          /*  *,r  */
                    if ((adrm & 3) == 2) { /* <IND */
                        c = 2;
                    }
                    else if ((adrm & 3) == 3) { /* >IND */
                        c = 3;
                    }
                    else if (exprflag & 2) { /* first pass: sym not def */
                        c = 3;               /* use longest possible    */
                                             /* addressing mode         */
                    }
                    else if (val == 0) { /* no ofs */
                        c = 0;
                    }
                    else if (indirect == 0 && val >= -16 && val <= 15) {
                        c = 1;  /* 5-bit ofs */
                    }
                    else if (val >= -128 && val <= 127) { /* 8-bit ofs */
                        c = 2;
                    }
                    else {      /* 16-bit ofs */
                        c = 3;
                    }

                    if (pass != LAST_PASS) {
                        adrsiztab[adrsizcnt++] = c;
                    }
                    else {
                        c = adrsiztab[adrsizcnt++];
                    }

                    switch (c) {
                    case 0:     /*   no ofs     */
                        dta[1] = 0x84;
                        break;
                    case 1:     /*   5-bit ofs  */
                        dta[1] = val & 0x1f;
                        break;
                    case 2:     /*   8-bit ofs  */
                        addsiz = 1;
                        dta[1] = 0x88;
                        dta[2] = val;
                        if (val < -128 || val > 255) {
                            error(BIGGERVAL, ptrsav);
                        }
                        break;
                    case 3:     /*   16-bit ofs */
                        addsiz = 2;
                        dta[1] = 0x89;
                        dta[2] = val >> 8;
                        dta[3] = val;
                        break;
                    }

                    c = AIND;
                    if (ireg < 8) {
                        dta[1] |= ireg << 5;
                    }
                    else {
                        c = -1;
                    }
                }
                break;

            /*   n,W   0x108 100001000         af       b0  */
            case 0x108:          /*   *n,W  */
                if (indirect) {
                    dta[1] = 0xb0;
                    indirect = 0;
                }
                else {
                    dta[1] = 0xaf;
                }
                addsiz = 2;
                dta[2] = val >> 8;
                dta[3] = val;
                c = AIND;
                break;

            default:
                c = -1;
                break;
            }

            if (indirect) {
                dta[1] |= 0x10;
            }
            break;

        default:
            c = -1;
            break;
        }

        if (savedoptype >= 0) {     /* continued hack for AIM etc */
            optype = savedoptype;
            dta[5] = dta[4];
            dta[4] = dta[3];
            dta[3] = dta[2];
            dta[2] = dta[1];
            dta[1] = immval;
            outform <<= 1;
            outform |= 2;
            addsiz++;
        }

        break;

    case REL:
    case REL16:
        switch(txt[txtptr]) { /* check '<' or '>'  */
        case '<':       // force short
            ++txtptr;
            adrm = 2;
            break;

        case '>':       // force long
            ++txtptr;
            adrm = 3;
            break;

        default:
            adrm = 0;
            break;
        }

        exprflag = 0;
        ptrsav = txtptr;    /* check 'nn'  */
        val = expr();
        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
        }

        rflag = 1;
        dta[10] = val >> 8;
        dta[11] = val;
        val &= 0xFFFF;

        if (optype == REL) { /* REL 8 */
            c = AREL;
            ofs = val - prgcnt - 2;
            dta[1] = ofs;

            if (pass == FIRST_PASS) {
                skipped = 0;
                adrsiztab[adrsizcnt] = 0;
            }
            else {
                skipped = adrsiztab[adrsizcnt]; // once it is set, it it never cleared
                if (skipped) {
                }
                else if (ofs == 0 && (optimizebranch&0x100) != 0 && adrm == 0) {
                    val2 = findbranchop(opcode, pageflag, optype);
                    if (val2 > 0) {
                        if (val2 & 0x10000) {
                            skipped = 1;
//                            fprintf(hdl_scr, " 0x%04X zeroofs   %05X\n", (uint) prgcnt, val2);
                            adrsiztab[adrsizcnt] = skipped;
                        }
                    }
                }
            }

            if (skipped) {
                error(INSTROPTAWAY, -1);
            }
            else if (ofs > 127 || ofs < -128) {
                if ((optimizebranch&255) == 0 || adrm == 2) {
                    error(BRANCHOUT, ptrsav);
                }
                else if (pass != FIRST_PASS) { // don't change during first pass
                    val2 = findbranchop(opcode, pageflag, optype);
                    if (val2 < 0) {
                        error(BRANCHOUT, ptrsav);
                    }
                    else {
                        if ((optimizebranch&255) > 1) {
                            error(BRCHSHORTTOLONG, -1);
//                            fprintf(hdl_scr, " 0x%04X s->l\n", (uint) prgcnt);
                        }
                        pageflag = (val2 >> 8) & 0xff;
                        opcode = val2 & 0xff;
                        optype = REL16;
                        c = AREL16;

                        opclen = 1;
                        if (pageflag) {
                            opclen = 2;
                        }
                        ofs = val - prgcnt - opclen - 2;
                        dta[1] = ofs >> 8;
                        dta[2] = ofs;
                    }
                }
            }

            branchstat(optype, ofs, prgcnt, ptrsav);
        }
        else {              /* REL 16 */
            c = AREL16;
            ofs = val - prgcnt - opclen - 2;
            dta[1] = ofs >> 8;
            dta[2] = ofs;
            if (pass == FIRST_PASS) {
                skipped = 0;
                adrsiztab[adrsizcnt] = 0;
            }

            if (ofs >= -128 && ofs <= 127) {
                if (pass == FIRST_PASS) {
                    skipped = 0;
                    adrsiztab[adrsizcnt] = 0;
                }
                else {
                    skipped = adrsiztab[adrsizcnt]; // once it is set, it it never cleared
                    if (skipped) {
                    }
                    else if (ofs == 0 && (optimizebranch&0x100) != 0 && adrm == 0) {
                        val2 = findbranchop(opcode, pageflag, optype);
                        if (val2 > 0) {
                            if (val2 & 0x10000) {
                                skipped = 1;
                                adrsiztab[adrsizcnt] = skipped;
                            }
                        }
                    }
                }

                if (skipped) {
                    error(INSTROPTAWAY, -1);
                }
                else if ((optimizebranch&255) == 0 || adrm == 3) {
                    if (lbwrn && adrm != 3) {   // keep silent if long forced
                        error(SHBRUSED, ptrsav);
                    }
                }
                else if (pass != FIRST_PASS) { // don't change during first pass
                    val2 = findbranchop(opcode, pageflag, optype);
                    if ((optimizebranch&255) > 1) {
                        error(BRCHLONGTOSHORT, -1);
//                            printf(" 0x%04X l->s   0x%04X\n",
//                                   (uint) prgcnt, (uint) val2);
                    }
                    pageflag = (val2 >> 8) & 0xff;
                    opcode = val2 & 0xff;
                    optype = REL;
                    c = AREL;

                    opclen = 1;
                    if (pageflag) {
                        opclen = 2;
                    }

                    ofs = val - prgcnt - 2;
                    dta[1] = ofs;
                }
            }

            branchstat(optype, ofs, prgcnt, ptrsav);
        }
        adrsizcnt++;
        break;

    case PSHS:                  /*       */
    case PSHU:                  /*       */
        if (txt[txtptr] == '#') {
            txtptr++;
            ptrsav = txtptr;    /* check 'nn'  */
            val = expr();
            if (ptrsav == txtptr) {
                error(NOEXPR, ptrsav);
            }
        }
        else if (optype == PSHS) {
            val = getreglistmask(pshsreg, txt+txtptr);
        }
        else {
            val = getreglistmask(pshureg, txt+txtptr);
        }

        dta[1] = val;
        c = ASPEC;
        oplen = 2;
        break;

    case CLRCC:
    case SETCC:
        if (txt[txtptr] == '#') {
            txtptr++;
            ptrsav = txtptr;    /* check 'nn'  */
            val = expr();
            if (ptrsav == txtptr) {
                error(NOEXPR, ptrsav);
            }
        }
        else if (optype == CLRCC) {
            val = ~getreglistmask(ccflags, txt+txtptr);
        }
        else {
            val = getreglistmask(ccflags, txt+txtptr);
        }

        optype = IMMON;
        c = AIMM8;
        dta[1] = val;
        break;

    case EXG:                   /*        */
        c = ASPEC;
        oplen = 2;
        ptrsav = txtptr;        /* check first reg  */
        i = 0;
        while (isalnum(txt[txtptr])) {
            r[i++] = txt[txtptr++];
        }
        r[i] = 0;
        val = getregval(exgreg, r);
        dta[1] = val << 4;

        if (val == -1) {
            error(REGUNREC, ptrsav);
        }
        else if (txt[txtptr] == ',') { /* ',' */
            txtptr++;
            ptrsav = txtptr;    /* check second reg  */
            i = 0;
            while (isalnum(txt[txtptr])) {
                r[i++] = txt[txtptr++];
            }
            r[i] = 0;
            val = getregval(exgreg, r);
            dta[1] |= val;

            if (val == -1) {
                error(REGUNREC, ptrsav);
            }
        }
        else {
            error(ADRMODBAD, ptrsav);
        }
        break;

    case TFM:               /*  HD6309  :  r+,r+ ; r-,r- ; r+,r ; r,r+    */
        c = ASPEC;
        oplen = 2;
        val2 = 0;
        val3 = 0;
        ptrsav = txtptr;        /* check first reg  */
        i = 0;
        while (isalpha(txt[txtptr])) {
            r[i++] = txt[txtptr++];
        }
        r[i] = 0;
        val = getregval(tfmreg, r);
        dta[1] = val << 4;

        if (txt[txtptr] == '+') {
            txtptr++;
            val2 = 1;
        }
        else if (txt[txtptr] == '-') {
            txtptr++;
            val2 = 2;
        }

        if (val == -1) {
            error(REGUNREC, ptrsav);
        }
        else if (txt[txtptr] == ',') { /* ',' */
            txtptr++;
            ptrsav = txtptr;    /* check second reg  */
            i = 0;
            while (isalpha(txt[txtptr])) {
                r[i++] = txt[txtptr++];
            }
            r[i] = 0;
            val = getregval(tfmreg, r);
            dta[1] |= val;

            if (val == -1) {
                error(REGUNREC, ptrsav);
            }

            if (txt[txtptr] == '+') {
                txtptr++;
                val3 = 1;
            }
            else if (txt[txtptr] == '-') {
                txtptr++;
                val3 = 2;
            }

            val2 += val3 * 3;
/* 0: r 1: r+ 2: r- */
/* 0   ,         */
/* 1  +,   *  0x3a */
/* 2  -,         */
/* 3   ,+  *  0x3b */
/* 4  +,+  *  0x38 */
/* 5  -,+        */
/* 6   ,-        */
/* 7  +,-        */
/* 8  -,-  *  0x39 */
            switch (val2) {
            case 1:         /*   r+,r   */
                opcode = 0x3a;
                break;

            case 3:         /*   r,r+   */
                opcode = 0x3b;
                break;

            case 4:         /*   r+,r+  */
                opcode = 0x38;
                break;

            case 8:         /*   r-,r-  */
                opcode = 0x39;
                break;

            default:
                error(ADRMODBAD, ptrsav);
                break;
            }
        }
        else {
            error(ADRMODBAD, ptrsav);
        }
        break;

    case BAND:                  /*  HD6309   */
        //  source format: reg,src,dst,addr
        //  bin format:   rrsssddd  addr
        ptrsav = txtptr;        /* get reg  */
        i = 0;
        while (isalpha(txt[txtptr])) {
            r[i++] = txt[txtptr++];
        }
        r[i] = 0;
        val = getregval(bandreg, r);
        dta[1] = val << 6;
        if (val == -1) {
            error(REGUNREC, ptrsav);
        }

        if (txt[txtptr] == ',') {
            txtptr++;
        }
        ptrsav = txtptr;    /* get src  */
        val = expr();
        dta[1] |= (val << 3) & 0x38;
        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
        }

        if (txt[txtptr] == ',') {
            txtptr++;
        }
        ptrsav = txtptr;    /* get dst  */
        val = expr();
        dta[1] |= val & 7;
        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
        }

        if (txt[txtptr] == ',') {
            txtptr++;
        }
        ptrsav = txtptr;    /* get addr  */
        val = expr();
        dta[2] = val;
        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
        }
        outform = FORM3;
        oplen = 3;
        c = ASPEC;
        break;
    }

    if (c == ADIR && optype == DIR2) {
        c = ADIR2;
    }

    if (skipped) {      // optimized away
        oplen = 0;
    }
    else if (c >= 0) {
        if (c == ASPEC) {
            dta[0] = opcode;
        }
        else {
            opcode += adrtab[c].addval;
            dta[0] = opcode;
            oplen = adrtab[c].length + addsiz;

            if ((adrtab[c].mask & adrcnv[optype]) == 0) {
                error(ADRMODBAD, ptrsav);
            }
        }

        if (pageflag != 0) {
            dta[6] = dta[5];
            dta[5] = dta[4];
            dta[4] = dta[3];
            dta[3] = dta[2];
            dta[2] = dta[1];
            dta[1] = dta[0];
            dta[0] = pagetab[pageflag];
            oplen += 1;
        }
        prgcnt += oplen;

// oplength statistics
        addoplenstat(oplen, opclen, pageflag, opcode);

    }
    else {
        error(ADRMODBAD, ptrsav);
    }

    return skipped;
}

int assigndirective_cpu(int opcode, int inptrsav)
{
    int ptrsav;
    int val;

    if (opcode == AD_REG) {  // set registerlist
        ptrsav = txtptr;
        exprflag = 0;
        val = getreglistmask(pshsreg, txt+txtptr);
        defineAndSetVar(0, SYMC_VAR, val, DEF_NORMAL);
        dta[0] = val >> 8;
        dta[1] = val;
        len = 2;
        dflag = FDBPRT;

        return 1;
    }

    return 0;
}

/*
 *   return: 0=no directive found, 1=directive found
 */
int directive(int dir)
{
    long val;
    short i;
    short ptrsav;

    if (dir == AD_SETDP) {          /* SETDP */
        ptrsav = txtptr;
        exprflag = 0x80;
        val = expr();

        if (pass != LAST_PASS) {
            adrsiztab[adrsizcnt++] = exprflag;
        }
        else {
            exprflag = adrsiztab[adrsizcnt++];
        }

        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
        }
        /* if there was an error on this line in pass 1 and no error is
         * found in pass 2, print out the saved error
         */
        if ((exprflag & 2) && errline != line) {
            error(SYMBMISS, ptrsav); /* symbols not defined */
        }

        if (val < 0) {
            dpregval = -1;
            dta[0] = 0xff;
            dta[1] = 0xff;
            len = 2;
            dflag = FDBPRT;
        }
        else {
            dpregval = (val << 8) & 0xff00;
            dta[0] = val;
            len = 1;
            dflag = FCBPRT;
        }

        return 1;
    }

    return 0;
}

/*
 *   return: 0=no opt found, 1=opt found
 */
int diropt2(char *r)
{

    return 0;
}

/*
   
   0         1         2         3         4
   01234567890123456789012345678901234567890
   .     .    .    .           .    .
   LLLLL AAAA DD   DD          FFFF
   LLLLL AAAA DD   DDDD
   LLLLL AAAA DDDD DD DD       FFFF
   LLLLL AAAA DDDD DD DDDD     FFFF
   LLLLL AAAA DD   DD DD DDDD  FFFF
   
   */
void formlst(FILE *hdl)
{
    int i;
    int erf;
    int txtp;
    int j;
    int ofm;
    char *p;
    char tt[32];


    i = strlen(txt) - 1;    /* shop off the added space at the end */
    if (txt[i] == ' ') {
        txt[i] = 0;
    }

    if (prntline) {
        /* LLLLL.:....:....:........:....str */
        sprintf(txtout, "%5d%*s", prntline, TXTSTART, " ");
    }
    else {
        sprintf(txtout, "     %*s", TXTSTART, " ");
    }

    if (aflag) {                /* AAAA */
        sprintf(txtout+6, "%04X", (uint) adr);
        txtout[10] = ' ';
    }

    erf = 0;
    switch(dflag) {
    case ASMPRT :
        p = txtout + 11;
        switch(opclen) {  
        case 1:                 /*DD .......... . str */
            sprintf(txtout+11, "%02X  ", dta[0]);
            txtout[15] = ' ';
            p = txtout + 16;
            break;

        case 2:                 /*DDDD .......... . str */
            sprintf(txtout+11, "%02X%02X", dta[0], dta[1]);
            txtout[15] = ' ';
            p = txtout + 16;
            break;
        }
        ofm = outform;
        for (i = opclen; i < len; i++) {
            if ((ofm & 1)) {
                *p++ = ' ';
            }
            sprintf(p, "%02X", dta[i]);
            p += 2;
            *p = ' ';
            ofm >>= 1;
        }

        if (rflag) {        /* relative op:s */
            sprintf(txtout+TXTSTART-5, "%02X%02X ", dta[10], dta[11]);
        }

        break;

    case FCBPRT :
        p = txtout + 10;
        for (i = 0; i < len; i++) {
            sprintf(tt, " %02X", dta[i]);
            if (p - txtout + strlen(tt) >= TXTSTART) {
                if (!erf) {
                    while ((p - txtout) < TXTSTART) {
                        *p++ = ' ';
                    }
                    strcpy(txtout+TXTSTART, txt);
                    fprintf(hdl, "%s\n", txtout);
                    formerr(hdl);
                    erf = 1;
                    if (hdl == hdl_err) {
                        break;
                    }
                }
                else {
                    fprintf(hdl, "%s\n", txtout);
                }
                strcpy(txtout,"          ");
                p = txtout + 10;
            }
            strcpy(p, tt);
            p += strlen(tt);
            tt[0] = 0;
        }

        if (!erf) {
            while ((p - txtout) < TXTSTART) {
                *p++ = ' ';
            }
        }
        break;

    case FDBPRT:
        p = txtout + 10;
        for (i = 0; i < len; i += 2) {
            sprintf(tt, " %02X%02X", dta[i], dta[i+1]);
            if (p - txtout + strlen(tt) >= TXTSTART) {
                if (!erf) {
                    while ((p - txtout) < TXTSTART) {
                        *p++ = ' ';
                    }
                    strcpy(txtout+TXTSTART, txt);
                    fprintf(hdl, "%s\n", txtout);
                    formerr(hdl);
                    erf = 1;
                    if (hdl == hdl_err) {
                        break;
                    }
                }
                else {
                    fprintf(hdl, "%s\n", txtout);
                }
                strcpy(txtout,"          ");
                p = txtout + 10;
            }
            strcpy(p, tt);
            p += strlen(tt);
            tt[0] = 0;
        }

        if (!erf) {
            while ((p - txtout) < TXTSTART) {
                *p++ = ' ';
            }
        }
        break;

    case FQBPRT:
        p = txtout + 10;
        for (i = 0; i < len; i += 4) {
            sprintf(tt, " %02X%02X%02X%02X",
                    dta[i], dta[i+1], dta[i+2], dta[i+3]);
            if (p - txtout + strlen(tt) >= TXTSTART) {
                if (!erf) {
                    while ((p - txtout) < TXTSTART) {
                        *p++ = ' ';
                    }
                    strcpy(txtout+TXTSTART, txt);
                    fprintf(hdl, "%s\n", txtout);
                    formerr(hdl);
                    erf = 1;
                    if (hdl == hdl_err) {
                        break;
                    }
                }
                else {
                    fprintf(hdl, "%s\n", txtout);
                }
                strcpy(txtout,"          ");
                p = txtout + 10;
            }
            strcpy(p, tt);
            p += strlen(tt);
            tt[0] = 0;
        }

        if (!erf) {
            while ((p - txtout) < TXTSTART) {
                *p++ = ' ';
            }
        }
        break;

    case FTBPRT:
        p = txtout + 10;
        for (i = 0; i < len; i += 3) {
            sprintf(tt, " %02X%02X%02X",
                    dta[i], dta[i+1], dta[i+2]);
            if (p - txtout + strlen(tt) >= TXTSTART) {
                if (!erf) {
                    while ((p - txtout) < TXTSTART) {
                        *p++ = ' ';
                    }
                    strcpy(txtout+TXTSTART, txt);
                    fprintf(hdl, "%s\n", txtout);
                    formerr(hdl);
                    erf = 1;
                    if (hdl == hdl_err) {
                        break;
                    }
                }
                else {
                    fprintf(hdl, "%s\n", txtout);
                }
                strcpy(txtout,"          ");
                p = txtout + 10;
            }
            strcpy(p, tt);
            p += strlen(tt);
            tt[0] = 0;
        }

        if (!erf) {
            while ((p - txtout) < TXTSTART) {
                *p++ = ' ';
            }
        }
        break;

    case PDTAPRT:
        p = txtout + 10;
        j = 0;
        for (i = 0; i < len; i += 2) {
            if (j >= FCBFDBLINES) {
                if (!erf) {
                    while ((p - txtout) < TXTSTART) {
                        *p++ = ' ';
                    }
                    strcpy(txtout+TXTSTART, txt);
                    fprintf(hdl,"%s\n",txtout);
                    formerr(hdl);
                    erf = 1;
                    if (hdl == hdl_err) {
                        break;
                    }
                }
                else {
                    fprintf(hdl,"%s\n",txtout);
                }
                strcpy(txtout,"          ");
                p = txtout + 10;
                j = 0;
            }
            sprintf(p," %02X%02X", pdta[i], pdta[i+1]);
            p += 5;
            j++;
        }

        if (!erf) {
            while ((p - txtout) < TXTSTART) {
                *p++ = ' ';
            }
        }
        break;
    }

    if (!erf) {
        strcpy(txtout+TXTSTART, txt);
    }
    fprintf(hdl, "%s\n", txtout);
    if (!erf) {
        formerr(hdl);
    }
}

/* end */

