/*******************************************/
/*      Definitioner                       */
/*******************************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    930818      Ported from Atari to Unix on Sun
*
*  0.02    941019   1. files asm03man.c and asm03ma2.c concatenated.
*                   2. tabs replaced with spaces.
*
*  0.03    970314   1. Version printout added.
*
*  0.04    970320   1. A space added after every mnem.
*
*  0.05    120517   1. ZMB added
*
*  0.06    171010   1. Directives added and reordered
*
*  0.07    180212   1. ABA,CBA and SBA added
*
*  0.08    180926   1. mnefixtbl added, used for splitted mnemonic: LDA A ...
*
*
*/
static char *version = "asm09tbl 0.08  2018-09-26";

#include <stdio.h>

#include "asmdef.h"

char *tblversion()
{

    return version;
}

#define MC6809  1
#define HD6309  2

/*tables */

CPUTYPE cputab[] = {
    "MC6809", MC6809,
    "HD6309", HD6309,
    "6809", MC6809,
    "6309", HD6309,
    "", 0
};

short pagetab[3] = {
    0,
    0x10,
    0x11
};

short adrcnv[] = { /* for optypes, contents: mask  */
               /* no type   exmpl allowed addressing modes  */
    0x0000,     /* 0  */
    0x0001,     /* 1  INH    NOP   inh                       */
    0x0074,     /* 2  IMM8   ADDA  imm8, dir, ind,ext        */
    0x0078,     /* 3  IMM16  ADDD  imm16,dir, ind,ext        */
    0x0070,     /* 4  DIR2   CLR         dir2,ind,ext        */
    0x0070,     /* 5  NOIMM  STD         dir, ind,ext        */
    0x0040,     /* 6  IND    LEAX             ind            */
    0x0002,     /* 7  REL    BRA   rel                       */
    0x0002,     /* 8  REL16  LBNE  rel16                     */
    0x0070,     /* 9  IMMADR AIM          dir2,ind,ext       */
    0x0001,     /* 10 PSHS   PSHS  reglist                   */
    0x0001,     /* 11 PSHU   PSHU  reglist                   */
    0x0001,     /* 12 EXG    EXG   reglist                   */
    0x0004,     /* 13 IMMON  ORCC  imm8                      */
    0x0000,     /* 14 */
    0x0000,     /* 15 */
    0x0000,     /* 16 */
    0x0000,     /* 17 */
    0x0000,     /* 18 */
    0x0000,     /* 19 */
    0x0000,     /* 20 */
    0x0000,     /* 21 */
    0x0000,     /* 22 */
    0x0170,     /* 23 IMM32    LDQ imm32,dir,ind,ext   (HD6309 only)*/
};

AdrType adrtab[] = { /* contents: reserved, mask, addval, length  */
    0, 0x0001, 0x00, 1,    /* 0  inh */
    0, 0x0002, 0x00, 2,    /* 1  rel */
    0, 0x0004, 0x00, 2,    /* 2  imm8 */
    0, 0x0008, 0x00, 3,    /* 3  imm16 */
    0, 0x0010, 0x10, 2,    /* 4  dir */
    0, 0x0020, 0x30, 3,    /* 5  ext */
    0, 0x0040, 0x20, 2,    /* 6  ind */
    0, 0x0010, 0xc0, 2,    /* 7  dir2 */
    0, 0x0000, 0x00, 0,    /* 8  -    */
    0, 0x0000, 0x00, 0,    /* 9  -    */
    0, 0x0000, 0x00, 0,    /* 10 -    */
    0, 0x0002, 0x00, 3,    /* 11 rel16 */
    0, 0x0001, 0x00, 2,    /* 12 ASPEC */
    0, 0x0100, 0x00, 5,    /* 13 imm32 */
};

/* contents:
 * mnemonic, opcode                   ,         type, cpu 
 *            imm,  dir,  ext,  ind
 */
MneType mnemic[] = {
    "ABA ",   0x26, TRANS, MC6809,                      // 6800
    "ABA ",   0x27, TRANS, HD6309,                      // 6800
    "ABX ",   0x3A, INH, 0,
    "ABY ",   0x01, TRANS, 0,
    "ADCA ",  0x89, MNEREADOP | IMM8, 0,
    "ADCB ",  0xC9, MNEREADOP | IMM8, 0,
    "ADCD ",  0x89, MNEREADOP | IMM16 | PAGE1, HD6309,
    "ADCR ",  0x31, EXG | PAGE1, HD6309,
    "ADDA ",  0x8B, MNEREADOP | IMM8, 0,
    "ADDB ",  0xCB, MNEREADOP | IMM8, 0,
    "ADDD ",  0xC3, MNEREADOP | IMM16, 0,
    "ADDE ",  0x8B, MNEREADOP | IMM8 | PAGE2, HD6309,
    "ADDF ",  0xCB, MNEREADOP | IMM8 | PAGE2, HD6309,
    "ADDR ",  0x30, EXG | PAGE1, HD6309,
    "ADDW ",  0x8B, MNEREADOP | IMM16 | PAGE1, HD6309,
    "AIM ",   0x42, MNEMODIFYOP | IMMADR, HD6309,
    "ANDA ",  0x84, MNEREADOP | IMM8, 0,
    "ANDB ",  0xC4, MNEREADOP | IMM8, 0,
    "ANDCC ", 0x1C, MNEREADOP | CLRCC, 0,
    "ANDD ",  0x84, MNEREADOP | IMM16 | PAGE1, HD6309,
    "ANDR ",  0x34, EXG | PAGE1, HD6309,
    "ASL ",   0x48, MNEMODIFYOP | DIR2, 0,
    "ASLA ",  0x48, INH, 0,
    "ASLB ",  0x58, INH, 0,
    "ASLD ",  0x48, INH | PAGE1, HD6309,
    "ASLD ",  0x22, TRANS, MC6809,
    "ASR ",   0x47, MNEMODIFYOP | DIR2, 0,
    "ASRA ",  0x47, INH, 0,
    "ASRB ",  0x57, INH, 0,
    "ASRD ",  0x47, INH | PAGE1, HD6309,
    "BAND ",  0x30, MNEREADOP | BAND | PAGE2, HD6309,
    "BCC ",   0x24, MNEREADOP | REL, 0,
    "BCS ",   0x25, MNEREADOP | REL, 0,
    "BEOR ",  0x34, MNEREADOP | BAND | PAGE2, HD6309,
    "BEQ ",   0x27, MNEREADOP | REL, 0,
    "BGE ",   0x2C, MNEREADOP | REL, 0,
    "BGT ",   0x2E, MNEREADOP | REL, 0,
    "BHI ",   0x22, MNEREADOP | REL, 0,
    "BHS ",   0x24, MNEREADOP | REL, 0,
    "BIAND ", 0x31, MNEREADOP | BAND | PAGE2, HD6309,
    "BIEOR ", 0x35, MNEREADOP | BAND | PAGE2, HD6309,
    "BIOR ",  0x33, MNEREADOP | BAND | PAGE2, HD6309,
    "BITA ",  0x85, MNEREADOP | IMM8, 0,
    "BITB ",  0xC5, MNEREADOP | IMM8, 0,
    "BITD ",  0x85, MNEREADOP | IMM16 | PAGE1, HD6309,
    "BITMD ", 0x3C, MNEREADOP | IMMON | PAGE2, HD6309,
    "BLE ",   0x2F, MNEREADOP | REL, 0,
    "BLO ",   0x25, MNEREADOP | REL, 0,
    "BLS ",   0x23, MNEREADOP | REL, 0,
    "BLT ",   0x2D, MNEREADOP | REL, 0,
    "BMI ",   0x2B, MNEREADOP | REL, 0,
    "BNE ",   0x26, MNEREADOP | REL, 0,
    "BOR ",   0x32, MNEREADOP | BAND | PAGE2, HD6309,
    "BPL ",   0x2A, MNEREADOP | REL, 0,
    "BRA ",   0x20, MNEREADOP | REL, 0,
    "BRN ",   0x21, MNEREADOP | REL, 0,
    "BSR ",   0x8D, MNEREADOP | REL, 0,
    "BVC ",   0x28, MNEREADOP | REL, 0,
    "BVS ",   0x29, MNEREADOP | REL, 0,
    "CBA ",   0x28, TRANS, MC6809,                      // 6800
    "CBA ",   0x29, TRANS, HD6309,                      // 6800
    "CLC ",   0x02, TRANS, 0,
    "CLF ",   0x16, TRANS, 0,
    "CLI ",   0x03, TRANS, 0,
    "CLIF ",  0x24, TRANS, 0,
    "CLR ",   0x4F, MNEMODIFYOP | DIR2, 0,
    "CLRA ",  0x4F, INH, 0,
    "CLRB ",  0x5F, INH, 0,
    "CLRD ",  0x4F, INH | PAGE1, HD6309,
    "CLRE ",  0x4F, INH | PAGE2, HD6309,
    "CLRF ",  0x5F, INH | PAGE2, HD6309,
    "CLRW ",  0x5F, INH | PAGE1, HD6309,
    "CLV ",   0x04, TRANS, 0,
    "CMPA ",  0x81, MNEREADOP | IMM8, 0,
    "CMPB ",  0xC1, MNEREADOP | IMM8, 0,
    "CMPD ",  0x83, MNEREADOP | IMM16 | PAGE1, 0,
    "CMPE ",  0x81, MNEREADOP | IMM8 | PAGE2, HD6309,
    "CMPF ",  0xC1, MNEREADOP | IMM8 | PAGE2, HD6309,
    "CMPR ",  0x37, EXG | PAGE1, HD6309,
    "CMPS ",  0x8C, MNEREADOP | IMM16 | PAGE2, 0,
    "CMPU ",  0x83, MNEREADOP | IMM16 | PAGE2, 0,
    "CMPW ",  0x81, MNEREADOP | IMM16 | PAGE1, HD6309,
    "CMPX ",  0x8C, MNEREADOP | IMM16, 0,
    "CMPY ",  0x8C, MNEREADOP | IMM16 | PAGE1, 0,
    "COM ",   0x43, MNEMODIFYOP | DIR2, 0,
    "COMA ",  0x43, INH, 0,
    "COMB ",  0x53, INH, 0,
    "COMD ",  0x43, INH | PAGE1, HD6309,
    "COME ",  0x43, INH | PAGE2, HD6309,
    "COMF ",  0x53, INH | PAGE2, HD6309,
    "COMW ",  0x53, INH | PAGE1, HD6309,
    "CPD ",   0x83, MNEREADOP | IMM16 | PAGE1, 0,
    "CPX ",   0x8C, MNEREADOP | IMM16, 0,
    "CPY ",   0x8C, MNEREADOP | IMM16 | PAGE1, 0,
    "CWAI ",  0x3C, MNEREADOP | CLRCC, 0,
    "DAA ",   0x19, INH, 0,
    "DEC ",   0x4A, MNEMODIFYOP | DIR2, 0,
    "DECA ",  0x4A, INH, 0,
    "DECB ",  0x5A, INH, 0,
    "DECD ",  0x4A, INH | PAGE1, HD6309,
    "DECE ",  0x4A, INH | PAGE2, HD6309,
    "DECF ",  0x5A, INH | PAGE2, HD6309,
    "DECW ",  0x5A, INH | PAGE1, HD6309,
    "DES ",   0x09, TRANS, 0,
    "DEX ",   0x05, TRANS, 0,
    "DEY ",   0x07, TRANS, 0,
    "DIVD ",  0x8D, MNEREADOP | IMM8 | PAGE2, HD6309,
    "DIVQ ",  0x8E, MNEREADOP | IMM16 | PAGE2, HD6309,
    "EIM ",   0x45, MNEMODIFYOP | IMMADR, HD6309,
    "EORA ",  0x88, MNEREADOP | IMM8, 0,
    "EORB ",  0xC8, MNEREADOP | IMM8, 0,
    "EORD ",  0x88, MNEREADOP | IMM16 | PAGE1, HD6309,
    "EORR ",  0x36, EXG | PAGE1, HD6309,
    "EXG ",   0x1E, EXG, 0,
    "INC ",   0x4C, MNEMODIFYOP | DIR2, 0,
    "INCA ",  0x4C, INH, 0,
    "INCB ",  0x5C, INH, 0,
    "INCD ",  0x4C, INH | PAGE1, HD6309,
    "INCE ",  0x4C, INH | PAGE2, HD6309,
    "INCF ",  0x5C, INH | PAGE2, HD6309,
    "INCW ",  0x5C, INH | PAGE1, HD6309,
    "INS ",   0x0A, TRANS, 0,
    "INX ",   0x06, TRANS, 0,
    "INY ",   0x08, TRANS, 0,
    "JMP ",   0x4E, MNEJUMP | MNEREADOP | DIR2, 0,
    "JSR ",   0x8D, MNEJUMP | MNEREADOP | NOIMM, 0,
    "LBCC ",  0x24, MNEREADOP | REL16 | PAGE1, 0,
    "LBCS ",  0x25, MNEREADOP | REL16 | PAGE1, 0,
    "LBEQ ",  0x27, MNEREADOP | REL16 | PAGE1, 0,
    "LBGE ",  0x2C, MNEREADOP | REL16 | PAGE1, 0,
    "LBGT ",  0x2E, MNEREADOP | REL16 | PAGE1, 0,
    "LBHI ",  0x22, MNEREADOP | REL16 | PAGE1, 0,
    "LBHS ",  0x24, MNEREADOP | REL16 | PAGE1, 0,
    "LBLE ",  0x2F, MNEREADOP | REL16 | PAGE1, 0,
    "LBLO ",  0x25, MNEREADOP | REL16 | PAGE1, 0,
    "LBLS ",  0x23, MNEREADOP | REL16 | PAGE1, 0,
    "LBLT ",  0x2D, MNEREADOP | REL16 | PAGE1, 0,
    "LBMI ",  0x2B, MNEREADOP | REL16 | PAGE1, 0,
    "LBNE ",  0x26, MNEREADOP | REL16 | PAGE1, 0,
    "LBPL ",  0x2A, MNEREADOP | REL16 | PAGE1, 0,
    "LBRA ",  0x16, MNEREADOP | REL16, 0,
    "LBRN ",  0x21, MNEREADOP | REL16 | PAGE1, 0,
    "LBSR ",  0x17, MNEREADOP | REL16, 0,
    "LBVC ",  0x28, MNEREADOP | REL16 | PAGE1, 0,
    "LBVS ",  0x29, MNEREADOP | REL16 | PAGE1, 0,
    "LDA ",   0x86, MNEREADOP | IMM8, 0,
    "LDAA ",  0x86, MNEREADOP | IMM8, 0,
    "LDAB ",  0xC6, MNEREADOP | IMM8, 0,
    "LDB ",   0xC6, MNEREADOP | IMM8, 0,
    "LDBT ",  0x36, MNEREADOP | BAND | PAGE2, HD6309,
    "LDD ",   0xCC, MNEREADOP | IMM16, 0,
    "LDE ",   0x86, MNEREADOP | IMM8 | PAGE2, HD6309,
    "LDF ",   0xC6, MNEREADOP | IMM8 | PAGE2, HD6309,
    "LDMD ",  0x3D, MNEREADOP | IMMON | PAGE2, HD6309,
    "LDQ ",   0xCC, MNEREADOP | IMM32 | PAGE1, HD6309,
    "LDS ",   0xCE, MNEREADOP | IMM16 | PAGE1, 0,
    "LDU ",   0xCE, MNEREADOP | IMM16, 0,
    "LDW ",   0x86, MNEREADOP | IMM16 | PAGE1, HD6309,
    "LDX ",   0x8E, MNEREADOP | IMM16, 0,
    "LDY ",   0x8E, MNEREADOP | IMM16 | PAGE1, 0,
    "LEAS ",  0x12, MNEREADOP | IND, 0,
    "LEAU ",  0x13, MNEREADOP | IND, 0,
    "LEAX ",  0x10, MNEREADOP | IND, 0,
    "LEAY ",  0x11, MNEREADOP | IND, 0,
    "LSL ",   0x48, MNEMODIFYOP | DIR2, 0,
    "LSLA ",  0x48, INH, 0,
    "LSLB ",  0x58, INH, 0,
    "LSLD ",  0x48, INH | PAGE1, HD6309,
    "LSLD ",  0x22, TRANS, MC6809,
    "LSR ",   0x44, MNEMODIFYOP | DIR2, 0,
    "LSRA ",  0x44, INH, 0,
    "LSRB ",  0x54, INH, 0,
    "LSRD ",  0x44, INH | PAGE1, HD6309,
    "LSRD ",  0x23, TRANS, MC6809,
    "LSRW ",  0x54, INH | PAGE1, HD6309,
    "MUL ",   0x3D, INH, 0,
    "MULD ",  0x8F, MNEREADOP | IMM16 | PAGE2, HD6309,
    "NEG ",   0x40, MNEMODIFYOP | DIR2, 0,
    "NEGA ",  0x40, INH, 0,
    "NEGB ",  0x50, INH, 0,
    "NEGD ",  0x40, INH | PAGE1, HD6309,
    "NOP ",   0x12, INH, 0,
    "OIM ",   0x41, MNEMODIFYOP | IMMADR, HD6309,
    "ORA ",   0x8A, MNEREADOP | IMM8, 0,
    "ORAA ",  0x8A, MNEREADOP | IMM8, 0,
    "ORAB ",  0xCA, MNEREADOP | IMM8, 0,
    "ORB ",   0xCA, MNEREADOP | IMM8, 0,
    "ORCC ",  0x1A, MNEREADOP | SETCC, 0,
    "ORD ",   0x8A, MNEREADOP | IMM16 | PAGE1, HD6309,
    "ORR ",   0x35, EXG | PAGE1, HD6309,
    "PSHA ",  0x18, TRANS, 0,
    "PSHB ",  0x19, TRANS, 0,
    "PSHX ",  0x1a, TRANS, 0,
    "PSHY ",  0x1b, TRANS, 0,
    "PSHS ",  0x34, PSHS, 0,
    "PSHSW ", 0x38, INH | PAGE1, HD6309,
    "PSHU ",  0x36, PSHU, 0,
    "PSHUW ", 0x3A, INH | PAGE1, HD6309,
    "PULA ",  0x1c, TRANS, 0,
    "PULB ",  0x1d, TRANS, 0,
    "PULX ",  0x1e, TRANS, 0,
    "PULY ",  0x1f, TRANS, 0,
    "PULS ",  0x35, PSHS, 0,
    "PULSW ", 0x39, INH | PAGE1, HD6309,
    "PULU ",  0x37, PSHU, 0,
    "PULUW ", 0x3B, INH | PAGE1, HD6309,
    "ROL ",   0x49, MNEMODIFYOP | DIR2, 0,
    "ROLA ",  0x49, INH, 0,
    "ROLB ",  0x59, INH, 0,
    "ROLD ",  0x49, INH | PAGE1, HD6309,
    "ROLW ",  0x59, INH | PAGE1, HD6309,
    "ROR ",   0x46, MNEMODIFYOP | DIR2, 0,
    "RORA ",  0x46, INH, 0,
    "RORB ",  0x56, INH, 0,
    "RORD ",  0x46, INH | PAGE1, HD6309,
    "RORW ",  0x56, INH | PAGE1, HD6309,
    "RTI ",   0x3B, INH, 0,
    "RTS ",   0x39, INH, 0,
    "SBA ",   0x2a, TRANS, MC6809,                      // 6800
    "SBA ",   0x2b, TRANS, HD6309,                      // 6800
    "SBCA ",  0x82, MNEREADOP | IMM8, 0,
    "SBCB ",  0xC2, MNEREADOP | IMM8, 0,
    "SBCD ",  0x82, MNEREADOP | IMM16 | PAGE1, HD6309,
    "SBCR ",  0x33, EXG | PAGE1, HD6309,
    "SEC ",   0x0B, TRANS, 0,
    "SEF ",   0x17, TRANS, 0,
    "SEI ",   0x0C, TRANS, 0,
    "SEIF ",  0x25, TRANS, 0,
    "SEV ",   0x0D, TRANS, 0,
    "SEX ",   0x1D, INH, 0,
    "SEXW ",  0x14, INH, HD6309,
    "STA ",   0x87, MNEWRITEOP | NOIMM, 0,
    "STAA ",  0x87, MNEWRITEOP | NOIMM, 0,
    "STAB ",  0xC7, MNEWRITEOP | NOIMM, 0,
    "STB ",   0xC7, MNEWRITEOP | NOIMM, 0,
    "STBT ",  0x37, MNEMODIFYOP | BAND | PAGE2, HD6309,
    "STD ",   0xCD, MNEWRITEOP | NOIMM, 0,
    "STE ",   0x87, MNEWRITEOP | NOIMM | PAGE2, HD6309,
    "STF ",   0xC7, MNEWRITEOP | NOIMM | PAGE2, HD6309,
    "STQ ",   0xCD, MNEWRITEOP | NOIMM | PAGE1, HD6309,
    "STS ",   0xCF, MNEWRITEOP | NOIMM | PAGE1, 0,
    "STU ",   0xCF, MNEWRITEOP | NOIMM, 0,
    "STW ",   0x87, MNEWRITEOP | NOIMM | PAGE1, HD6309,
    "STX ",   0x8F, MNEWRITEOP | NOIMM, 0,
    "STY ",   0x8F, MNEWRITEOP | NOIMM | PAGE1, 0,
    "SUBA ",  0x80, MNEREADOP | IMM8, 0,
    "SUBB ",  0xC0, MNEREADOP | IMM8, 0,
    "SUBD ",  0x83, MNEREADOP | IMM16, 0,
    "SUBE ",  0x80, MNEREADOP | IMM8 | PAGE2, HD6309,
    "SUBF ",  0xC0, MNEREADOP | IMM8 | PAGE2, HD6309,
    "SUBR ",  0x32, EXG | PAGE1, HD6309,
    "SUBW ",  0x80, MNEREADOP | IMM16 | PAGE1, HD6309,
    "SWI ",   0x3F, INH, 0,
    "SWI2 ",  0x3F, INH | PAGE1, 0,
    "SWI3 ",  0x3F, INH | PAGE2, 0,
    "SYNC ",  0x13, INH, 0,
    "TAB ",   0x20, TRANS, 0,
    "TAP ",   0x0E, TRANS, 0,
    "TBA ",   0x21, TRANS, 0,
    "TFM ",   0x38, TFM | PAGE2, HD6309,
    "TFR ",   0x1F, EXG, 0,
    "TIM ",   0x4B, MNEMODIFYOP | IMMADR, HD6309,
    "TPA ",   0x0F, TRANS, 0,
    "TST ",   0x4D, MNEREADOP | DIR2, 0,
    "TSTA ",  0x4D, INH, 0,
    "TSTB ",  0x5D, INH, 0,
    "TSTD ",  0x4D, INH | PAGE1, HD6309,
    "TSTE ",  0x4D, INH | PAGE2, HD6309,
    "TSTF ",  0x5D, INH | PAGE2, HD6309,
    "TSTW ",  0x5D, INH | PAGE1, HD6309,
    "TSX ",   0x10, TRANS, 0,
    "TSY ",   0x11, TRANS, 0,
    "TXS ",   0x12, TRANS, 0,
    "TYS ",   0x13, TRANS, 0,
    "XGDX ",  0x14, TRANS, 0,
    "XGDY ",  0x15, TRANS, 0,

    "ALIGN ", AD_ALIG, ASMDIR, 0,
    "CPU ",   AD_CPU,  ASMDIR, 0,
    "DATA ",  AD_DATA, ASMDIR, 0,
    "DC.B ",  AD_FCC,  ASMDIR | DIR_GENERATE, 0,
    "DC.W ",  AD_FDB,  ASMDIR | DIR_GENERATE, 0,
    "DC.T ",  AD_FTB,  ASMDIR | DIR_GENERATE, 0,
    "DC.L ",  AD_FQB,  ASMDIR | DIR_GENERATE, 0,
//    "DS   ",  AD_FCC,  ASMDIR | DIR_GENERATE, 0,
    "DW ",    AD_FDB,  ASMDIR | DIR_GENERATE, 0,
    "DT ",    AD_FTB,  ASMDIR | DIR_GENERATE, 0,
    "DQ ",    AD_FQB,  ASMDIR | DIR_GENERATE, 0,
    "DS ",    AD_RMB,  ASMDIR, 0,
    "END ",   AD_END,  ASMDIR, 0,
    "ENDM ",  AD_ENDM, ASMDIR, 0,
    "ENDS",   AD_ENDSTRUCT, ASMDIR, 0,
    "EQU ",   AD_EQU,  ASMDIR | DIR_ASSIGN, 0,
    "EXIT ",  AD_EXIT, ASMDIR, 0,
    "FCB ",   AD_FCB,  ASMDIR | DIR_GENERATE, 0,
    "FCC ",   AD_FCC,  ASMDIR | DIR_GENERATE, 0,
    "FDB ",   AD_FDB,  ASMDIR | DIR_GENERATE, 0,
    "FTB ",   AD_FTB,  ASMDIR | DIR_GENERATE, 0,
    "FQB ",   AD_FQB,  ASMDIR | DIR_GENERATE, 0,
    "IF ",    AD_IF,   ASMDIR | DIR_CONDITIONAL, 0,
    "IFC ",   AD_IFC,  ASMDIR | DIR_CONDITIONAL, 0,
    "IFNC ",  AD_IFNC, ASMDIR | DIR_CONDITIONAL, 0,
    "IFEQ ",  AD_IFEQ, ASMDIR | DIR_CONDITIONAL, 0,
    "IFNE ",  AD_IFNE, ASMDIR | DIR_CONDITIONAL, 0,
    "IFGT ",  AD_IFGT, ASMDIR | DIR_CONDITIONAL, 0,
    "IFGE ",  AD_IFGE, ASMDIR | DIR_CONDITIONAL, 0,
    "IFLT ",  AD_IFLT, ASMDIR | DIR_CONDITIONAL, 0,
    "IFLE ",  AD_IFLE, ASMDIR | DIR_CONDITIONAL, 0,
    "ELSE ",  AD_ELSE, ASMDIR | DIR_CONDITIONAL, 0,
    "ELSEC ", AD_ELSEC, ASMDIR | DIR_CONDITIONAL, 0,
    "ENDC ",  AD_ENDC, ASMDIR | DIR_CONDITIONAL, 0,
    "INCL",   AD_INCL, ASMDIR, 0,
    "MACR",   AD_MACRO, ASMDIR | DIR_ASSIGN, 0,
    "NAM ",   AD_NAM,  ASMDIR, 0,
    "OPT ",   AD_OPT,  ASMDIR, 0,
    "ORG ",   AD_ORG,  ASMDIR, 0,
    "PAG",    AD_PAGE, ASMDIR, 0,
    "REG ",   AD_REG,  ASMDIR | DIR_ASSIGN, 0,
    "RMB ",   AD_RMB,  ASMDIR, 0,
    "SECTION ", AD_SECTION, ASMDIR, 0,
    "SET ",   AD_SET,  ASMDIR | DIR_ASSIGN, 0,
    "SETDP ", AD_SETDP, ASMDIR, 0,
    "SPC ",   AD_PAGE, ASMDIR, 0,        // ignored directive
    "STRUCT ", AD_STRUCT, ASMDIR | DIR_ASSIGN, 0,
    "TEXT ",  AD_TEXT, ASMDIR, 0,
    "TTL ",   AD_TTL,  ASMDIR, 0,
    "ZMB ",   AD_ZMB,  ASMDIR | DIR_GENERATE, 0,
    "BSZ ",   AD_ZMB,  ASMDIR | DIR_GENERATE, 0,
    "FAIL ",  AD_FAIL, ASMDIR, 0,
    "XDEF ",  AD_XDEF, ASMDIR, 0,
    "XREF ",  AD_XREF, ASMDIR, 0,
    ".BYTE ", AD_FCC,  ASMDIR | DIR_GENERATE, 0,
    ".WORD ", AD_FDB,  ASMDIR | DIR_GENERATE, 0,
    ".RWORD ",AD_FDBR, ASMDIR | DIR_GENERATE, 0,
    ".TRIP",  AD_FTB,  ASMDIR | DIR_GENERATE, 0,  // TRIPLE
    ".LONG ", AD_FQB,  ASMDIR | DIR_GENERATE, 0,
    ".QUAD ", AD_FQB,  ASMDIR | DIR_GENERATE, 0,
    ".ERR",   AD_ERROR, ASMDIR, 0,
    ".MSG ",  AD_INFO, ASMDIR, 0,
    ".MESS",  AD_INFO, ASMDIR, 0,
    ".WARN",  AD_WARN, ASMDIR, 0,
    ".GENLAB ", AD_GENLAB, ASMDIR, 0,   // generate and push label
    ".POP ",  AD_POP, ASMDIR, 0,
    ".EXCH ", AD_EXCH, ASMDIR, 0,
    "*", 0, -1, 0,
    "", 0, -1, 0
};

/*  count, form, opcode          */
TRNSL transtab[] = {
    0x0102,  0, 0x00, 0x00, 0, 0,       /* 0                           */
    0x0102,  0, 0x31, 0xA5, 0, 0,       /* 1  ABY   LEAY  B,Y          */
    0x0102,  0, 0x1C, 0xFE, 0, 0,       /* 2  CLC   ANDCC #$FE         */
    0x0102,  0, 0x1C, 0xEF, 0, 0,       /* 3  CLI   ANDCC #$EF         */
    0x0102,  0, 0x1C, 0xFD, 0, 0,       /* 4  CLV   ANDCC #$FD         */
    0x0102,  0, 0x30, 0x1F, 0, 0,       /* 5  DEX   LEAX  -1,X         */
    0x0102,  0, 0x30, 0x01, 0, 0,       /* 6  INX   LEAX  1,X          */
    0x0102,  0, 0x31, 0x3F, 0, 0,       /* 7  DEY   LEAY  -1,Y         */
    0x0102,  0, 0x31, 0x21, 0, 0,       /* 8  INY   LEAY  1,Y          */
    0x0102,  0, 0x32, 0x7F, 0, 0,       /* 9  DES   LEAS  -1,S         */
    0x0102,  0, 0x32, 0x61, 0, 0,       /* a  INS   LEAS  1,S          */
    0x0102,  0, 0x1A, 0x01, 0, 0,       /* b  SEC   ORCC  #$01         */
    0x0102,  0, 0x1A, 0x10, 0, 0,       /* c  SEI   ORCC  #$10         */
    0x0102,  0, 0x1A, 0x02, 0, 0,       /* d  SEV   ORCC  #$02         */
    0x0102,  0, 0x1F, 0x8A, 0, 0,       /* e  TAP   TFR   A,CC         */
    0x0102,  0, 0x1F, 0xA8, 0, 0,       /* f  TPA   TFR   CC,A         */
    0x0102,  0, 0x1F, 0x41, 0, 0,       /*10  TSX   TFR   S,X          */
    0x0102,  0, 0x1F, 0x42, 0, 0,       /*11  TSY   TFR   S,Y          */
    0x0102,  0, 0x1F, 0x14, 0, 0,       /*12  TXS   TFR   X,S          */
    0x0102,  0, 0x1F, 0x24, 0, 0,       /*13  TYS   TFR   Y,S          */
    0x0102,  0, 0x1E, 0x01, 0, 0,       /*14  XGDX  EXG   D,X          */
    0x0102,  0, 0x1E, 0x02, 0, 0,       /*15  XGDY  EXG   D,Y          */
    0x0102,  0, 0x1C, 0xBF, 0, 0,       /*16  CLF   ANDCC #$BF         */
    0x0102,  0, 0x1A, 0x40, 0, 0,       /*17  SEF   ORCC  #$40         */
    0x0102,  0, 0x34, 0x02, 0, 0,       /*18  PSHA  PSHS  A            */
    0x0102,  0, 0x34, 0x04, 0, 0,       /*19  PSHB  PSHS  B            */
    0x0102,  0, 0x34, 0x10, 0, 0,       /*1a  PSHX  PSHS  X            */
    0x0102,  0, 0x34, 0x20, 0, 0,       /*1b  PSHY  PSHS  Y            */
    0x0102,  0, 0x35, 0x02, 0, 0,       /*1c  PULA  PULS  A            */
    0x0102,  0, 0x35, 0x04, 0, 0,       /*1d  PULB  PULS  B            */
    0x0102,  0, 0x35, 0x10, 0, 0,       /*1e  PULX  PULS  X            */
    0x0102,  0, 0x35, 0x20, 0, 0,       /*1f  PULY  PULS  Y            */
    0x0102,  0, 0x1F, 0x89, 0, 0,       /*20  TAB   TFR   A,B          */
    0x0102,  0, 0x1F, 0x98, 0, 0,       /*21  TBA   TFR   B,A          */
    0x0102,  0, 0x58, 0x49, 0, 0,       /*22  LSLD  LSLB;ROLA          */
    0x0102,  0, 0x44, 0x56, 0, 0,       /*23  LSRD  LSRA;RORB          */
    0x0102,  0, 0x1C, 0xAF, 0, 0,       /*24  CLIF  ANDCC #$AF         */
    0x0102,  0, 0x1A, 0x50, 0, 0,       /*25  SEIF  ORCC  #$50         */
    0x0004, -2, 0x34, 0x04, 0xab, 0xe0, /*26  ABA   PSHS B; ADDA ,S+      */
    0x0203,  0, 0x10, 0x30, 0x98, 0,    /*27  ABA   ADDR B,A           HD */
    0x0004, -2, 0x34, 0x04, 0xa1, 0xe0, /*28  CBA   PSHS B; CMPA ,S+      */
    0x0203,  0, 0x10, 0x37, 0x98, 0,    /*29  CBA   CMPR B,A           HD */
    0x0004, -2, 0x34, 0x04, 0xa0, 0xe0, /*2a  SBA   PSHS B; SUBA ,S+      */
    0x0203,  0, 0x10, 0x32, 0x98, 0,    /*2b  SBA   SUBR B,A           HD */
};

/*
 * Old 6800 assemblers allowed a space (or maybe more) before the register
 * specification, this table is included in a hack to take care of that.
 *
 */
MneFix mnefixtbl[] = {
    { "ADC ", "AB" },
    { "ADD ", "AB" },
    { "AND ", "AB" },
    { "ASL ", "AB" },
    { "ASR ", "AB" },
    { "BIT ", "AB" },
    { "CLR ", "AB" },
    { "CMP ", "AB" },
    { "COM ", "AB" },
    { "DEC ", "AB" },
    { "EOR ", "AB" },
    { "INC ", "AB" },
    { "LDA ", "AB" },
    { "LSL ", "AB" },
    { "LSR ", "AB" },
    { "NEG ", "AB" },
    { "ORA ", "AB" },
    { "PSH ", "AB" },
    { "PUL ", "AB" },
    { "ROL ", "AB" },
    { "ROR ", "AB" },
    { "SBC ", "AB" },
    { "STA ", "AB" },
    { "SUB ", "AB" },
    { "TST ", "AB" },
    {"", ""}
};


BranchOp braoptab[] = {
//      opcode    page
//    short long  s  l  z
    { 0x20, 0x16, 0, 0, 1 },      // "BRA ", 
//    { 0x21, 0x21, 0, 1, 1 },      // "BRN ", 
    { 0x22, 0x22, 0, 1, 1 },      // "BHI ", 
    { 0x23, 0x23, 0, 1, 1 },      // "BLS ", 
    { 0x24, 0x24, 0, 1, 1 },      // "BCC ", 
    { 0x25, 0x25, 0, 1, 1 },      // "BCS ", 
    { 0x26, 0x26, 0, 1, 1 },      // "BNE ", 
    { 0x27, 0x27, 0, 1, 1 },      // "BEQ ", 
    { 0x28, 0x28, 0, 1, 1 },      // "BVC ", 
    { 0x29, 0x29, 0, 1, 1 },      // "BVS ", 
    { 0x2A, 0x2A, 0, 1, 1 },      // "BPL ", 
    { 0x2B, 0x2B, 0, 1, 1 },      // "BMI ", 
    { 0x2C, 0x2C, 0, 1, 1 },      // "BGE ", 
    { 0x2D, 0x2D, 0, 1, 1 },      // "BLT ", 
    { 0x2E, 0x2E, 0, 1, 1 },      // "BGT ", 
    { 0x2F, 0x2F, 0, 1, 1 },      // "BLE ", 
    { 0x8D, 0x17, 0, 0, 0 },      // "BSR ", 
    { 0,    0,   -1, -1, 0 }      // endmark
};

REGNAM pshsreg[] = {
	"PC", 0x80, 0,
	"U",  0x40, 0,
	"Y",  0x20, 0,
	"X",  0x10, 0,
	"D",  0x06, 0,
	"DP", 0x08, 0,
	"DPR",0x08, 0,
	"B",  0x04, 0,
	"A",  0x02, 0,
	"CC", 0x01, 0,
	"CCR",0x01, 0,
	"",   -1, 0
};

REGNAM pshureg[] = {
	"PC", 0x80, 0,
	"S",  0x40, 0,
	"Y",  0x20, 0,
	"X",  0x10, 0,
	"D",  0x06, 0,
	"DP", 0x08, 0,
	"DPR",0x08, 0,
	"B",  0x04, 0,
	"A",  0x02, 0,
	"CC", 0x01, 0,
	"CCR",0x01, 0,
	"",   -1, 0
};  
  
REGNAM exgreg[] = {
  	"D",   0, 0,
  	"X",   1, 0,
 	"Y",   2, 0,
	"U",   3, 0,
	"S",   4, 0,
	"PC",  5, 0,
	"W",   6, HD6309,
	"V",   7, HD6309,
	"A",   8, 0,
	"B",   9, 0,
	"CC", 10, 0,
	"CCR",10, 0,
	"DP", 11, 0,
	"DPR",11, 0,
	"0",  12, HD6309,
	"E",  14, HD6309,
	"F",  15, HD6309,
	"",   -1, 0
};

REGNAM indofsreg[] = {
  	"A",   1, 0,
 	"B",   2, 0,
	"D",   3, 0,
	"E",   4, HD6309,
	"F",   5, HD6309,
	"W",   6, HD6309,
	"",   -1, 0
};

REGNAM indreg[] = {
  	"X",   0, 0,
 	"Y",   1, 0,
	"U",   2, 0,
	"S",   3, 0,
	"PCR", 8, 0,
	"PC",  9, 0,
	"W",  32, HD6309,
	"",   -1, 0
};

REGNAM tfmreg[] = {
  	"D",   0, 0,
  	"X",   1, 0,
 	"Y",   2, 0,
	"U",   3, 0,
	"S",   4, 0,
	"",   -1, 0
};

REGNAM bandreg[] = {
	"A",   1, 0,
	"B",   2, 0,
	"CC",  0, 0,
	"",   -1, 0
};

char *ccflagstr = "EFHINZVC";
REGNAM ccflags[] = {
	"E", 0x80, 0,
	"F", 0x40, 0,
	"H", 0x20, 0,
	"I", 0x10, 0,
	"N", 0x08, 0,
	"Z", 0x04, 0,
	"V", 0x02, 0,
	"C", 0x01, 0,
	"",   -1, 0
};


unsigned short mnemiccnttab[sizeof(mnemic)/sizeof(MneType)];

long getmnemicsize()
{

   return sizeof(mnemic);
}

/* end */

