/*******************************/
/*   assembler for 68xx        */
/*       Main part             */
/*******************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    930818      Ported from Atari to Unix on Sun
*
*  0.02    941019   1. files asm03man.c and asm03ma2.c concatenated,
*                       asmcom.c created.
*                   2. tabs replaced with spaces.
*
*  0.03    970314   1. New output formats: Motorola S-rec, Intel hex and
*                       prepared for Tectronics hex.
*                   2. Lots of new command line options added to support
*                       the new format and to specify the output file names.
*                   3. Version printout added.
*
*  0.04    980929   1. A colon ':' accepted after label.
*
*  0.05    040113   1. A semicolon ';' accepted as start of comment char
*
*  0.06    040817   1. New option -t tablist: set tabs
*
*  0.07    091124   1. ??
*
*  0.08    120517   1. accept spaces after comma for fcb and fdb
*                   2. asmdir ZMB added, some minor changes made to get the
*                       printout nice
*                   3. <stdlib.h> included to make gcc on linux happier
*                   4. option -noprio: set prio of * same as +
*
*  0.09    161025   1. asmdir FQB added
*                   2. Output format of FCB,FCC,FDB changed
*
*
*  0.10    170124   1. Pass variable: FIRST_PASS, MIDDLE_PASS and LAST_PASS
*                   2. flag for symbol changed added
*                   3. Multipass, the MIDDLE_PASS runs until symbols not changed
*                   4. Conditional directives added: IF, IFC etc.
*
*  0.11    171007   1. Input files read to internal data during first pass
*                   2. Macros added: MACR, EXIT, ENDM. Including local labels
*                       '\.n' and params '\n' or '\{n}' where <n> is a value
*                   3. Support for control flow added: Generate unique labels
*                       Label stack, dir .GENLAB, .POP, .EXCH and read of label
*                       stack with '\:[n]'.
*
*  0.12    180127   1. Asmdir FTB added (Form Triple Byte)
*                   2. FCC no longer takes any char as delimiter.
*                       Limited to / \ " ' `
*                   3. Added test in readasm() to catch the last line
*
*  0.13    180926   1. Now handle mnemonic of type: LDA A ...
*                   2. FCC takes form:  length,text
*
*  0.14    181114   1. Asmdir .rword: store reversed word
*
*/
static char *version = "asmcom   0.14  2018-11-14";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "asmdef.h"

#define MAXTABS 32


NAME_S *inflh;
NAME_S *infle;


TXTHEAD srctxt;
TXTLINE *txtlp;

int cputypeflag = 0;        // cputypemask is not set
int nocondlist;
char *actfname;
int beautify = 0;
int xrefcnt = 0;
int recoutstate = 0;
int recordlen = 0;
int tabs = -1;
char tabtxt[MAXTXTLEN];
int tablist[MAXTABS];
int tabcnt = 0;
char errorstring[120];

int noifelse = 0;
int condskip = 0;
int condlevel = 0;
int skiplevel = 0;

int macrodef;
int showmacros;

extern int macroLevel;
extern int macexpand;
extern MneFix mnefixtbl[];


int addInfile(char *name);
int setcputype(char *name);
int asmline();
void parsetablist();
void settab();
int asmdirective();
int assigndirective(int opcode, int inptrsav);
int conddirective(int opcode, int inptrsav);
void dircpu();
int dirzmb();
void dirfqb();
void dirftb();
void dirfcb();
void dirfcc();
void dirfdb();
void dirfdbr();
void diralign();
void dirorg();
void dirend();
void dirttl();
void dir_error(int);
void dir_fail(int);
void diropt();
void dirpage();
void dirincl();
void dirrmb();
void dir_xref();

long factor2();
int dirgenlab();
int dirpop(int);
int direxch(int);
void dirstruct();
void direndstruct();


void prntout();
void prntlst();
void prnterr();
void prntbin();
void prntbuff();
void prntsymtab();
void prntxreftab();
void prntstatistics();
void printmacros(FILE* hdl);
void printmacstat(FILE* hdl);
void toterrprnt();
void printChangedSymbols();

int readasm();
void buffbin();
int isadrfield();
int open_listfile(char *name);
void close_files();
void beautifylist(char *name, char *targetname);
SECTION *addSection(char *name, int type, int prgmode);
void clearSections();
void clearSymbols();
int copystring(char *d, int len);
void storetxt(TXTHEAD *headp, int lineno, char *txt, char *fname);
int readMacroDef();
int expandMacro();
int checkMacroDef();
int checkStructDef();
void initPassExp();
void afterPassExp();
void expandShortFormLabel(char *dest);
int insertChar(char *txt, int count);
int isin(int c, char *s);
SymType *definevar(int symclass, int flag);
void addVarToStruct(char *name);


char *comversion()
{

    return version;
}

void printver()
{

    fprintf(hdl_scr, "\n %s: versions\n", cmdbase);
    fprintf(hdl_scr, "   %s\n", defversion);
    fprintf(hdl_scr, "   %s\n", manversion());
    fprintf(hdl_scr, "   %s\n", tblversion());
    fprintf(hdl_scr, "   %s\n", comversion());
    fprintf(hdl_scr, "   %s\n", expversion());
    fprintf(hdl_scr, "   %s\n\n", filversion());

}
void printsizes()
{

    printf("short*:  %d\n", (int) sizeof(short*));
    printf("short :  %d\n", (int) sizeof(short));
    printf("int   :  %d\n", (int) sizeof(int   ));
    printf("long  :  %d\n", (int) sizeof(long  ));
    printf("long l:  %d\n", (int) sizeof(long long int));

    printf("MneType: %d\n", (int) sizeof(MneType));
    printf("REGNAM:  %d\n", (int) sizeof(REGNAM));
    printf("SymType: %d\n", (int) sizeof(SymType));

}

void usage()
{

    fprintf(hdl_scr, "\nusage: %s [-v] [-h] [-cre] [-x] [-stat] [-o[par] [objfile]]\n",
            cmdbase);
    fprintf(hdl_scr, "       [-l [listfile]] [-] infile\n\n");
    fprintf(hdl_scr, " -v            Print versions\n");
    fprintf(hdl_scr, " -h            Print help\n");
    fprintf(hdl_scr, " -b            Beautify source\n");
    fprintf(hdl_scr, " -cpu name     Set CPU type\n");
    fprintf(hdl_scr, " -cre or -x    Generate cross reference list\n");
    fprintf(hdl_scr, " -braopt       Optimize short/long branches, and\n");
//    fprintf(hdl_scr, " -braopt[z]    Optimize short/long branches, and\n");
//    fprintf(hdl_scr, "                 with <z> also remove branches with zero offset\n");
    fprintf(hdl_scr, " -ncl or       Do not list conditional directives\n");
    fprintf(hdl_scr, " -nocondlist\n");
    fprintf(hdl_scr, " -maccom       Include comment in macro expansion\n");
    fprintf(hdl_scr, " -noprio       No priority for * / etc in expressions (useful for source prepared for as09)\n");
    fprintf(hdl_scr, " -noif         Do not interpret IF or ELSE in conditional directives\n");
    fprintf(hdl_scr, " -noforwopt    No forward optimization of PC-relative addresses\n");
    fprintf(hdl_scr, " -stat         Generate statistics of used mnemonics\n");
    fprintf(hdl_scr, " -opar [objfile]  Object file (file name optional) where <par> is:\n");
    fprintf(hdl_scr, "                 s    object file in Motorola s19 format (default)\n");
    fprintf(hdl_scr, "                 i    object file in Intel hex format\n");
    fprintf(hdl_scr, "                 t    object file in Tectronic format (not yet impl.)\n");
    fprintf(hdl_scr, "                 o    object file in own format\n");
    fprintf(hdl_scr, "                   and in addition to the 's','i' or 't' the length\n");
    fprintf(hdl_scr, "                   of the records can be appended after the letter.\n");
    fprintf(hdl_scr, " -l [listfile]  Listing to this file, otherwise to stdout,\n");
    fprintf(hdl_scr, "                 if no file name is given the input file\n");
    fprintf(hdl_scr, "                 with the extension changed to .lst is used\n");
    fprintf(hdl_scr, "               If any of the file names above are not given, the\n");
    fprintf(hdl_scr, "               extension from the input file is chopped of and\n");
    fprintf(hdl_scr, "               a '.s19', '.obj' or '.lst' is added\n");
    fprintf(hdl_scr, " -             Just in case\n");
    fprintf(hdl_scr, " -t tablist    Specify tabs\n");
    fprintf(hdl_scr, " infile...     Input file(s)\n");
    fprintf(hdl_scr, "\n");
    fprintf(hdl_scr, "  Some of the parameters above can be set by the OPT directive in\n");
    fprintf(hdl_scr, "  the  assembler list:\n");
    fprintf(hdl_scr, "\n");
    fprintf(hdl_scr, "\n");





}

enum {
    NXTARG_LST = 1,
    NXTARG_OBJ,
    NXTARG_TAB,
    NXTARG_CPU,
};

int GetParameter(int argc, char *argv[])
{
    int i, j;
    int nxtarg;
    int lastarg;
    char *p;

    xreflist = 0;
    statistics = 0;

    cmdbase = argv[0];

    lstfilename[0] = 0;
    objfilename[0] = 0;
    objfileformat = 0;
    files[0].filesname[0] = 0;

    nxtarg = 0;
    lastarg = 0;
    for (j = 0, i = 1; i < argc; i++) {
        if (*argv[i] == '-') {
            nxtarg = 0;
            lastarg = 0;
            if (argv[i][1] == 0) {
            }
            else if (!strcmp(argv[i]+1, "l")) {
                listflag = 2;
                nxtarg = NXTARG_LST;
            }
            else if (argv[i][1] == 'o') {
                if (argv[i][2] == 0) {
                    objfileformat = S19OBJ;
                    recordlen = S19LEN;
                }
                else if (argv[i][2] == 'o') {
                    objfileformat = OWNOBJ;
                }
                else if (argv[i][2] == 's') {
                    objfileformat = S19OBJ;
                    recordlen = 0;
                    if (isdigit(argv[i][3])) {
                        recordlen = strtol(argv[i]+3, 0, 10);
                    }
                    if (recordlen < 1) {
                        recordlen = S19LEN;
                    }
                    else if (recordlen > 120) {
                        recordlen = 120;
                    }
                }
                else if (argv[i][2] == 'i') {
                    objfileformat = INTOBJ;
                    recordlen = 0;
                    if (isdigit(argv[i][3])) {
                        recordlen = strtol(argv[i]+3, 0, 10);
                    }
                    if (recordlen < 1) {
                        recordlen = INTLEN;
                    }
                    else if (recordlen > 120) {
                        recordlen = 120;
                    }
                }
                else if (argv[i][2] == 't') {
                    objfileformat = TECOBJ;
                    recordlen = 0;
                    if (isdigit(argv[i][3])) {
                        recordlen = strtol(argv[i]+3, 0, 10);
                    }
                    if (recordlen < 1) {
                        recordlen = TECLEN;
                    }
                    else if (recordlen > 120) {
                        recordlen = 120;
                    }
                }
                nxtarg = NXTARG_OBJ;
            }
            else if (!strcmp(argv[i]+1, "h")) {
                usage();
                exit(0);
            }
            else if (!strcmp(argv[i]+1, "v")) {
                printver();
                exit(0);
            }
            else if (!strncmp(argv[i]+1, "cpu", 3)) {
                if (argv[i][4]) {
                    if (setcputype(argv[i]+4) == 0) {
                        fprintf(hdl_scr, "CPU type not found\n");
                    }
                }
                else {
                    nxtarg = NXTARG_CPU;
                }
            }
            else if (!strcmp(argv[i]+1, "cre") || !strcmp(argv[i]+1, "x")) {
                xreflist = 1;
            }
            else if (!strcmp(argv[i]+1, "noprio")) {
                mulnoprio = 1;
            }
            else if (!strcmp(argv[i]+1, "noif")) {
                noifelse = 1;
            }
            else if (!strcmp(argv[i]+1, "nocondlist") || !strcmp(argv[i]+1, "ncl")) {
                nocondlist = 1;
            }
            else if (!strcmp(argv[i]+1, "stat")) {
                statistics = 1;
            }
            else if (!strcmp(argv[i]+1, "pmac")) {
                showmacros = 1;
            }
            else if (!strcmp(argv[i]+1, "nomaccom")) {
                macexpand &= ~1;
            }
            else if (!strcmp(argv[i]+1, "maccom")) {
                macexpand |= 1;
            }
            else if (!strncmp(argv[i]+1, "braopt", 6)) { /* optimize long/short branch */
                p = argv[i] + 7;
                optimizebranch = 1;
                if (tolower(*p) == 'z') {
                    p++;
                    optimizebranch |= 0x100;
                }
                if (isdigit(*p)) {
                    optimizebranch |= strtol(p, &p, 10) & 255;
                }
                fprintf(hdl_scr, "%s: option braopt %d\n", cmdbase, optimizebranch);
            }
            else if (!strcmp(argv[i]+1, "noforwopt")) { /* No forward optimization */
                noforwardopt = 1;
            }
            else if (!strcmp(argv[i]+1, "b")) {
                beautify = 1;
            }
            else if (argv[i][1] == 't') {
                p = argv[i] + 2;
                if (*p == 0) {
                    nxtarg = NXTARG_TAB;
                }
                else {
                    parsetablist(p);
                }
            }
            else if (!strcmp(argv[i]+1, "size")) {
                printsizes();
                exit(0);
            }
            else {
                fprintf(hdl_scr, "%s: Unknown option '%s'\n", cmdbase, argv[i]);
                usage();
                exit(4);
            }
        }
        else {
            p = argv[i];
            switch (nxtarg) {
            case 0:
                addInfile(p);
                lastarg = 0;
                break;
            case NXTARG_LST:
                strcpy(lstfilename, p);
                nxtarg = 0;
                lastarg = NXTARG_LST;
                break;
            case NXTARG_OBJ:
                strcpy(objfilename, p);
                nxtarg = 0;
                lastarg = NXTARG_OBJ;
                break;
            case NXTARG_TAB:
                parsetablist(p);
                nxtarg = 0;
                break;
            case NXTARG_CPU:
                if (setcputype(p)) {
                    fprintf(hdl_scr, "CPU type not found\n");
                }
                nxtarg = 0;
                break;
            }
        }
    }

    if (inflh == 0) {
        fprintf(hdl_scr, "%s: Input file missing\n", cmdbase);
        usage();
        exit(4);
    }

    if (objfileformat == 0) {
//        objfileformat = OWNOBJ;
        objfileformat = S19OBJ;
        recordlen = S19LEN;
    }

    settab();
    return j;
}

int addInfile(char *name)
{
    NAME_S *np;

    np = malloc(sizeof(NAME_S));

    memset(np, 0, sizeof(NAME_S));

    np->name = strdup(name);

    if (inflh == 0) {
        inflh = np;
        infle = np;
    }
    else {
        infle->nxt = np;
        infle = np;
    }

    return 0;
}

void settab()
{
    int i;
    int val;

    if (tabs == -1) {
        val = 8;

        tabtxt[0] = 1;
        for (i = 1; i < MAXTXTLEN; i++) {
            if ((i % val) == 0) {
                tabtxt[i] = 0;  /* tab-pos */
                if (tabcnt < MAXTABS) {
                    tablist[tabcnt++] = i;
                }
            }
            else {
                tabtxt[i] = 1;
            }
        }
    }
}

void parsetablist(char *p)
{
    int i;
    int val;
    int type;
    int pos;

    for (i = 0; i < MAXTXTLEN; i++) {
        tabtxt[i] = 1;
    }

    type = 0;                   /* relative default */
    if (tolower(*p) == 'a') {
        p++;
        type = 1;               /* absolute */
    }
    else if (tolower(*p) == 'r') {
        p++;
        type = 0;               /* relative */
    }
    else {
    }

    val = 8;
    if (type == 0) {            /* relative */
        i = 0;
        pos = 0;
        while (*p) {
            if (isdigit(*p)) {
                if (*p == '=') { /* absolute insertion */
                    p++;
                    val = strtol(p, &p, 10);
                    pos = val;
                }
                else {
                    val = strtol(p, &p, 10);
                    pos += val;
                }
                if (pos >= MAXTXTLEN) {
                    break;
                }
                tabtxt[pos] = 0;
                if (tabcnt < MAXTABS) {
                    tablist[tabcnt++] = pos;
                }
                i++;
            }
            else {
                break;
            }

            if (*p == ',') {
                p++;
            }
        }
    }
    else {                      /* absolute */
        i = 0;
        pos = 0;
        while (*p) {
            if (isdigit(*p)) {
                if (*p == '+') { /* relative insertion */
                    p++;
                    val = strtol(p, &p, 10);
                    pos += val;
                }
                else {
                    val = strtol(p, &p, 10);
                    pos = val;
                }
                if (pos >= MAXTXTLEN) {
                    break;
                }
                tabtxt[pos] = 0;
                if (tabcnt < MAXTABS) {
                    tablist[tabcnt++] = pos;
                }
                i++;
            }
            else {
                break;
            }

            if (*p == ',') {
                p++;
            }
        }
    }

    if (i <= 1) {               /* none or only one pos set, assume val to be repeated */
        if (val < 1) {
            val = 1;
        }
        tabs = val;
        pos = 1;
    }
    else {
        val = 8;
    }

    for (i = pos; i < MAXTXTLEN; i++) {
        if ((i % val) == 0) {
            tabtxt[i] = 0;
            if (tabcnt < MAXTABS) {
                tablist[tabcnt++] = i;
            }
        }
    }

}

void init()
{

    lbwrn = 0;
    listflag = 1;

    symbolptr = 0;
    linelen = 101;
/*   hdl_scr = fopen("CON:", "w"); */
    hdl_scr = stdout;

    cputypemask = 1;
    mulnoprio = 0;

    textsection = addSection("TEXT", SECT_TEXT, PRGTEXT);
    datasection = addSection("DATA", SECT_BSS, PRGDATA);

}

int iscomchar(int c)
{

    return c == '*' || c == ';';
}

void assemble(char *name, char *targetname)
{
    short prgmodeold;
    int passcnt = 1;
    SECTION * oldcursection;
    long *prgcntptr;
    int ret;
    char *p;

    if (beautify) {
        beautifylist(name, targetname);
        return;
    }

    strcpy(files[0].filesname, inflh->name);

    actfname = files[0].filesname;

    runadr = 0;
    symbmaxlen = 6;      /* shortest symbol length used in printout*/
    pass = FIRST_PASS;                        /* cleared once */
    fprintf(hdl_scr, "%s Assembler\n\n", targetname);
    fflush(hdl_scr);
    while (1) {
        if (!open_infile(actfname)) {
            return;
        }
        if (!open_outfile(actfname)) {
            return;
        }

        txtlp = srctxt.txtlh;

        endoffil = 0;
        errline = wrnline = 0;           /* cleared for each pass */
        errcnt = wrncnt = 0;
        infocnt = 0;
        infoline = 0;
        errlinelow = -1;
        wrnlinelow = -1;
        infolinelow = -1;

        condskip = 0;
        condlevel = 0;
        skiplevel = 0;

        line = 0;
        adrsizcnt = 0;

        prgmode = PRGTEXT;
        cursection = textsection;
        structdef = STRUCT_CLOSED;

        initPassExp();
        prgcntdata = prgcnttext = 0;
        prgcnt = startadr = 0;
        bufflg = 0;
        bufptr = 4;

        strncpy(buftab+4, actfname, strlen(actfname));
        bufptr += strlen(actfname);
        prntbuff(0);
        bufptr = 4;

        fprintf(hdl_scr, "Starting pass %d  %s\n", passcnt, pass == LAST_PASS ? "Last pass" : "");
        if (passcnt > 24) {
            fprintf(hdl_scr, "WARNING: max number of passes: %d\n", passcnt);
            pass = LAST_PASS;
        }

        while(1) { 
            errpos = 0;               /* cleared for each line */
            errout[0] = 0;
            txtptr = 0;
            lblflag = 0;
            aflag = dflag = rflag = 0;
            oplen = len = 0;

            if (macroLevel && expandMacro()) { // expand macro
                prntline = 0;          /* print no line number */
                txtptr = 1;
            }
            else {
                ret = readasm();          // read from input file
                prntline = line;          /* print this line number */
                if (ret == EOF) {    // end of input
                    break;
                }
                else if (ret == 2) { // include file closed
                    prntline = 0;          /* print no line number */
                }
                txtptr = 0;
                if (txt[txtptr] == '+' || txt[txtptr] == '-') { // line prefix
                    txtptr++;               // '+' show this line is from macro
                }                           // '-' don't show this line at all
            }

//            printf("%d '%s'\n", line, txt);


            if (structdef != STRUCT_CLOSED) {
                prgcntptr = &structprgcnt;
                prgmode = PRGDATA;
            }
            else {
                oldcursection = cursection;
                prgcntptr = &cursection->secprgcnt;
                prgmode = cursection->progmode;
            }

            prgcnt = *prgcntptr; // read the program counter
            adr = prgcnt;

            ret = 0;
            if (txt[txtptr] == 0) { /* empty line */
                ret = condskip ? 1 : 0;
            }
            else if (txt[txtptr] == '#') { /* assembly directive */
                ret = condskip ? 1 : 0;
            }
            else if (txt[txtptr] == '!') { /* debug comment from pass 1 */
                ret = 0;
            }
            else if (iscomchar(txt[txtptr])) { /* comment line */
                ret = condskip ? 1 : 0;
            }
            else {              /* it is a label/symbol definition */
                p = strchr(txt, '\\');
                if (p) {
                    expandShortFormLabel(txt);
                }
                if (txt[txtptr] != ' ') { /* line starts with a label */
                    copyvar();
                    lblflag = 1;
                }
                ret = asmline();
            }

            *prgcntptr = prgcnt;        // write back the updated program counter

//            if (structdef != STRUCT_CLOSED) {
//                structprgcnt = prgcnt;
//                if (structdef == STRUCT_CLOSING) {
//                    structdef = STRUCT_CLOSED;
//                }
//            }
//            else {
//                oldcursection->secprgcnt = prgcnt;
//            }
            if (ret == 0) {
                if (pass == FIRST_PASS) {   // store input in text list
//                    printf("add: %04d %s\n", prntline, txt);
                    storetxt(&srctxt, prntline, txt, actfname);
                }
                prntout();
            }
        }

        afterPassExp();
//        fprintf(hdl_scr, "Symbols changed: %d,ec=%d \n", symbolchanged, errcnt);
//        prntsymtab(hdl_scr);
        if (pass == LAST_PASS) {
            break;
        }
        else {
//            if (symbolchanged && symboldoubledef == 0) {
            if (pass == FIRST_PASS || (symbolchanged != 0 && errcnt == 0)) {
                fprintf(hdl_scr, "   Symbols changed: %d\n", symbolchanged);
                if (symbolchanged < 20) {
                    printChangedSymbols();
                }
                pass = MIDDLE_PASS;
            }
            else {
                pass = LAST_PASS;
            }
        }
        passcnt++;
//        printf("skip %d  c-lvl %d s-lvl %d\n", condskip, condlevel, skiplevel);
    }

//    printf("skip %d  c-lvl %d s-lvl %d\n", condskip, condlevel, skiplevel);
    if (bufflg) {
        prntbuff(1);
        bufflg = 0;
    }
    startadr = runadr;
    prntbuff(2);
    toterrprnt();


    if (xreflist) {
        prntxreftab(hdl_lst);
    }
    else {
        prntsymtab(hdl_lst);
    }

    if (showmacros) {
        printmacros(hdl_lst);
    }

    if (statistics) {
        prntstatistics(hdl_lst);
        printmacstat(hdl_lst);
    }

    close_files();

}


/*
 *
 *
 * return: =0: output list and obj, !=0: skip this line
 */
int asmline()
{
    int mnetyp, ptrsav;
    long val;
    int found = 0;
    int ret = 0;
    int skipped;

    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    ptrsav = txtptr;
    opref = 0;
//    printf("'%s'\n", txt+txtptr);
    mnetyp = mnefind(&opcode, &optype, &opref, cputypemask, 0);
    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    if (condskip == 0) {
        ret = 0;
//    printf("'%s' (%d)\n", txt+txtptr, mnetyp);
        if (mnetyp == 3) {
            if (!lblflag) {
                error(SYMBMISS, 0);
            }
            found = assigndirective(opcode, ptrsav);
        }
        else {
            if (lblflag) {
                definevar(SYMC_VAR, DEF_NORMAL);
                setvar(prgcnt, DEF_NORMAL);
                aflag = 1;          // print address
            }
            switch (mnetyp) {
            case -1:             /* end of line */
                found = 1;
                break;

            case 1:              /* Mnemonic     */
                if (prgmode == PRGDATA) {
                    error(NALWINDS, ptrsav);
                    break;
                }
                bufflg = 1;
                skipped = adrfield();
                len = oplen;
                if (skipped == 0) {
                    aflag = 1;              /* print address */
                    dflag = ASMPRT;         /* print data    */
                }
                found = 1;
                break;

            case 2:              /* DIRECTIVE     */
                if (optype & DIR_CONDITIONAL) {
//                    printf("1.'%s' (%d)   %d %d\n", txt, condskip, condlevel, skiplevel);
                    found = conddirective(opcode, ptrsav);
//                    printf("2.'%s' (%d)   %d %d\n", txt, condskip, condlevel, skiplevel);
                    if (condskip == 0 || found == 4) {
                        ret = 0;
                    }
                    else {
                        ret = 1;
                    }
                }
                else {
                    if (prgmode == PRGDATA && (optype & DIR_GENERATE)) {
                        error(NALWINDS, ptrsav);
                        break;
                    }
                    found = asmdirective(opcode, ptrsav);
                }
                break;
            }
        }

        if (found == 0) {              /* mnemonic/ directive not found */
            txtptr = ptrsav;
            macrodef = checkMacroDef();
            if (macrodef == 0) {
                found = checkStructDef();
                if (found == 0) {
                    if (txt[txtptr] && !iscomchar(txt[txtptr])) {
                        error(MNEMUNREC, txtptr);
                    }
                }
            }
        }
    }
    else {
        ret = 1;
        if (mnetyp == 2) {
            if (optype & DIR_CONDITIONAL) {
//                printf("3.'%s' (%d)   %d %d\n", txt, condskip, condlevel, skiplevel);
                found = conddirective(opcode, ptrsav);
//                printf("4.'%s' (%d)   %d %d\n", txt, condskip, condlevel, skiplevel);
                if (condskip == 0 || found == 4) {
                    ret = 0;
                }
                else {
                    ret = 1;
                }
            }
        }
    }

    return ret;
}

long readexpression()
{
    int ptrsav;
    long val;

    ptrsav = txtptr;
    exprflag = 0x80;
    val = expr();
    dta[0] = val >> 8;
    dta[1] = val;
    len = 2;
    dflag = FDBPRT;

    if (pass != LAST_PASS) {
        adrsiztab[adrsizcnt++] = exprflag;
    }
    else {
        exprflag = adrsiztab[adrsizcnt++];
    }

    if ((ptrsav == txtptr)) {
        error(NOEXPR, ptrsav);               /* symbols not defined */
    }
/* if there was an error on this line in pass 1 and no error is
 * found in pass 2, print out the saved error
 */
    if ((exprflag & 2) && errline != line) {
        error(SYMBMISS, ptrsav);               /* symbols not defined */
    }

    return val;
}
/*
 *   
 *   IF 
 *     IF
 *       mne
 *     ELSE
 *       mne
 *     ENDC
 *     mne
 *
 *   ELSE
 *     IF
 *       mne
 *     ELSE
 *       mne
 *     ENDC
 *     mne
 *   ENDC
 *   
 *
 */
int conddirective(int opcode, int inptrsav)
{
    int ptrsav;
    long val;
    char str1[100];
    char str2[100];
    int ret = 1;

    if (condskip == 0 || condlevel < skiplevel) {
        switch (opcode) {
        case AD_IF:
            if (noifelse) {
                return 0;
            }
            // no break
        case AD_IFNE:
            condlevel++;
            val = readexpression();

            if (pass == FIRST_PASS) {
                if (val) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFEQ:
            condlevel++;
            val = readexpression();
            if (pass == FIRST_PASS) {
                if (val == 0) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFGT:
            condlevel++;
            val = readexpression();
            if (pass == FIRST_PASS) {
                if (val > 0) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFGE:
            condlevel++;
            val = readexpression();
            if (pass == FIRST_PASS) {
                if (val >= 0) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFLT:
            condlevel++;
            val = readexpression();
            if (pass == FIRST_PASS) {
                if (val < 0) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFLE:
            condlevel++;
            val = readexpression();
            if (pass == FIRST_PASS) {
                if (val <= 0) {
                    condskip = 0;
                }
                else {
                    condskip = 1;
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_IFC:     //  IFC string1,string2
        case AD_IFNC:    //  IFNC string1,string2
            condlevel++;
            copystring(str1, 100);
            str2[0] = 0;
            if (txt[txtptr] == ',') {
                txtptr++;
                copystring(str2, 100);
            }

            if (!strcmp(str1, str2)) {
                condskip = 0;
            }
            else {
                condskip = 1;
            }
            if (opcode == AD_IFNC) {
                condskip = !condskip;
            }

            dta[0] = condskip >> 8;
            dta[1] = condskip;
            len = 2;
            dflag = FDBPRT;

            if (pass == FIRST_PASS) {
                if (condskip) {
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            else {
                condskip = 0;
            }
            break;

        case AD_ELSE:
            if (noifelse) {
                return 0;
            }
            // falling through
        case AD_ELSEC:
            if (pass == FIRST_PASS) {
                if (condlevel == 0) {
                    error(ENDCWITHNOIF, inptrsav);
                }
                else {
                    condskip = !condskip;
                    if (condskip) {
                        skiplevel = condlevel;
                        ret = 4;
                    }
                }
            }
            break;

        case AD_ENDC:
            if (pass == FIRST_PASS) {
                condlevel--;
                if (condlevel < 0) {
                    error(ENDCWITHNOIF, inptrsav);
                    condlevel = 0;
                }
            }
            break;

        default:
            error(MNEMUNREC, inptrsav);
            return 0;
            break;
        }
    }
    else {
        if (pass != FIRST_PASS) {
            return 1;
        }

        switch (opcode) {
        case AD_IF:
            if (noifelse) {
                return 0;
            }
            // no break
        case AD_IFNE:
        case AD_IFEQ:
        case AD_IFGT:
        case AD_IFGE:
        case AD_IFLT:
        case AD_IFLE:
        case AD_IFC:     //  IFC string1,string2
        case AD_IFNC:    //  IFNC string1,string2
            condlevel++;
            break;

        case AD_ELSE:
            if (noifelse) {
                return 0;
            }
            // falling through
        case AD_ELSEC:
            if (condlevel == skiplevel) {
                condskip = !condskip;
                if (condskip) {
                    skiplevel = condlevel;
                    ret = 4;
                }
            }
            break;

        case AD_ENDC:
            if (condlevel == skiplevel) {
                condskip = 0;
            }
            condlevel--;
            if (condlevel < 0) {
                error(ENDCWITHNOIF, inptrsav);
                condlevel = 0;
            }
            break;

        default:
            error(MNEMUNREC, inptrsav);
            return 0;
            break;
        }
    }

    if (nocondlist) {
        if (txt[0] == '+' || txt[0] == '-') {
            txt[0] = '-';
        }
        else {
            insertChar(txt, 1);
            txt[0] = '-';
            txtptr++;
        }
    }

    return ret;
}

int assigndirective(int opcode, int inptrsav)
{
    int ptrsav;
    long val;

    switch (opcode) {
    case AD_EQU:
        val = readexpression();
        defineAndSetVar(0, SYMC_VAR, val, DEF_NORMAL);
        break;

    case AD_SET:
        definevar(SYMC_VAR, DEF_BYSET);
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        dta[0] = val >> 8;
        dta[1] = val;
        len = 2;
        dflag = FDBPRT;

        if (ptrsav != txtptr) {
            if (lblflag) {
                setvar(val, DEF_BYSET);
            }
        }
        else {
            error(NOEXPR, ptrsav);
        }
        break;

    case AD_STRUCT:
        if (structdef == STRUCT_OPEN) { // error
            error(MACROINMACRO, ptrsav);
        }
        else {
            structdef = STRUCT_OPEN;
            structprgcnt = 0;
            dirstruct();
        }
        break;

    case AD_MACRO:  // define macro
        readMacroDef();
        break;

    default:
        if (!assigndirective_cpu(opcode, inptrsav)) {
            error(MNEMUNREC, txtptr);
            return 0;
        }
        break;
    }

    return 1;
}

int asmdirective(int dir, int ptrsav)
{

    switch (opcode) {
    case AD_FCB:
        bufflg = 1;
        dirfcb();
        aflag = 1;           /* print address */
        dflag = FCBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_FCC:
        bufflg = 1;
        dirfcc();
        aflag = 1;           /* print address */
        dflag = FCBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_FDB:
        bufflg = 1;
        dirfdb();
        aflag = 1;           /* print address */
        dflag = FDBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_FDBR:
        bufflg = 1;
        dirfdbr();
        aflag = 1;           /* print address */
        dflag = FDBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_FTB:
        bufflg = 1;
        dirftb();
        aflag = 1;           /* print address */
        dflag = FTBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_FQB:
        bufflg = 1;
        dirfqb();
        aflag = 1;           /* print address */
        dflag = FQBPRT;      /* print data    */
        len = oplen;
        break;

    case AD_ORG:
        dirorg();
        aflag = 1;           /* print address  */
        dflag = 0;           /* not print data */
        break;

    case AD_END:
        dirend();
        break;

    case AD_RMB:
        dirrmb();
        aflag = 1;              /* print address */
        dflag = FDBPRT;         /* print data    */
        len = 2;
        break;

    case AD_ZMB:
        bufflg = 1;
        len = dirzmb();
        aflag = 1;              /* print address */
        break;

    case AD_TTL:
        dirttl();
        break;

    case AD_OPT:
        diropt();
        break;

    case AD_DATA:                 /* DATA */
        cursection = datasection;
        prgmode = PRGDATA;
        break;

    case AD_TEXT:                 /* TEXT */
        cursection = textsection;
        prgmode = PRGTEXT;
        break;

    case AD_INCL:                 /* INCL(UDE) */
        dirincl();
        break;

    case AD_ALIG:                 /* ALIGN */
        diralign();
        break;

    case AD_PAGE:
        dirpage();
        break;

    case AD_CPU:
        dircpu();
        break;

    case AD_FAIL:
        dir_fail(CUSTOMEERROR);
        break;

    case AD_ERROR:
        dir_error(CUSTOMEERROR);
        break;

    case AD_WARN:
        dir_error(CUSTOMEWARNING);
        break;

    case AD_INFO:
        dir_error(CUSTOMEINFO);
        break;

    case AD_ENDSTRUCT:
        if (structdef == STRUCT_CLOSED) { // error
        }
        else {
            aflag = 1;              /* print address */
            direndstruct();
            structdef = STRUCT_CLOSED;
        }
        break;

    case AD_EXIT:   // exit macro
    case AD_ENDM:   // end of macro

        break;

    case AD_XDEF:       // ignore
        break;

    case AD_XREF:       // do a weak definition of symbols
        dir_xref();
        break;

    case AD_NAM:       // ignore
        break;

    case AD_POP:       //
        dirpop(ptrsav);
        break;

    case AD_GENLAB:    //
        dirgenlab();
        break;

    case AD_EXCH:      //
        direxch(ptrsav);
        break;

    default:
        if (!directive(opcode)) {
            error(MNEMUNREC, txtptr);
            return 0;
        }
        break;
    }

    return 1;
}

void dircpu()
{
    char name[50];
    char ch, enclosure;
    int i;
    int ptrsav;
    int found;

    if (txt[txtptr] == 0) {
        error(NOEXPR, txtptr - 1);
    }
    else {
        ptrsav = txtptr;
        if (txt[txtptr] == '"' || txt[txtptr] == '\'') {
            enclosure = txt[txtptr++];
        }
        else {
            enclosure = 0;
        }
        for (i = 0; i < 49; i++) {
            ch = txt[txtptr];
            if (ch == ' ' || ch == 0) {
                break;
            }
            else if (ch == enclosure) {
                break;
            }
            name[i] = toupper(ch);
            txtptr++;
        }
        name[i] = 0;
        if (enclosure != 0) {
            if (txt[txtptr] == enclosure) {
                txtptr++;
            }
            else {
                error(STRENDMIS, ptrsav);
            }
        }

        found = setcputype(name);
        if (found == 0) {
            error(VALOTORNG, ptrsav);
        }
    }
}

int setcputype(char *name)
{
    int i;
    int found;

    found = 0;
    for (i = 0; cputab[i].name[0]; i++) {
        if (!strcmp(cputab[i].name, name)) {
            found = 1;
            break;
        }
    }

    if (cputypeflag == 0) {
        if (found) {
            cputypemask = cputab[i].mask;
            cputypeflag = 1;
        }
    }

    return found;
}

void diralign()
{
    int val;
    int filler;
    int ptrsav;
    int c;
    int newpc;

    oplen = 0;
    ptrsav = txtptr;
    exprflag = 0;
    val = expr();               /* get align value */
    if (ptrsav == txtptr) {
        val = 1;
    }
    if (txt[txtptr] == ',') {
        txtptr++;
        ptrsav = txtptr;
        exprflag = 0;
        filler = expr();
        if (ptrsav == txtptr) {
            error(NOEXPR, ptrsav);
            filler = 0xff;
        }
        else if (filler > 255) {
            error(BIGGERVAL, ptrsav);
        }
    }
    else {
        filler = 0xff;
    }
    if (val < 1) {
        val = 1;
    }
    newpc = (prgcnt + val - 1) / val;
    newpc *= val;
    if (prgmode == PRGTEXT) {
        while (prgcnt < newpc) {
            dta[oplen++] = filler;
            prgcnt++;
        }
    }
    else {
        prgcnt = newpc;
    }
}

void dirpage()
{

    txt[0] = 0;     // just leave an empty source line

}

/*
 *
 *
 * FCB  data[,data]*
 *
 */
void dirfcb()
{
    long val;
    int ptrsav;
    int c;

    oplen = 0;
    do {
        while (txt[txtptr] == ' ') {           // accept spaces after comma
            txtptr++;
        }
        ptrsav = txtptr;
        if (txt[txtptr] == '"') {
            txtptr++;
            while (1) {
                if (txt[txtptr] == 0) {
                    error(STRENDMIS, ptrsav);
                    break;
                }
                if (txt[txtptr] == '"') {
                    txtptr++;
                    if (txt[txtptr] == '"') {
                        dta[oplen++] = txt[txtptr++];
                    }
                    else {
                        break;
                    }
                }
//                else if (txt[txtptr] == '\\') {
//                    txtptr++;
//                    if (txt[txtptr]) {
//                        dta[oplen++] = txt[txtptr++];
//                    }
//                }
                else {
                    dta[oplen++] = txt[txtptr++];
                }
            }
        }
        else {
            ptrsav = txtptr;
            exprflag = 0;
            val = expr();
            if (ptrsav != txtptr) {
                dta[oplen++] = val;
                if (val < -128 || val > 255) {
                    error(BIGGERVAL, ptrsav);
                }
            }
            else {
                error(NOEXPR, ptrsav);
                break;
                
            }
        }
    } while (txt[txtptr++] == ',');
    prgcnt += oplen;
}

/*
 *
 *
 * FCC  data[,data]*
 * FCC  length,text
 *
 */
void dirfcc()
{
    long val;
    int ch;
    int enclosure;
    int ptrsav;

    oplen = 0;
    if (isdigit(txt[txtptr])) {     // add the form length,text
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        if (ptrsav != txtptr) {
            if (val > 200) {
                val = 200;
            }
            if (txt[txtptr] == ',') {
                txtptr++;

                while (val > 0) {
                    ch = txt[txtptr];
                    if (ch != 0) {
                        txtptr++;
                    }
                    else {
                        ch = ' ';   // append with spaces when the line is too short
                    }
                    dta[oplen++] = ch;
                    val--;
                }
            }
        }
    }
    else {
        do {
            while (txt[txtptr] == ' ') {           // accept spaces after comma
                txtptr++;
            }
            ptrsav = txtptr;
            if (isin(txt[txtptr], "\"'/`\\")) {
                ptrsav = txtptr;
                enclosure = txt[txtptr++];
                while (1) {
                    if (txt[txtptr] == 0) {
                        error(STRENDMIS, ptrsav);
                        break;
                    }
                    ch = txt[txtptr++];
                    if (ch == enclosure) {
                        if (txt[txtptr] != enclosure) {
                            break;
                        }
                        txtptr++;
                    }
                    dta[oplen++] = ch;
                }
            }
            else {
                ptrsav = txtptr;
                exprflag = 0;
                val = expr();
                if (ptrsav != txtptr) {
                    dta[oplen++] = val;
                    if (val < -128 || val > 255) {
                        error(BIGGERVAL, ptrsav);
                    }
                }
                else {
                    error(NOEXPR, ptrsav);
                    break;
                    
                }
            }
        } while (txt[txtptr++] == ',');
    }
    
    prgcnt += oplen;

}

/*
 *
 *
 * FDB  data[,data]*
 *
 */
void dirfdb()
{
    long val;
    short ptrsav;

    oplen = 0;
    do {
        while (txt[txtptr] == ' ') {           // accept spaces after comma
            txtptr++;
        }
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        if (ptrsav != txtptr) {
            dta[oplen++] = val >> 8;
            dta[oplen++] = val;
        }
        else {
            error(NOEXPR, ptrsav);
            break;
        }

    } while (txt[txtptr++] == ',');
    prgcnt += oplen;

}

/*
 *
 *
 * .RWORD  data[,data]*
 *
 */
void dirfdbr()
{
    long val;
    short ptrsav;

    oplen = 0;
    do {
        while (txt[txtptr] == ' ') {           // accept spaces after comma
            txtptr++;
        }
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        if (ptrsav != txtptr) {
            dta[oplen++] = val;
            dta[oplen++] = val >> 8;
        }
        else {
            error(NOEXPR, ptrsav);
            break;
        }

    } while (txt[txtptr++] == ',');
    prgcnt += oplen;

}

/*
 *
 *
 * FTB  data[,data]*
 *
 */
void dirftb()
{
    long val;
    short ptrsav;

    oplen = 0;
    do {
        while (txt[txtptr] == ' ') {           // accept spaces after comma
            txtptr++;
        }
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        if (ptrsav != txtptr) {
            dta[oplen++] = val >> 16;
            dta[oplen++] = val >> 8;
            dta[oplen++] = val;
        }
        else {
            error(NOEXPR, ptrsav);
            break;
        }

    } while (txt[txtptr++] == ',');
    prgcnt += oplen;

}

/*
 *
 *
 * FQB  data[,data]*
 *
 */
void dirfqb()
{
    long val;
    short ptrsav;

    oplen = 0;
    do {
        while (txt[txtptr] == ' ') {           // accept spaces after comma
            txtptr++;
        }
        ptrsav = txtptr;
        exprflag = 0;
        val = expr();
        if (ptrsav != txtptr) {
            dta[oplen++] = val >> 24;
            dta[oplen++] = val >> 16;
            dta[oplen++] = val >> 8;
            dta[oplen++] = val;
        }
        else {
            error(NOEXPR, ptrsav);
            break;
        }

    } while (txt[txtptr++] == ',');
    prgcnt += oplen;

}

void dirorg()
{
    long val;
    short ptrsav;

    ptrsav = txtptr;
    exprflag = 0x80;
    val = expr();

    if (pass != LAST_PASS) {
        adrsiztab[adrsizcnt++] = exprflag;
    }
    else {
        exprflag = adrsiztab[adrsizcnt++];
    }

    if (ptrsav == txtptr) {
        error(NOEXPR, ptrsav);
      
    }
/* if there was an error on this line in pass 1 and no error is
 * found in pass 2, print out the saved error
 */
    else if ((exprflag & 2) && errline != line) {
        error(SYMBMISS, ptrsav);               /* symbols not defined */
    }
    else {
        if (bufflg) {
            prntbuff(1);
            bufflg = 0;
        }
        prgcnt = val & 0xFFFF;
        startadr = adr = prgcnt;
    }
}

void dirend()
{

/* ehhh??? */
    runadr = expr();
    if (bufflg) {
        prntbuff(1);
        bufflg = 0;
    }

    startadr = runadr;
    prntbuff(2);

}

void dirttl()
{
    short i;

    for(i = 0 ; txt[txtptr] && i < 45 ; title[i++] = txt[txtptr++])
        ;

    title[i] = 0;
}

/*
 *
 *
 * OPT  option[,option]*
 *
 *
 * Some of the options are are defined by Motorola assembler and
 * are silently 'eaten up'
 *
 */
void diropt()
{
    long val;
    short i;
    short ptrsav;
    char r[16];

    while (1) {
        ptrsav = txtptr;

        for (i = 0; txt[txtptr+i] != ',' && txt[txtptr+i] != ' '; i++) {
            r[i] = toupper(txt[txtptr+i]);
        }
        r[i] = 0;
        if (!strcmp(r, "ABS")) { /* absolute allocated code */
            txtptr += 3;
        }
        else if (!strncmp(r, "LLE=", 4)) { /* line length*/
            txtptr += 4;
            val = factor2();
            linelen = val;
        }
        else if (!strncmp(r, "LLEN=", 5)) { /* line length*/
            txtptr += 5;
            val = factor2();
            linelen = val;
        }
        else if (!strncmp(r, "PLE=", 4)) { /* paper length */
            txtptr += 4;
            val = factor2();
            pagelen = val;
        }
        else if (!strcmp(r, "CRE")) { /* cross ref listing */
            txtptr += 3;
            xreflist = 1;
        }
        else if (!strcmp(r, "X")) { /* cross ref listing */
            txtptr += 1;
            xreflist = 1;
        }
        else if (!strcmp(r, "NOG")) { /*  */
            txtptr += 3;
        }
        else if (!strcmp(r, "LST")) { /* list file */
            txtptr += 3;
            listflag  = 2;
        }
        else if (!strcmp(r, "-LST")) { /* no list file */
            txtptr += 4;
            listflag = 0;
        }
        else if (!strcmp(r, "L")) { /* list file */
            txtptr += 1;
        }
        else if (!strcmp(r, "NOL")) { /* no list file */
            txtptr += 3;
        }
        else if (!strcmp(r, "ELBW")) { /* Enable long branch wrn */
            txtptr += 4;
            lbwrn = 1;
        }
        else if (!strncmp(r, "BRAOPT", 6)) { /* optimize long/short branch */
            txtptr += 6;
            if (r[6] == 'Z') {
                txtptr++;
                optimizebranch = 0x100;
            }
            else {
                optimizebranch = 0;
            }

            if (txt[txtptr] == '=') {
                txtptr++;
                val = factor2();
                optimizebranch |= val & 255;
            }
            else {
                optimizebranch |= 1;
            }
        }
        else if (!strcmp(r, "NOFORWOPT")) { /* No forward optimization */
            txtptr += 8;
            noforwardopt = 1;
        }
        else if (!strcmp(r, "NOPRIO")) { /* No priority of mul etc. */
            txtptr += 6;
            mulnoprio = 1;
        }
        else if (!strcmp(r, "NOIF")) { /* Don't interpret IF and ELSE as directives */
            txtptr += 4;
            noifelse = 1;
        }
        else if (!strcmp(r, "S")) { /* Symbol listing */
            txtptr += 1;
        }
        else if (diropt2(r)) {
        }
        else {
            txtptr += i;
            error(UNRECOPT, ptrsav); /* UNRECOPT */
        }

        if (txt[txtptr] != ',') {
            break;
        }
        txtptr++;
    }
}

int copystring(char *d, int len)
{
    int enclosure = 0;
    int i;
    int ch;
    short ptrsav;

    len--;

    if (isin(txt[txtptr], "\"'")) {
        enclosure = txt[txtptr++];
        ptrsav = txtptr;
        i = 0;
        while (txt[txtptr] && i < len) {
            ch = txt[txtptr++];
            if (ch == enclosure) {
                if (txt[txtptr] != enclosure) {
                    break;
                }
                txtptr++;
            }
            *d++ = ch;
            i++;
        }
        *d = 0;
    }
    else {
        i = 0;
        while (txt[txtptr] && i < len) {
            ch = txt[txtptr];
            if (isin(ch, " ,")) {
                break;
            }
            txtptr++;
            *d++ = ch;
            i++;
        }
        *d = 0;
    }

    return i;
}

void dirincl()        /*    INCL    "filename.asm"   */
{
    char tt[200];
    char name[128];
    char *p;
    int ch;
    int enclosure = 0;
    int i;
    short ptrsav;
    FILE *fh;

    printf("%s\n", txt);

    if (pass != FIRST_PASS) {
        return;
    }

    if (txt[txtptr] == 0) {
        error(NOEXPR, txtptr);
    }
    else {
        copystring(name, 50);
        if (name[0] == '/') {
            p = name;
        }
        else {
            strcpy(tt, files[actfile].filesname);
            p = strrchr(tt, '/');
            if (p == 0) {
                p = name;
            }
            else {
                strcpy(p+1, name);
                p = tt;
            }
        }

        if (actfile < (MAXOPENFILE - 1)) {
            if (!(fh = open_inclfile(p))) {
                error(CANTOPNFIL, ptrsav);
                return;
            }
            files[actfile].fileshdl = hdl_asm;
            files[actfile].filesline = line;
            hdl_asm = fh;

            actfile++;
            strcpy(files[actfile].filesname, p);
            actfname = files[actfile].filesname;
            line = 0;
        }
        else {
            error(TOOMANYFIL, ptrsav);
        }
    }

}

void dirrmb()
{
    long val;
    short ptrsav;

    oplen = 0;
    ptrsav = txtptr;
    exprflag = 0x80;
    val = expr();

    if (pass != LAST_PASS) {
        adrsiztab[adrsizcnt++] = exprflag;
    }
    else {
        exprflag = adrsiztab[adrsizcnt++];
    }

    if (ptrsav == txtptr) {
        error(NOEXPR, ptrsav);
    }
/* if there was an error on this line in pass 1 and no error is
 * found in pass 2, print out the saved error
 */
    else if ((exprflag & 2) && errline != line) {
        error(SYMBMISS, ptrsav);               /* symbols not defined */
    }
    else {
        setvarsize(0, val);
        addVarToStruct(0);
        dta[0] = val >> 8;
        dta[1] = val;
        prgcnt += val;
        if (prgmode == PRGTEXT) {
            if (bufflg) {
                prntbuff(1);
                bufflg = 0;
            }
            startadr = prgcnt;
        }
    }
}

/*
 *
 *
 * ZMB  count[,filler]
 *
 */
int dirzmb()
{
    long val;
    short ptrsav;
    int i;
    int filler = 0;
    int rv = 2;

    dflag = PDTAPRT;         /* print p-data    */

    oplen = 0;
    ptrsav = txtptr;
    exprflag = 0x80;
    val = expr();

    if (pass != LAST_PASS) {
        adrsiztab[adrsizcnt++] = exprflag;
    }
    else {
        exprflag = adrsiztab[adrsizcnt++];
    }

    if (ptrsav == txtptr) {
        error(NOEXPR, ptrsav);
    }
/* if there was an error on this line in pass 1 and no error is
 * found in pass 2, print out the saved error
 */
    else if ((exprflag & 2) && errline != line) {
        error(SYMBMISS, ptrsav);               /* symbols not defined */
    }
    else {
        pdta[0] = val >> 8;
        pdta[1] = val;

        if (txt[txtptr] == ',') {
            txtptr++;

            ptrsav = txtptr;
            filler = expr();
            if (ptrsav == txtptr) {
                error(NOEXPR, ptrsav);
            }
            pdta[2] = 0;
            pdta[3] = filler;
            rv = 4;
        }

        for (i = 0; i < val; i++) {
            dta[i] = filler;
        }

        oplen = val;
        prgcnt += val;

        if (val <= 12) {
            dflag = FCBPRT;      /* print data    */
            rv = val;
        }
    }

    return rv;
}


/*
 *
 *
 * .ERROR  [condition,]string
 * .ERROR  string[,condition]
 *
 */
void dir_error(int errcode)
{
    char ch;
    char enclosure;
    int i = 0;
    short ptrsav;
    long val;
    int flag = -1;

    errorstring[0] = 0;

    ptrsav = txtptr;
    exprflag = 0;
    val = expr();
    if (ptrsav != txtptr) {
        if (val) {
            flag = 1;
        }
        else {
            flag = 0;
        }
        if (txt[txtptr] == ',') {
            txtptr++;
        }
    }

    if (txt[txtptr] == '"') {
        txtptr++;
        ptrsav = txtptr;
        while (1) {
            if (txt[txtptr] == 0) {
                error(STRENDMIS, ptrsav);
                break;
            }
            if (txt[txtptr] == '"') {
                txtptr++;
                if (txt[txtptr] == '"') {
                    errorstring[i++] = txt[txtptr++];
                }
                else {
                    break;
                }
            }
            else {
                errorstring[i++] = txt[txtptr++];
            }
        }
        errorstring[i] = 0;
    }

    if (flag < 0) {
        if (txt[txtptr] == ',') {
            txtptr++;
            ptrsav = txtptr;
            exprflag = 0;
            val = expr();
            if (ptrsav != txtptr) {
                if (val) {
                    flag = 1;
                }
                else {
                    flag = 0;
                }
            }
        }
    }

//    printf("%d -- %d -- '%s'\n", val, flag, errorstring);
    if (errorstring[0] == 0) {
        error(NOEXPR, -1);
    }
    else if (flag) {
        error(errcode, -1);
    }
}

/*
 *
 * Take the rest of the line as error text
 * 
 * FAIL     text
 *
 */
void dir_fail(int errcode)
{
    char ch;
    char enclosure;
    int i = 0;
    short ptrsav;
    long val;
    int flag = -1;

    errorstring[0] = 0;

    ptrsav = txtptr;

    ptrsav = txtptr;
    while (txt[txtptr]) {
        errorstring[i++] = txt[txtptr++];
    }
    errorstring[i] = 0;

//    printf("%d -- %d -- '%s'\n", val, flag, errorstring);
    if (errorstring[0] == 0) {
        error(NOEXPR, -1);
    }
    else if (flag) {
        error(errcode, -1);
    }
}

void dir_xref()
{
    long val;
    short i;
    short ptrsav;
    char r[16];

    while (1) {
        ptrsav = txtptr;
        copyvar();
        definevar(SYMC_VAR, DEF_WEAK);

        if (txt[txtptr] != ',') {
            break;
        }
        txtptr++;
    }

}

/*
 *
 * check for mnemonic of type: LDA A ...
 * and change them to:         LDAA  ...
 *
 */
void mnefixcheck(char *tmp, int *pp)
{
    MneFix *mfp;
    int p;
    int found = 0;
    int i;
    int ch;

    for (mfp = mnefixtbl; mfp->mnem[0]; mfp++) {
        if (!strcmp(mfp->mnem, tmp)) {
            found = 1;
            break;
        }
    }

    if (found == 0) {   // no match
        return;
    }

    i = strlen(tmp);
    p = *pp;

    if (txt[p] == ' ' && txt[p+2] == ' ' ) {
        ch = toupper(txt[p+1]);
        if (isin(ch, mfp->reg)) {
            tmp[i-1] = ch;
            tmp[i++] = ' ';
            tmp[i] = 0;
            p += 2;
            *pp = p;
        }
    }
}

/*
 *
 *
 *
 * int *code:     Pointer to opcode variable
 * int *type:     Pointer to opcode variable
 * int *ref:      Pointer to reference variable
 * int cputype:   Flag for allowed cpu
 * uint modeflag: Flag for assembler mode   0=normal mode
 *
 * return: 0=nofind, 1=mnemonic, 2=directiv, 3=EQU, -1=empty line
 */
int mnefind(int *code, int *type, int *ref, int cputype, uint modeflag)
{
    MneType *mneptr;
    int i;
    char tmp[16];
    int p;
    int found;

    if (txt[txtptr] == 0) {
        return -1;
    }

/*   strncpy(tmp, txt + txtptr, 4); */
    i = 0;
    p = txtptr;
    while (txt[p] != 0 && txt[p] != ' ' && i < 10) {
        tmp[i++] = toupper(txt[p++]);
    }
    tmp[i++] = ' ';
    tmp[i] = 0;

    mnefixcheck(tmp, &p);

//  mneptr = mnemic;
//  printf("3.1  '%s'\n", tmp);
//  printf("3.2  '%s'\n", mneptr->mnem);
//  printf("3.3  %08X   %08X\n", mneptr, txt);

    found = 0;
    for (mneptr = mnemic; mneptr->mnem[0] != '*'; mneptr++) {
        if (!strncmp(tmp, mneptr->mnem, strlen(mneptr->mnem))) {
            if ((mneptr->allowed & ALLOWCPUMASK) == 0 || (mneptr->allowed & cputype) != 0) {
                found = 1;
                break;
            }
        }
    }

    if (found == 0) {
        return 0;
    }
// printf("1.1  '%s'\n", mneptr->mnem);
// printf("1.2  %08X   %08X  \n", mneptr, txt);

// statistics
    if (pass == FIRST_PASS) {
        i = mneptr - mnemic;
        mnemiccnttab[i]++;
        mnemiccnt++;
    }

    txtptr = p;
    if ((mneptr->type & TYPEMASK) == ASMDIR) {
        if (pass == FIRST_PASS) {
            mnedircnt++;
        }
        *code = mneptr->opcode;
        *type = mneptr->type & DIR_FLAGMASK;
        if ((mneptr->opcode & 0x80) || (mneptr->type & DIR_ASSIGN)) {
            return 3;
        }
        else {
            return 2;
        }
    }
    else {
        *code = mneptr->opcode;
        *type = mneptr->type & (TYPEMASK | PAGEMASK);
        *ref = (mneptr->type >> MNEACCSHIFT);
        return 1;
    }

    return 0;
}


/*
 *  param:
 *   type:   REL or REL16
 *   ofs:    branch offset
 *   prgcnt: address of branch instruction
 *   txtpos: text position of address parameter (used only in case of error)
 */
void branchstat(int type, int ofs, int prgcnt, int txtpos)
{
    int neardist = 10;

    if (type == REL) {
    }
    else if (type == REL16) {
        if (lbwrn) {
            if (ofs >= -(128+neardist) && ofs < -128) { /* check for 10 out of range */
                error(LBRLOWOFFS, txtpos);
            }
            else if (ofs >= 128 && ofs < (128+neardist)) { /* check for 10 out of range */
                error(LBRLOWOFFS, txtpos);
            }
        }
    }

}

void error(int errcode, short tp)
{
    char *p;
    char tt[32];

    if (pass == LAST_PASS) {
        if (errcode < 0) {
            if (errcode == CUSTOMEERROR) {
                errcnt++;
                if (errlinelow == -1) {
                    errlinelow = line;      
                }

                if (line > errline) {
                    errline = line;
                }
            }
            else if (errcode == CUSTOMEWARNING) {
                wrncnt++;
                if (wrnlinelow == -1) {
                    wrnlinelow = line;      
                }

                if (line > wrnline) {
                    wrnline = line;      
                }
            }
        }
        else {
            p = errstr[errcode & 0x7F];
            if (*p == 'S') {
                p++;
            }
            if (*p == 'E') {
                errcnt++;
                if (errlinelow == -1) {
                    errlinelow = line;      
                }

                if (line > errline) {
                    errline = line;
                }
            }
            else if (*p == 'W') {
                wrncnt++;
                if (wrnlinelow == -1) {
                    wrnlinelow = line;      
                }

                if (line > wrnline) {
                    wrnline = line;      
                }
            }
            else if (*p == 'I') {
                infocnt++;
                if (infolinelow == -1) {
                    infolinelow = line;      
                }

                if (line > infoline) {
                    infoline = line;      
                }
            }
            else {
            }
        }

        if (errpos < 10) {
            errpostab[errpos] = tp;
            errcodtab[errpos++] = errcode;
        }
    }
    else {
        if (errcode < 0) {
            if (errcode == CUSTOMEERROR) {
                errcnt++;
            }
        }
        else {
            p = errstr[errcode & 0x7F];
            if (*p == 'S') {
                p++;
                sprintf(tt ,"(%d,%d)", errcode, tp);
//                strcat(errout, tt);
                printf("%s\n", errout);
            }
            if (*p == 'E') {
                errcnt++;
            }
//            printf("%d :  error=%d   %s\n", line, errcode, p);
        }
    }
}

void formerr(FILE* hdl)
{
    char *strptr;
    short i;
    short hipos;
    short pos;
    short errcode;
    char istr[8];
    char errchar;

    if (errpos) {
        for (i = 0 ; i < 128 ; errout[i++] = ' ')
            ;

        hipos = 0;

        for (i = 0 ; i < errpos ; i++) {
            pos = errpostab[i] + txtstart;
            if (errpostab[i] >= 0) {
                while (errout[pos] != ' ') {
                    pos++;
                }

                errout[pos] = i + '0';
                if (pos > hipos) {
                    hipos = pos;
                }
            }
        }

        errout[hipos + 1] = 0;
        fprintf(hdl, "%s\n", errout);
        for (i = 0; i < errpos; i++) {
            errcode = errcodtab[i];
            if (errcode < 0){
                if (errcode == CUSTOMEERROR) {
                    errchar = 'E';
                }
                else if (errcode == CUSTOMEWARNING) {
                    errchar = 'W';
                }
                else if (errcode == CUSTOMEINFO) {
                    errchar = 'I';
                }
                strptr = errorstring;
            }
            else {
                strptr = errstr[errcode & 0x7F];
                if (*strptr == 'S') {
                    strptr++;
                }
                errchar = *strptr++;
            }

            if (errpostab[i] >= 0) {
                sprintf(istr, "%d.", i);
            }
            else {
                strcpy(istr, "  ");
            }

            if (errchar == 'E') {
                fprintf(hdl, "%s:  %s   ---ERROR---    %s\n",
                        actfname, istr, strptr);
            }
            else if (errchar == 'W') {
                fprintf(hdl, "%s:  %s   ---WARNING---  %s\n",
                        actfname, istr, strptr);
            }
            else if (errchar == 'I') {
                fprintf(hdl, "%s:  %s   ---INFORMATION---  %s\n",
                        actfname, istr, strptr);
            }
        }
    }
}

void prntout()
{

    if  (pass == LAST_PASS) { // output files
        buffbin();

        if (hdl_lst) {
            prntlst();
        }
        if (hdl_err) {
            prnterr();
        }
    }
}

void prntlst()
{

/* print string to out file */

    if (txt[0] != '-') {
        formlst(hdl_lst);
    }
}

void prnterr()
{

/* Is there any errors */

    if (errpos) {

/* first print line to out file */

        formlst(hdl_err);

    }
}

void prntbin(int len)
{
    int i;

    for (i = 0; i < len ; i++) {
        fputc(buftab[i], hdl_bin);
    }
}


int hexch(int val)
{
    static char hextab[] = { "0123456789ABCDEF" };

    return hextab[val & 0x0f];
}

void prntihex(int type, int addr, int count, uchar buf[])
{
    int csum;
    int i;
    int typech;

	switch (type) {
	case 0:
        return;
		break;

	case 1:
		typech = 0;
		break;

	case 2:
        if (recoutstate > 1) {
            return;
        }
        recoutstate = 2;
        typech = 1;
		break;

	}

    fputc(':', hdl_bin);
    fputc(hexch(count>>4), hdl_bin);
	fputc(hexch(count),hdl_bin);
	fputc(hexch((int)addr>>12),hdl_bin);
	fputc(hexch((int)addr>>8),hdl_bin);
	fputc(hexch((int)addr>>4),hdl_bin);
	fputc(hexch((int)addr),hdl_bin);
    fputc(hexch(typech>>4), hdl_bin);
	fputc(hexch(typech),hdl_bin);

	csum = ((addr >> 8) & 0xff) + (addr & 0xff) + (count & 0xff);
	csum += typech & 0xff;

    for (i = 0; i < count; i++) {
        csum += buf[i] & 0xff;
        fputc(hexch(buf[i]>>4), hdl_bin);
        fputc(hexch(buf[i]), hdl_bin);
    }

	csum = (-csum) & 0xff;
	fputc(hexch(csum>>4), hdl_bin);
	fputc(hexch(csum), hdl_bin);
	fputc('\n',hdl_bin);

}

void prnts19(int type, int addr, int count, uchar buf[])
{
    int csum;
    int i;
    int typech;

	switch (type) {
	case 0:
		typech = '0';
		break;

	case 1:
		typech = '1';
		break;

	case 2:
        if (recoutstate > 1) {
            return;
        }
        recoutstate = 2;
        typech = '9';
		break;

	}

    fputc('S', hdl_bin);
    fputc(typech, hdl_bin);

    count += 3;
    fputc(hexch(count>>4), hdl_bin);
	fputc(hexch(count),hdl_bin);
	fputc(hexch((int)addr>>12),hdl_bin);
	fputc(hexch((int)addr>>8),hdl_bin);
	fputc(hexch((int)addr>>4),hdl_bin);
	fputc(hexch((int)addr),hdl_bin);

    csum = ((addr >> 8) & 0xff) + (addr & 0xff) + (count & 0xff);
    count -= 3;

    for (i = 0; i < count; i++) {
        csum += buf[i];
        fputc(hexch(buf[i]>>4), hdl_bin);
        fputc(hexch(buf[i]), hdl_bin);
    }

	csum = (~csum) & 0xff;
	fputc(hexch(csum>>4), hdl_bin);
	fputc(hexch(csum), hdl_bin);
	fputc('\n',hdl_bin);

}

/**
 * prntbuff
 *
 *
 * param int type: 0=initial record, 1=middle record, 2=end record
 *
 *
 */
void prntbuff(int type)
{
    int addr;
    int i;
    int cnt;

    if (pass == LAST_PASS) {
        switch (objfileformat) {
        case OWNOBJ:
            if (type < 2) {
                buftab[0] = startadr >> 8;
                buftab[1] = startadr;
                buftab[2] = (bufptr + startadr - 5) >> 8;
                buftab[3] = bufptr + startadr - 5;
                prntbin(bufptr);
                bufptr = 4;
            }
            break;
        case S19OBJ:
            if (type == 1) {
                addr = startadr - 4;
                i = 4;
                while (i < bufptr) {
                    cnt = bufptr - i;
                    if (cnt > recordlen) {
                        cnt = recordlen;
                    }
                    prnts19(type, addr+i, cnt, buftab+i);
                    i += cnt;
                }
            }
            else {
                addr = startadr;
                cnt = bufptr - 4;
                prnts19(type, addr, cnt, buftab+4);
            }
            bufptr = 4;
            break;

        case INTOBJ:
            if (type == 1) {
                addr = startadr - 4;
                i = 4;
                while (i < bufptr) {
                    cnt = bufptr - i;
                    if (cnt > recordlen) {
                        cnt = recordlen;
                    }
                    prntihex(type, addr+i, cnt, buftab+i);
                    i += cnt;
                }
            }
            else {
                addr = startadr;
                cnt = bufptr - 4;
                prntihex(type, addr, cnt, buftab+4);
            }
            bufptr = 4;
            break;
        case TECOBJ:
            bufptr = 4;
            break;
        }
    }
}

int maxoplen = 0;
int maxopclen = 0;
int *oplentab = 0;
int maxpages = 0;

/*
 *
 * param int oplen:  longest opcode including data
 * param int opclen: longest opcode, instruction part
 * param int pages:  number of opcode pages
 *
 */
void setmaxoplen(int oplen, int opclen, int pages)
{

    maxoplen = oplen + 1;
    maxopclen = opclen + 1;

    oplentab = malloc(maxoplen * maxopclen * sizeof(int));
    memset(oplentab, 0, maxoplen * maxopclen * sizeof(int));
    maxpages = pages;
}

void addoplenstat(int oplen, int opclen, int page, int opcode)
{
    int ind;

    if (pass != LAST_PASS) {
        return;
    }

    ind = opclen * maxoplen + oplen;
//    printf("ind=%d, maxind=%d\n", ind, maxoplen * maxopclen);

    oplentab[ind]++;

}

void prntoplenstatistics(FILE *fh)
{
    int sum = 0;
    int op;
    int opc;
    int ind;
    int cnt1;
    int cnt2;
    int csum1;
    float pcent1;
    float pcent2;
    float psum1;
    float psum2;

    for (opc = 1; opc < maxopclen; opc++) {
        ind = opc * maxoplen;
        for (op = 1; op < maxoplen; op++) {
            oplentab[ind] += oplentab[ind+op];
        }
        sum += oplentab[ind];
    }


    fprintf(fh, "\n");
    fprintf(fh, "Instruction length:\n");

    fprintf(fh, "     ");
    for (opc = 1; opc < maxopclen; opc++) {
        fprintf(fh, "%d               ", opc);
    }
    fprintf(fh, "<- opcode length\n");

    fprintf(fh, "   ");
    for (opc = 1; opc < maxopclen; opc++) {
        fprintf(fh, "  --------------");
    }
    fprintf(fh, "\n");

    fprintf(fh, "len");
    for (opc = 1; opc < maxopclen; opc++) {
        fprintf(fh, "  count  percent");
    }
    if (maxopclen > 2) {
        fprintf(fh, "  count  percent");
    }
    fprintf(fh, "\n");

    fprintf(fh, "---");
    for (opc = 1; opc < maxopclen; opc++) {
        fprintf(fh, "  -----  -------");
    }
    if (maxopclen > 2) {
        fprintf(fh, "  -----  -------");
    }
    fprintf(fh, "\n");


    for (op = 1; op < maxoplen; op++) {
        psum1 = 0.0;
        csum1 = 0;
        fprintf(fh, "%3d", op);
        for (opc = 1; opc < maxopclen; opc++) {
            cnt1 = oplentab[opc * maxoplen + op];
            csum1 += cnt1;
            pcent1 = ((float) cnt1 * 100) / sum;
            psum1 += pcent1;
            fprintf(fh, "  %5d  %5.1f  ", cnt1, pcent1);
        }

        if (maxopclen > 2) {
            fprintf(fh, "  %5d  %5.1f  ", csum1, psum1);
        }
        fprintf(fh, "\n");
    }

    fprintf(fh, "---");
    for (opc = 1; opc < maxopclen; opc++) {
        fprintf(fh, "  -----  -------");
    }
    if (maxopclen > 2) {
        fprintf(fh, "  -----  -------");
    }
    fprintf(fh, "\n");


    psum1 = 0.0;
    csum1 = 0;
    fprintf(fh, "tot");
    for (opc = 1; opc < maxopclen; opc++) {
        cnt1 = oplentab[opc * maxoplen];
        csum1 += cnt1;
        pcent1 = ((float) cnt1 * 100) / sum;
        psum1 += pcent1;
        fprintf(fh, "  %5d  %5.1f  ", cnt1, pcent1);
    }

    if (maxopclen > 2) {
        fprintf(fh, "  %5d  %5.1f  ", csum1, psum1);
    }
    fprintf(fh, "\n");

}


void prntstatistics(FILE *hdl)
{
    int i;
    int ttlen;
    int index;
    int tabsize;
    char tt[16];

    tabsize = getmnemicsize() / sizeof(MneType);

    fprintf(hdl, "\n Statistics:  (%d statements including %d directives)\n",
            mnemiccnt, mnedircnt);

    i = 0;
    for (index = 0; index < tabsize; index++) {
        if (mnemiccnttab[index]) {
            sprintf(tt, "  %-6s%5d",
                    mnemic[index].mnem,
                    mnemiccnttab[index]);

            ttlen = strlen(tt);

            if ((i + ttlen) > linelen) {
                fprintf(hdl, "%s\n", txtout);
                i = 0;
            }

            strcpy(txtout + i, tt);
            i += ttlen;
        }
    }

    if (i) {
        fprintf(hdl, "%s\n", txtout);
    }

    prntoplenstatistics(hdl);
}

void buffbin()
{
    int i;

    if (bufptr + oplen > MAXBUFLEN) {
        prntbuff(1);
    }

    for (i = 0; i < oplen; i++) {
        buftab[bufptr + i] = dta[i];
    }
    bufptr += oplen;
}

TXTLINE *maketxtline(int lineno, char *txt, char *fname)
{
    TXTLINE *tlp;

    tlp = malloc(sizeof(TXTLINE));

    memset(tlp, 0, sizeof(TXTLINE));

    tlp->lineno = lineno;
    tlp->txt = strdup(txt);
    tlp->filename = fname;

    return tlp;
}

/*
 * storetxt
 *
 * store a text, for example a source code line, inlude a line number
 * and a file name reference.
 *  space is allocated for the text, but NOT for the reference
 *
 */
void storetxt(TXTHEAD *headp, int lineno, char *txt, char *fname)
{
    TXTLINE *tlp;

//    printf("add: %04d %s\n", lineno, txt);

    tlp = maketxtline(lineno, txt, fname);
    if (errout[0]) {
        tlp->errtxt = strdup(errout);
    }

    if (headp->txtlh == 0) {
        headp->txtlh = tlp;
        headp->txtle = tlp;
    }
    else {
        headp->txtle->nxt = tlp;
        headp->txtle = tlp;
    }

}

/*
 *
 *
 */
void addStoredErrors(char *p)
{
    int errcode;
    int tp;

    while (*p) {
        if (*p == '(') {
            p++;
            errcode = strtol(p, &p, 10);
            p++;                    // step over ',' assumed to be there...
            tp = strtol(p, &p, 10);
            p++;                    // and the ')'
            error(errcode, tp);
        }
    }
}


/*
 *  returns:  0=not end of file, EOF=end of file, 2=include file closed
 */
int readasm()
{
    int i, c;

    if (pass != FIRST_PASS && srctxt.txtlh) {
        txtptr = 0;
        if (txtlp) {
            line = txtlp->lineno;
            actfname = txtlp->filename;
            strcpy(txt, txtlp->txt);

// check stored errors
            if (txtlp->errtxt) {
                addStoredErrors(txtlp->errtxt);
            }

            txtlp = txtlp->nxt;
            return 0;
        }
        else {
            return EOF;
        }
    }

    i = 0;
    while ((c = fgetc(hdl_asm)) != EOF && c != '\n') {
        if (i < (MAXTXTLEN-2)) {
            if (c == '\r' || c == 0x1a) {    /* ignore CR or MS end-of-file */
            }
            else if (c == '\t') {
                if (i < (MAXTXTLEN-2)) {
                    txt[i++] = ' ';
                }
                while (tabtxt[i] && i < (MAXTXTLEN-2)) {
                    txt[i++] = ' ';
                }
            }
            else {
                txt[i++] = c;
            }
        }
    }

    txt[i++] = ' ';   /* add a space at the end of the line */
    txt[i] = 0;

    txtptr = 0;
    line++;

    if (c == EOF && i == 1) {   // don't end until the last line is empty
        if (actfile == 0) {
            return EOF;
        }
        else {
//            files[actfile].filesname[0] = 0;
            actfile--;

            fclose(hdl_asm);

            hdl_asm = files[actfile].fileshdl;
            line = files[actfile].filesline;
            actfname = files[actfile].filesname;
            return 2;
        }
    }

    return 0;
}

void toterrprnt()
{
    char tte[200];
    char ttw[200];
    char tti[200];

    sprintf(tte,"  Total errors %d", errcnt);
    if (errline) {
        sprintf(tte+strlen(tte)," (First/Last on line %d/%d)",
                errlinelow, errline);
    }

    sprintf(ttw,"  Total warnings %d", wrncnt);
    if (wrnline) {
        sprintf(ttw+strlen(ttw)," (First/Last on line %d/%d)",
                wrnlinelow, wrnline);
    }

    tti[0] = 0;
    if (infocnt) {
        sprintf(tti,"  Total info msg %d", infocnt);
        if (infoline) {
            sprintf(tti+strlen(tti)," (First/Last on line %d/%d)",
                    infolinelow, infoline);
        }
    }

    if (hdl_lst) {
        fprintf(hdl_lst,"%s\n", tte);
        fprintf(hdl_lst,"%s\n", ttw);
        if (tti[0]) {
            fprintf(hdl_lst,"%s\n", tti);
        }
    }

    if (hdl_err) {
        fprintf(hdl_err,"%s\n", tte);
        fprintf(hdl_err,"%s\n", ttw);
        if (tti[0]) {
            fprintf(hdl_err,"%s\n", tti);
        }
    }
}

/*
 * insertChar
 * move text to make space to insert characters
 *
 *
 *
 * param char *txt: point for insertion
 * param int count: number of characters to insert
 *
 */
int insertChar(char *txt, int count)
{
    char *d;
    char *s;

    s = txt + strlen(txt);
    d = s + count;

    while (s >= txt) {
        *d-- = *s--;
    }

    return 0;
}

/*
 *  returns true if 'c' is a char found in string 's'.
 *
 */
int isin(int c, char *s)
{

    while (*s) {
        if (c == *s) {
            return 1;
        }
        s++;
    }
    return 0;
}

char *scanto(char *s, char *e)
{

    while (*s && !isin(*s, e)) {
        s++;
    }

    return s;
}

/*
 * Copy string until but not including delimiter
 *
 * @param char *d: destination
 * @param char *s: source
 * @param char *e: delimiter(s)
 */
char *strdcpy(char *d, char *s, char *e)
{

    while (*s && !isin(*s, e)) {
        *d++ = *s++;
    }
    *d = 0;
    return s;
}

/*
 * copy string until but not including delimiter
 *  or until n-1 chars are copied, rest of chars are skipped
 *
 *  char *d:  destination
 *  char *s:  source
 *  char *e:  delimiter(s)
 *  int n:    size of destination
 *
 * returns pointer to next char
 */
char *strndcpy(char *d, char *s, char *e, int n)
{

    n--;
    while (*s && !isin(*s, e)) {
        if (n > 0) {
            *d++ = *s;
            n--;
        }
        s++;
    }
    *d = 0;
    return s;
}

char *esp(char *p)
{

    while (*p == ' ') {
        p++;
    }
    return p;
}

int getregval(REGNAM *t, char *s)
{

    while (t->regname[0]) {
        if (!strcasecmp(t->regname, s)) {
            if (t->allowed == 0 || (t->allowed & cputypemask)) {
                return t->regvalue;
            }
        }
        ++t;
    }
    return -1;
}

int getreglistmask(REGNAM *table, char *s)
{
    int i;
    int val;
    int mask = 0;
    char reg[16];
    int ptrsav;

    while (1) {
        i = 0;
        ptrsav = txtptr;
        while (isalnum(txt[txtptr])) {
            reg[i++] = txt[txtptr++];
        }
        reg[i] = 0;
//        s = strndcpy(reg, s, " ,", 16);

        val = getregval(table, reg);
        mask |= val;
        if (val == -1) {
            error(REGUNREC, ptrsav);
        }

        if (txt[txtptr] != ',') {
            break;
        }

        txtptr++;
    }

    return mask;
}

/*****************************************************************************/

int bealine()
{
    int mnetyp, ptrsav;
    int val = 0;

    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    ptrsav = txtptr;
    opref = 0;

    mnetyp = mnefind(&opcode, &optype, &opref, cputypemask, 0);
    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    if (mnetyp == 3) {
        val = 1;
    }
    else {
        switch (mnetyp) {
        case -1:             /* end of line */
            val = -1;
            break;

        case 0:              /* not found */
            val = 1;
            break;

        case 1:              /* Mnemonics     */
            val = isadrfield();
            break;

        case 2:              /* DIRECTIVS     */
            val = 1;
            break;
        }
    }

    return val;
}

void beautifylist(char *name, char *targetname)
{
    int fcount;
    char *p;
    char *label;
    char *mnem;     // pointers in txt
    char *adrf;
    char *comment;
    int len;

    if (!open_infile(name)) {
        return;
    }
    if (!open_listfile(name)) {
        return;
    }

    if (tabs == -1) {                // tabs not specified
        tabcnt = 0;
        parsetablist("a8,16,40");    // set default
    }

    line = 0;
    adrsizcnt = 0;

    while(1) { 
        errpos = 0;               /* cleared for each line */
        txtptr = 0;
        lblflag = 0;
        aflag = dflag = rflag = 0;
        oplen = len = 0;

        if (readasm() == EOF) { // read line from file
            break;
        }
        prntline = line;          /* print this line number */


        fcount = -1;
        if (txt[txtptr] == ' ') { /* line starts with a space */
            fcount = bealine();
        }
        else if (txt[txtptr] == 0) { /* empty line */
        }
        else if (txt[txtptr] == '#') { /* assembly directive */
        }
        else if (iscomchar(txt[txtptr])) { /* comment line */
        }
        else {              /* it is a label/symbol definition */
            p = scanto(txt+txtptr, " ");
            txtptr = p - txt;
            fcount = bealine();
        }

        len = strlen(txt) - 1;    /* shop off the added space at the end */
        if (txt[len] == ' ') {
            txt[len] = 0;
        }

        if (fcount < 0) {   // output line as is
            fprintf(hdl_lst, "%s\n", txt);
        }
        else {  // split and join line with new column positions
            label = 0;
            mnem = 0;
            adrf = 0;
            comment = 0;

            txtout[0] = 0;
            p = txt;
            if (*p != ' ') {    // check if label
                label = p;
                p = scanto(p, " ");
                if (*p) {
                    *p++ = 0;
                }
                strcpy(txtout, txt);
            }

            p = esp(p);
            if (*p) {
                mnem = p;
                p = scanto(p, " ");
                if (*p) {
                    *p++ = 0;
                }

                if (fcount > 0) {
                    p = esp(p);
                    if (*p) {
                        adrf = p;
                        p = scanto(p, " ");
                        if (*p) {
                            *p++ = 0;
                        }
                    }                
                }

                p = esp(p);
                if (*p) {
                    comment = p;
                }
            }

            if (mnem) {
                len = strlen(txtout);
                p = txtout + len;
                *p++ = ' ';
                while ((p - txtout) < tablist[0]) {
                    *p++ = ' ';
                }
                strcpy(p, mnem);
            }

            if (adrf) {
                len = strlen(txtout);
                p = txtout + len;
                *p++ = ' ';
                while ((p - txtout) < tablist[1]) {
                    *p++ = ' ';
                }
                strcpy(p, adrf);
            }

            if (comment) {
                len = strlen(txtout);
                p = txtout + len;
                *p++ = ' ';
                while ((p - txtout) < tablist[2]) {
                    *p++ = ' ';
                }
                strcpy(p, comment);
            }

            fprintf(hdl_lst, "%s\n", txtout);
        }
    }
}

/* end */
