/***********************************/
/*   Externdeklarationer           */
/***********************************/

#ifndef ASMDEF_H
#define ASMDEF_H

/* sizes */

#define MAXTXTLEN    100
#define MAXBUFLEN    0x8000
#define MAXOPENFILE  10
#define OWNOBJ 1
#define S19OBJ 2
#define INTOBJ 3
#define TECOBJ 4
#define S19LEN 32
#define INTLEN 32
#define TECLEN 32


/* instruction types */
#define INH       1     /* 03, 05, 09, 11 12 */
#define IMM8      2     /* 03, 05, 09, 11 12 */
#define IMM16     3     /* 03,     09, 11 12 */
#define NODIR     4     /* 03, 05,     11 12 */
#define DIR2      4     /*         09        */
#define NOIMM     5     /* 03, 05, 09, 11 12 */
#define NOEXT     6     /*     05            */
#define IND       6     /*         09     12 */
#define REL       7     /* 03, 05, 09, 11 12 */
#define INH2      8     /*             11    */
#define REL16     8     /*         09     12 */
#define BCLR      9     /*     05,     11 12 */
#define IMMADR    9     /* 03      09        */
#define BRCLR     10    /*     05,     11 12 */
#define PSHS      10    /*         09        */
#define PSHU      11    /*         09        */
#define IM16X     11    /*             11    */
#define NODIR2    11    /*                12 */
#define IM16Y     12    /*             11    */
#define EXG       12    /*         09     12 */
#define NOIMX     13    /*             11    */
#define IMMON     13    /*         09     12 */
#define NOIMY     14    /*             11    */
#define NOIMM2    14    /*                12 */
#define CPD       15    /*             11    */
#define IND2      15    /*                12 */
#define IND3      16    /*                12 */
#define BAND      16    /*         09        */
#define EXTON     17    /*                12 */
#define TFM       17    /*         09        */
#define CALL      18    /*                12 */
#define MOVB      19    /*                12 */
#define MOVW      20    /*                12 */
#define TRAP      21    /*                12 */
#define LOOP      22    /*                12 */
#define IMM32     23    /*         09        */
#define CLRCC     24    /*         09        */
#define SETCC     25    /*         09        */
#define TRANS     30    /*         09     12 */
#define ASMDIR    31

#define TYPEMASK 0x1f

#define PAGE0	0x00    /*                   */
#define PAGE1	0x20    /*         09     12 */
#define PAGE2	0x40    /*         09        */
#define PAGE3	0x60    /*                   */
#define PAGEMASK 0x60

// access flags for assembler mnemonic
#define MNEACCSHIFT  8
#define MNEREADOP    0x000  /* op read value    ex LDA */
#define MNEWRITEOP   0x100  /* op write value   ex STA */
#define MNEMODIFYOP  0x200  /* op modify value  ex INC */
#define MNEJUMP      0x4000 /* op of type jump: JMP,JSR... Used for long branch warning */
#define READREF    0
#define WRITEREF   1
#define MODIFYREF  2
#define DEFINEREF  3

// directive flags
#define DIR_FLAGMASK    0xff00
#define DIR_ASSIGN      0x0100
#define DIR_GENERATE    0x0200
#define DIR_CONDITIONAL 0x0400


/* addressing modes */
#define AINH      0     /* 03, 05, 09, 11 12 */
#define AREL      1     /* 03, 05, 09, 11 12 */
#define AIMM8     2     /* 03, 05, 09, 11 12 */
#define AIMM16    3     /* 03,     09, 11 12 */
#define ADIR      4     /* 03, 05, 09, 11 12 */
#define AEXT      5     /* 03, 05, 09, 11 12 */
#define AIND      6     /* 03,     09, 11    */
#define AIND0     6     /*     05         12 */
#define AIND1     7     /*     05         12 */
#define AIMMIND   7     /* 03                */
#define ADIR2     7     /*         09     12 */
#define BCAIND    7     /*             11    */
#define AIND2     8     /*     05         12 */
#define BRCAIND   8     /*             11    */
#define AIMMDIR   8     /* 03                */
#define BCADIR    9     /*     05      11    */
#define AINDID    9     /*                12 */
#define BRCADIR   10    /*     05      11    */
#define AINDIN    10    /*                12 */
#define AREL16    11    /*         09     12 - */
#define ASPEC     12    /*         09     12 - */
#define AIMM32    13    /*         09        */


enum {    // assembler directives
    AD_FCB = 0,
    AD_FCC = 1,
    AD_FDB = 2,
    AD_FQB = 9,
    AD_ORG = 3,
    AD_END = 4,
    AD_RMB = 6,
    AD_TTL = 7,
    AD_OPT = 8,
    AD_DATA = 0x10,
    AD_TEXT = 0x11,
    AD_INCL = 0x12,
    AD_ALIG = 0x13,
    AD_PAGE = 0x20,
    AD_CPU = 0x21,
    AD_ZMB = 0x22,
    AD_ERROR,
    AD_FAIL,
    AD_WARN,
    AD_INFO,
    AD_IF,
    AD_IFC,
    AD_IFNC,
    AD_IFEQ,
    AD_IFNE,
    AD_IFGT,
    AD_IFGE,
    AD_IFLT,
    AD_IFLE,
    AD_ELSE,
    AD_ELSEC,
    AD_ENDC,
    AD_XREF,
    AD_XDEF,
    AD_NAM,
    AD_ENDM,
    AD_EXIT,
    AD_ENDSTRUCT,
    AD_REG,
    AD_SECTION,
    AD_SET,
    AD_STRUCT,
    AD_MACRO,
    AD_GENLAB,
    AD_POP,
    AD_EXCH,
    AD_FTB,
    AD_FDBR,
    AD_SETDP = 0x70,
    AD_EQU = 0x85,
};

// allow flags
#define ALLOWCPUMASK    0xffff
#define ALLOWMODEMASK   0xffff0000




enum {
    STRUCT_CLOSED = 0,
    STRUCT_OPEN,
    STRUCT_CLOSING
};

enum {
    MACRO_CLOSED = 0,
    MACRO_OPEN,
};


// PASS
enum {
    FIRST_PASS = 1,
    MIDDLE_PASS,
    LAST_PASS,
};


/*    */
#define PRGTEXT   0
#define PRGDATA   1

#define ASMPRT    1
#define FCBPRT    2
#define FDBPRT    3
#define PDTAPRT   4
#define FQBPRT    5
#define FTBPRT    6

#define SYMBOLLEN 16
#define EOL       10

/* Error codes  */
#define CUSTOMEINFO     -3      /*I  */
#define CUSTOMEWARNING  -2      /*W  */
#define CUSTOMEERROR    -1      /*E  */
#define ADRMODBAD       0       /*E  addressing mode is not allowed*/
#define BIGGERVAL       1       /*W  value exceeds available field width */
#define BRANCHOUT       2       /*E2 offset in branch out of range */
#define DOUBLEDEF       3       /*E1 symbol/macro double defined         */
#define ILLEGSYMB       4       /*E  symbol name is illegal        */
#define MNEMUNREC       5       /*E  unrecognised mnemonic         */
#define NOTBINNO        6       /*E  value is not a binary number  */
#define NOTHEXNO        7       /*E  value is not a hex number     */
#define PRNTSMIS        8       /*E  end parentes is missing       */
#define SYMBMISS        9       /*E  symbol is missing in EQU stat.*/
#define NOEXPR          10      /*E  no expression in address field */
#define SYMNOTDEF       11      /*E2 symbol is not defined         */
#define UNRECOPT        12      /*E  unrecognised option           */
#define ORGADDRERR      13      /*E  Address in ORG is in old area */
#define NOTYETIMP       14      /*E  Function not yet implemented */
#define NALWINDS        15      /*E  Not allowed in data segment */
#define TOOMANYFIL      16      /*E  Too many open files */
#define CANTOPNFIL      17      /*E  Can't Open File */
#define REGUNREC        18      /*E  Unknown register specification */
#define SHBRUSED        19      /*W  Short branch can be used */
#define VALOTORNG       20      /*E  value out of range */
#define STRENDMIS       21      /*E  missing end of string */
#define BASEOTORNG      22      /*E  base out of range */
#define NOTVALIDNO      23      /*E  not a valid number */
#define LBRLOWOFFS      24      /*I  long branch offset is near short range */
#define BRNCHOVERBRNCH  25      /*I  branch over branch */
#define BRCHSHORTTOLONG 26      /*I  Short branch changed to long */
#define BRCHLONGTOSHORT 27      /*I  Long branch changed to short */
#define DIVBYZERO       28      /*E  Divide by zero               */
#define ENDCWITHNOIF    29      /*E  ENDC/ELSE without IF         */
#define MISSINGENDC     30      /*E  Missing ENDC   */
#define MISSINGENDM     31      /*E  Missing ENDM   */
#define LBLSTACKERROR   32      /*E  Labelstack under/overflow   */
#define MISSINGPOP      33      /*E  Missing .POP   */
#define MACROINMACRO    34      /*E  Macro definition in macro  */
#define INSTROPTAWAY    35      /*I  Instruction optimized away */


typedef unsigned int uint;
typedef unsigned short ushort;
typedef unsigned char uchar;

typedef struct name_s {
    struct name_s *nxt;
    char *name;
} NAME_S;

typedef struct {
    short opcode[2];
    short page[2];
    int zeroopt;
} BranchOp;

typedef struct mne {
   char mnem[10];
   short opcode;
   short type;
   uint allowed;
} MneType;

typedef struct {
   char mnem[10];
   char reg[10];
} MneFix;

typedef struct xrefs {
    struct xrefs *xrefnxt;
    short xreftype; /* 0 read ' ', 1 write 'W', 2 modify 'M', 3 defined '*' */
    int xrefline;
} XrefType;

enum {
    DEF_NORMAL = 0,
    DEF_BYSET,
    DEF_WEAK,
};

/*
 *
 *
 */
enum {
    SYM_DBLDEF = 1,
    SYM_VALID = 2,
    SYM_WEAKDEF = 4,
};

enum {  // symclass
    SYMC_NOTDEF = 0,
    SYMC_VAR,           // ordinary; def by EQU, SET or program labels
    SYMC_STRUCT,        // STRUCT
};

typedef struct symb {
    struct symb *nxt;
    char *symbol;
    int symclass;
    short symflag;      /* b0 set when double defined */
                        /* b1 set when symbol has a valid value */
    short changed;
    long symvalue;
    int symsize;
    struct symb *lst;   // used for struct
    struct symb *nxt2;  // used for struct
    XrefType *symxref;
    XrefType *symxend;
} SymType;


typedef struct adr {
   short reserved;
   short mask;
   short addval;
   short length;
} AdrType;

typedef struct {
   char filesname[128];
   int filesline;
   FILE *fileshdl;
} FILE_S;

typedef struct {
   int trcount;
   int troutform;
   char trdta[4];
} TRNSL;

typedef struct {
	char regname[6];
	short regvalue;
    short allowed;
} REGNAM;

typedef struct {
    char *name;
    int mask;
} CPUTYPE;

enum {
    SECT_TEXT = 0,
    SECT_BSS
};

typedef struct section {
    struct section *nxt;
    char *secname;
    int sectype;
    int progmode;
    long secprgcnt;
} SECTION;

typedef struct txtline {
    struct txtline *nxt;
    char *txt;
    char *errtxt;
    int lineno;
    char *filename;
} TXTLINE;

typedef struct {
    TXTLINE *txtlh;
    TXTLINE *txtle;
} TXTHEAD;


typedef struct tratxt {
    struct tratxt *nxt;
    char *abbr;
    char *trans;
} TRATXT;

typedef struct {
    TRATXT *tlh;
    TRATXT *tle;
} TRATXTHEAD;



typedef struct macro {
    struct macro *nxt;
    char *macname;
    TXTHEAD mactxt;
    TXTLINE *txtlp;
    char **parlist;
    int parcount;
    char **parnamelist;
    int parnamecount;
    TRATXTHEAD labels;
    int condlevel;      // store the levels on entry of macro
    int skiplevel;      // to restore them at EXIT of macro
    int condskip;
    int maccount[3];    // reference counters
    char *macfile;      // file where the macro is defined
} MACRO;


#ifndef ASMDEF_C

extern char *defversion;
extern int bufptr;
extern int bufflg;

extern char buftab[MAXBUFLEN];

extern char *errstr[];


/*tables */  
extern CPUTYPE cputab[];

extern short pagetab[];
extern short adrcnv[];
extern AdrType adrtab[];
extern MneType mnemic[];
extern MneType mnemicend[];
extern unsigned short mnemiccnttab[];
extern int mnemiccnt;
extern int mnedircnt;
extern MneFix mnefixtbl[];


//extern REGNAM exgreg[];
//extern REGNAM exgregsrc[];
//extern REGNAM exgregdst[];
//extern REGNAM loopreg[];
//extern REGNAM indreg[];
//extern REGNAM accindreg[];


extern SymType *symtabh;
extern SymType *symtabe;
extern int symbolchanged;
extern int symboldoubledef;

/* variables */
extern long errstrtab[];
extern FILE *hdl_asm;
extern FILE *hdl_bin;
extern FILE *hdl_lst;
extern FILE *hdl_err;
extern FILE *hdl_scr;
extern long adr;

extern long prgcnt;
extern long prgcntdata;
extern long prgcnttext;
extern int prgmode;
extern SECTION *textsection;
extern SECTION *datasection;
extern SECTION *cursection;



extern long structprgcnt;
extern int structdef;

extern long symbolptr;
extern long startadr;
extern long runadr;
extern int xreflist;
extern int statistics;
extern int adrsizcnt;

extern int page;
extern int pagelen;
extern int linelen;
extern int line;
extern int prntline;
extern int pass;
extern int errcnt;
extern int errline;
extern int errlinelow;
extern int wrncnt;
extern int wrnline;
extern int wrnlinelow;
extern int infocnt;
extern int infoline;
extern int infolinelow;

extern int optimizebranch;
extern int noforwardopt;
extern int lbwrn;
extern int lblflag;
extern int aflag;
extern int dflag;
extern int rflag;
extern int endoffil;
extern int listflag;
extern int exprflag;
extern int txtptr;
extern int errpos;
extern int txtstart;
extern int opcode;
extern int optype;
extern int oplen;
extern int len;
extern int opref;
extern int opclen;
extern int errpostab[];
extern int errcodtab[];
extern int symbmaxlen;
extern int mulnoprio;

extern char *cmd;
extern char *cmdbase;
extern char txt[];
extern char mnesav[];
extern char txtout[];
extern char errout[];
extern char title[];
extern char sel[];
extern char path[];

extern char lstfilename[];
extern char objfilename[];
extern FILE_S files[];
extern int actfile;
extern int objfileformat;
extern int cputypemask;

extern unsigned char adrsiztab[];
extern unsigned char dta[];
extern unsigned char pdta[];  /* print data if different from output data */

/* Funktions in ASMxxMAN.C */

/* main()   */
char *manversion();
int adrfield();
int assigndirective_cpu(int opcode, int inptrsav);
int directive();
int diropt2();
void formlst();

/* Funktioner i ASMxxTBL.C */

char *tblversion();
long getmnemicsize();

/* Funktioner i ASMEXP.C */

int generateLabel(char *buf);
char *expversion();
long expr();
void copyvar();
long readvar();
void setvar();
void setvarsize(char *s, int val);
void getvar();
SymType *findvar();
void defineAndSetVar(char *name, int symclass, long val, int flag);

/* Funktioner i ASMFIL.C */

char *filversion();
FILE *open_inclfile();
int open_infile();
int open_outfile();
void getfname();
FILE *opfi();

/* Funktioner i ASMCOM.C */

void init();
int mnefind();
void error(int errcod, short tp);
void formerr();

#endif   // ASMDEF_C
#endif   // ASMDEF_H

/* end */
