/******************************************************/
/* code for decoding of aritmetics expressions        */
/******************************************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    930818      Ported from Atari to Unix on Sun
*
*  0.02    941019   1. files asm03man.c and asm03ma2.c concatenated,
*                       asmcom.c created.
*                   2. tabs replaced with spaces.
*
*  0.03    970314   1. Version printout added.
*
*  0.04    120517   1. Possible to set prio of * same as + (-noprio-flag)
*
*  0.05    161215   1. Relation ops added: &&, ||
*                   2. conditional ops added: ==, !=, >, <, >=, <=
*
*  0.06    170127   1. op: !< and !> added
*                   2. More Motorola-style op: !+ , !x and !.
*
*  0.07    171007   1. Macros added: MACR, EXIT, ENDM. Including local labels
*                       '\.n' and parameters '\n' where <n> is a value
*                   2. Support for control flow added: Generate unique labels
*                       Label stack, dir .GENLAB, .POP, .EXCH and read of label
*                       stack with '\:[n]'.
*
*  0.08    180202   1. struct added
*
*/
static char *version = "asmexp   0.08  2018-02-02";

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "asmdef.h"


typedef struct avk {
   char code[12];
   short type;
} AvkType;

#define IGNORE_CASE 256

enum {
    ADD_OP = 1,
    SUB_OP,
    EOR_OP,
    OR_OP,

    MUL_OP,
    DIV_OP,
    MOD_OP,
    AND_OP,
    SHR_OP,
    SHL_OP,

    RELAND_OP,
    RELOR_OP,

    EQ_OP,
    NE_OP,
    GT_OP,
    LT_OP,
    GE_OP,
    LE_OP,

    POS_OP,
    NEG_OP,
    COM_OP,

    BINNUM_OP,
    HEXNUM_OP,
    PAREN_OP,
    PRGCNT_OP,
    CHAR_OP,
    STR_OP,  
    SIZEOF_OP,
    BANKOF_OP,
    NARG_OP,
};



AvkType avktabrelation[] = {
    "&&", RELAND_OP,
    "||", RELOR_OP,
    " ",  -1,
};

AvkType avktabcond[] = {
    "==", EQ_OP,
    "!=", NE_OP,
    ">",  GT_OP,
    "<",  LT_OP,
    ">=", GE_OP,
    "<=", LE_OP,
    " ",  -1,
};


AvkType avktabterm2[] = {  // used when * / are not prio over + -
    "+",  ADD_OP,       // add
    "-",  SUB_OP,       // sub
    ":",  EOR_OP,       // eor  Acorn Atom BASIC style
    "^",  EOR_OP,       // eor  C style
    "||", 0,        // Q&D hack, protect this from decoding here
    "|",  OR_OP,        // or
    "*",  MUL_OP,       // mul
    "/",  DIV_OP,       // div
    "%",  MOD_OP,       // mod
    "&&", 0,        // Q&D hack, protect this from decoding here
    "&",  AND_OP,       // and
    ">>", SHR_OP,       // shift right
    "<<", SHL_OP,       // shift left
    "!+", OR_OP,        // or  Motorola style
    "!x", EOR_OP,       // eor  Motorola style
    "!X", EOR_OP,       // eor  Motorola style
    "!.", AND_OP,       // and  Motorola style
    "!>", SHR_OP,       // shift right  Motorola style
    "!<", SHL_OP,       // shift left  Motorola style
    " ",  -1,
};

AvkType avktabterm[] = {
    "+",  ADD_OP,       // add
    "-",  SUB_OP,       // sub
    ":",  EOR_OP,       // eor  Acorn Atom BASIC style
    "^",  EOR_OP,       // eor  C style
    "||", 0,        // Q&D hack, protect this from decoding here
    "|",  OR_OP,        // or
    "!+", OR_OP,        // or  Motorola style
    "!x", EOR_OP,       // eor  Motorola style
    "!X", EOR_OP,       // eor  Motorola style
    " ",  -1,
};

AvkType avktabfactor[] = {
    "*",  MUL_OP,       // mul
    "/",  DIV_OP,       // div
    "%",  MOD_OP,       // mod
    "&&", 0,        // Q&D hack, protect this from decoding here
    "&",  AND_OP,       // and
    "!.", AND_OP,       // and  Motorola style
    ">>", SHR_OP,       // shift right
    "!>", SHR_OP,       // shift right  Motorola style
    "<<", SHL_OP,       // shift left
    "!<", SHL_OP,       // shift left  Motorola style
    " ",  -1,
};

AvkType avktabsign[] = {
    "-",  NEG_OP,
    "~",  COM_OP,
    "+",  POS_OP,
    " ",  -1,
};

AvkType avktaboper[] = {
    "%",  BINNUM_OP,            // binary number
    "$",  HEXNUM_OP,            // hex number
    "(",  PAREN_OP,            // open parenthesis
    "*",  PRGCNT_OP,            // read prog counter
    "'",  CHAR_OP,            // ascii char
    "\"", STR_OP,            // string
    "sizeof(",  SIZEOF_OP,      // size of structs
    "bankof(",  BANKOF_OP,      // -- which bank is a symbol in, not implemented
    "narg",  NARG_OP+IGNORE_CASE, // narg, number of args in macro
    " ",  -1
};

#define MAXSYMLEN 63
#define MAXMACROLEVEL  20
#define LABELSTACKSIZE 20


int labelCounter;       // used to generate labels
char *labelstack[LABELSTACKSIZE];
int labelstackind;

int symboldoubledef;
int symbolchanged;
char label[MAXSYMLEN+1];
char symbol[MAXSYMLEN+1];
SymType *symtabh = 0;
SymType *symtabe = 0;
SECTION *sectlh = 0;
SECTION *sectle = 0;

MACRO *maclh;
MACRO *macle;
int macroLevel;
MACRO *macrotable[MAXMACROLEVEL];
int macexpand = 0;  //  b0=1:sow comment lines

SymType *structptr; // holds pointer to open struct between STRUCT and ENDSTRUCT

extern int xrefcnt;
extern char *actfname;
extern TXTHEAD srctxt;
extern int condlevel;
extern int skiplevel;
extern int condskip;


long expr1();
long term();
long factor();
long factor2();
int avk();
long relation();
long condition();
void storetxt(TXTHEAD *headp, int lineno, char *txt, char *fname);
void prntout();
int readasm();
char *strdcpy(char *d, char *s, char *e);
char *strndcpy(char *d, char *s, char *e, int len);
int iscomchar(int c);
int readsize();
void dirincl();
void prntbuff(int type);
char *esp(char *p);



char *expversion()
{

    return version;
}




long expr()
{
    int c;
    long val;

    if (txt[txtptr] == '<') {   /* take upper part (b15-b8)  */
        txtptr++;
        val = relation();
        val >>= 8;
        val &= 0xff;
    }
    else if (txt[txtptr] == '>') { /* take lower part (b7-b0) */
        txtptr++;
        val = relation();
        val &= 0xff;
    }
    else {
        val = relation();
    }
    return val;
}

long relation()
{
    int c;
    long val1;
    long val2;

    val1 = condition();

    while (c = avk(avktabrelation)) {
        switch (c) {
        case RELAND_OP:                /* '&&' and */
            val2 = condition();
            if (val1 && val2) {
                val1 = 1;
            }
            else {
                val1 = 0;
            }
            break;

        case RELOR_OP:                /* '||' or */
            val2 = condition();
            if (val1 || val2) {
                val1 = 1;
            }
            else {
                val1 = 0;
            }
            break;

        }
    }

    return val1;
}

long condition()
{
    int c;
    long val1;
    long val2;

    val1 = expr1();

    switch (avk(avktabcond)) {
    case EQ_OP:                   /* equal */
        val2 = expr1();
        val1 = (val1 == val2);
        break;

    case NE_OP:                   /* not equal */
        val2 = expr1();
        val1 = (val1 != val2);
        break;

    case GT_OP:                   /* higher */
        val2 = expr1();
//        printf("%d -- %d  '>'\n", val1, val2);
        val1 = (val1 > val2);
        break;

    case LT_OP:                   /* lower */
        val2 = expr1();
//        printf("%d -- %d  '<'\n", val1, val2);
        val1 = (val1 < val2);
        break;

    case GE_OP:                   /* higher or equal */
        val2 = expr1();
        val1 = (val1 >= val2);
        break;

    case LE_OP:                   /* lower or equal */
        val2 = expr1();
        val1 = (val1 <= val2);
        break;
    }

    return val1;
}

long expr1()
{
    int c;
    long val;
    long val2;
    int ptrsav;

    val = term();

    if (mulnoprio) {
        c = avk(avktabterm2);
    }
    else {
        c = avk(avktabterm);
    }

    while (c) {
        switch (c) {
        case ADD_OP:                /* '+' add */
            val += term();
            break;

        case SUB_OP:                /* '-' sub */
            val -= term();
            break;

        case EOR_OP:                /* ':' eor */
            val ^= term();
            break;

        case OR_OP:                /* '|' or */
            val |= term();
            break;

        case MUL_OP:                /* '*' mul */
            val *= factor();
            break;

        case DIV_OP:                /* '/' div */
            ptrsav = txtptr;
            val2 = factor();
            if (val2) {
                val /= val2;
            }
            else {
                error(DIVBYZERO, ptrsav);
            }
            break;

        case MOD_OP:                /* '%' mod */
            ptrsav = txtptr;
            val2 = factor();
            if (val2) {
                val %= val2;
            }
            else {
                error(DIVBYZERO, ptrsav);
            }
            break;

        case AND_OP:                /* '&' and */
            val &= factor();
            break;

        case SHR_OP:                /* '>>' shift right */
            val >>= factor();
            break;

        case SHL_OP:                /* '<<' shift left */
            val <<= factor();
            break;

        }
        if (mulnoprio) {
            c = avk(avktabterm2);
        }
        else {
            c = avk(avktabterm);
        }
    }
    return val;
}

long term()
{
    int c;
    long val;
    long val2;
    int ptrsav;

    val = factor();

    if (mulnoprio) {
        return val;
    }

    while (c = avk(avktabfactor)) {
        switch (c) {
        case MUL_OP:                /* '*' mul */
            val *= factor();
            break;

        case DIV_OP:                /* '/' div */
            ptrsav = txtptr;
            val2 = factor();
            if (val2) {
                val /= val2;
            }
            else {
                error(DIVBYZERO, ptrsav);
            }
            break;

        case MOD_OP:                /* '%' mod */
            ptrsav = txtptr;
            val2 = factor();
            if (val2) {
                val %= val2;
            }
            else {
                error(DIVBYZERO, ptrsav);
            }
            break;

        case AND_OP:                /* '&' and */
            val &= factor();
            break;

        case SHR_OP:                /* '>>' shift right */
            val >>= factor();
            break;

        case SHL_OP:                /* '<<' shift left */
            val <<= factor();
            break;

        }
    }
    return val;
}

long factor()
{
    int c;
    long val;

    switch (avk(avktabsign)) {
    case 0:                   /* ''  nothing */
        val = factor2();
        break;

    case POS_OP:                   /* '+' positive */
        val = factor();
        break;

    case NEG_OP:                   /* '-' negative */
        val = -factor();
        break;

    case COM_OP:                   /* '~' complement */
        val = ~factor();
        break;

    }
    return val;
}

int chartoval(int ch)
{

    if (ch >= '0' && ch <= '9') {
        return ch - '0';
    }
    else if (ch >= 'A' && ch <= 'Z') {
        return ch - 'A' + 10;
    }
    else if (ch >= 'a' && ch <= 'z') {
        return ch - 'a' + 10;
    }
    return -1;
}

int isinbase(int ch, int base)
{

    ch = chartoval(ch);
    if (ch >= base) {
        return -1;
    }
    return ch;
}

long factor2()
{
    char tt[200];
    char *p;
    int i;
    int ch;
    int c, flag;
    int val;
    int base;
    int ptrsav;

    switch (avk(avktaboper)) {
    case 0:                   /* constant or symbol */
        if (isdigit(txt[txtptr])) {
            val = 0;
            ptrsav = txtptr;
            i = 0;
            while (isinbase(ch = txt[txtptr], 36) >= 0) {
                tt[i++] = ch;
                txtptr++;
            }
            tt[i] = 0;
            i--;
            if (toupper(tt[i]) == 'H') {        // 0abcdH
                tt[i] = 0;
                val = strtol(tt, &p, 16);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else if (toupper(tt[i]) == 'O') {   // 01234o
                tt[i] = 0;
                val = strtol(tt, &p, 8);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else if (toupper(tt[i]) == 'D') {   // 01278d
                tt[i] = 0;
                val = strtol(tt, &p, 10);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else if (toupper(tt[i]) == 'B') {   // 0101b
                tt[i] = 0;
                val = strtol(tt, &p, 2);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else if (txt[txtptr] == '#') {      // 8#1234
                txtptr++;
                base = strtol(tt, &p, 10);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
                if (base < 2 || base > 36) {
                    error(BASEOTORNG, ptrsav);
                }
                val = 0;
                while (isinbase(ch = txt[txtptr], base) >= 0) {
                    val = val * base + chartoval(ch);
                    flag = 0;
                    txtptr++;
                }
            }
            else if (tt[0] == '0' && toupper(tt[1]) == 'X') { // 0xabcd
                val = strtol(tt+2, &p, 16);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else if (tt[0] == '0' && toupper(tt[1]) == 'B') { // 0b0101
                val = strtol(tt+2, &p, 2);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
            else {                                          // 1234
                val = strtol(tt, &p, 10);
                if (*p) {
                    error(NOTVALIDNO, ptrsav);
                }
            }
        }
        else {
            val = readvar();
        }
        break;

    case BINNUM_OP:                   /* '%' binary number */
        val = 0;
        flag = 1;
        ptrsav = txtptr;
        while ((ch = txt[txtptr]) == '0' || ch == '1') {
            val = (val << 1) + chartoval(ch);
            flag = 0;
            txtptr++;
        }

        if (flag) {
            error(NOTBINNO, ptrsav);
        }
        break;

    case HEXNUM_OP:                   /* '$' hex number */
        val = 0;
        flag = 1;
        ptrsav = txtptr;
        while (isxdigit(txt[txtptr])) {
            ch = txt[txtptr];
            txtptr++;
            val = (val << 4) + chartoval(ch);
            flag = 0;
        }

        if (flag) {
            error(NOTHEXNO, ptrsav);
        }
        break;

    case PAREN_OP:                   /* '(' */
        val = expr();
        if (txt[txtptr] == ')') {
            txtptr++;
        }
        else {
            error(PRNTSMIS, txtptr);
        }
        break;

    case PRGCNT_OP:                   /* '*', read pc */
        val = prgcnt;
        break;

    case CHAR_OP:                   /* ''', singel 'fnutt', read as ascii */
        val = txt[txtptr];
        if (val) {
            txtptr++;
        }
        if (txt[txtptr] == '\'') {
            txtptr++;
        }
        break;
    case STR_OP:                     /* '"' string */
        val = 0;

        ptrsav = txtptr-1;
        while (1) {
            if (txt[txtptr] == 0) {
                error(STRENDMIS, ptrsav);
                break;
            }
            else if (txt[txtptr] == '"') {
                txtptr++;
                if (txt[txtptr] == '"') {
                    val = val * 256 + txt[txtptr++];
                }
                else {
                    break;
                }
            }
            else {
                val = val * 256 + txt[txtptr++];
            }
        }
        break;
    case SIZEOF_OP:                     /* sizeof( */
        val = readsize();
        if (txt[txtptr] == ')') {
            txtptr++;
        }
        else {
            error(PRNTSMIS, txtptr);
        }
        break;
    case BANKOF_OP:                     /* bankof(   */
        val = 0;
//        val = expr();
        if (txt[txtptr] == ')') {
            txtptr++;
        }
        else {
            error(PRNTSMIS, txtptr);
        }
    case NARG_OP:                     /* narg (only for macros)  */
        if (macroLevel > 0) {
            val = macrotable[macroLevel-1]->parcount;
        }
        else {
            val = 0;
        }

    }
    return val;
}

int avk(AvkType *avkptr)
{

    while (avkptr->type != -1) {
        if (!strncasecmp(avkptr->code, txt + txtptr,
                     strlen(avkptr->code))) {
            if ((avkptr->type & 0xff) != 0) {   // Q&D hack, to protect some pattern
                txtptr += strlen(avkptr->code);
            }
            return avkptr->type & 0xff;
        }
        avkptr++;
    }
    return 0;
}


/******************************************************/
/*    symbol handling                                 */
/******************************************************/


void insertxref(SymType *sym, short ref)
{
    XrefType *xptr;

    if (pass != FIRST_PASS) {
        return;
    }

    xrefcnt++;
    xptr = (XrefType *)malloc(sizeof(XrefType));
    if (xptr) {
        memset(xptr, 0, sizeof(XrefType));
        if (sym->symxref == 0) {
            sym->symxref = xptr;
        }
        if (sym->symxend) {
            sym->symxend->xrefnxt = xptr;
        }
        sym->symxend = xptr;

        xptr->xrefnxt = 0;
        xptr->xreftype = ref;
        xptr->xrefline = line;
    }
}

SymType *insertsym(char *name)
{
    SymType *stp;

    stp = (SymType*) malloc(sizeof(SymType));

    memset(stp, 0, sizeof(SymType));
    stp->symbol = strdup(name);

    if (symtabh == 0) {
        symtabh = stp;
        symtabe = stp;
    }
    else {
        symtabe->nxt = stp;
        symtabe = stp;
    }
    symbolptr++;
    return stp;
}

SymType *findvar(char *name)
{
    SymType *stp;
    int i;

    for (stp = symtabh; stp; stp = stp->nxt) {
        if (stp->symclass == SYMC_VAR) {
            if (!strcmp(name, stp->symbol)) {
                return stp;
            }
        }
    }

    return 0;
}

SymType *findstruct(char *name)
{
    SymType *stp;
    int i;

    for (stp = symtabh; stp; stp = stp->nxt) {
        if (stp->symclass == SYMC_STRUCT) {
            if (!strcmp(name, stp->symbol)) {
                return stp;
            }
        }
    }

    return 0;
}

SymType *findsym(char *name)
{
    SymType *stp;
    int i;

    for (stp = symtabh; stp; stp = stp->nxt) {
        if (!strcmp(name, stp->symbol)) {
            return stp;
        }
    }

    return 0;
}


void clearSymbols()
{
    SymType *stp;

    for (stp = symtabh; stp; stp = stp->nxt) {
        stp->changed = 0;
    }
}



int copyvarptrsav;

void copyvar()
{

    copyvarptrsav = txtptr;
    getvar(label);
    if (txt[txtptr] == ':') {
        txtptr++;
    }

}

/*
 * definevar
 *
 * define a symbol
 *
 * param int symclass:
 * param int flag:  0=normal set, 1=set by SET directive, 2=weak definition
 *
 */
SymType *definevar(int symclass, int flag)
{
    SymType *ptr;

//    printf("definevar %s\n", label);
    ptr = findsym(label);
    if (ptr == 0) {             /* symbol not defined */
        ptr = insertsym(label);
        ptr->symclass = symclass;
        if (flag == DEF_WEAK) {
            ptr->symflag |= SYM_WEAKDEF | SYM_VALID;
            ptr->symvalue = 0;
        }
    }
    else if (ptr->symflag & SYM_VALID) {    /* symbol defined and valid */
        if (pass == FIRST_PASS) {
            if (flag == DEF_NORMAL) {
                if ((ptr->symflag & SYM_WEAKDEF)) {
                    ptr->symflag &= ~SYM_WEAKDEF;
                }
                else {
                    ptr->symflag |= SYM_DBLDEF;
                    symboldoubledef++;
                }
            }
            else if (flag == DEF_WEAK) {
                ptr->symflag |= SYM_WEAKDEF;
            }
        }
        else if (pass == LAST_PASS) {
            if (ptr->symflag & SYM_DBLDEF) {
                error(DOUBLEDEF, copyvarptrsav);
            }
        }
    }
    else {                      /* symbol defined, but not valid */
        if (flag == DEF_WEAK) {
            ptr->symflag |= SYM_WEAKDEF | SYM_VALID;
            ptr->symvalue = 0;
        }
    }

    return ptr;
}

/*
 * defineAndSetVar
 *
 * define a symbol and assign value
 *
 * pram char *name:   Name of symbol, use variable label if null
 * param int symclass:
 * param long val: value
 * param int flag:  0=normal set, 1=set by SET directive, 2=weak definition
 *
 */
void defineAndSetVar(char *name, int symclass, long val, int flag)
{
    SymType *ptr;

    if (name == 0) {
        name = label;
    }

//    printf("defineAndSetVar %s\n", name);
    ptr = findsym(name);
    if (ptr == 0) {             /* symbol not defined */
        ptr = insertsym(name);
        ptr->symclass = symclass;
        if (flag == DEF_NORMAL) { // set by EQU, or ordinare label
            ptr->changed++;
            symbolchanged++;
        }
        else if (flag == DEF_WEAK) {
            ptr->symflag |= SYM_WEAKDEF;
        }
        ptr->symvalue = val;
        ptr->symflag |= SYM_VALID;
        if (xreflist) { /* insert XREF */
            insertxref(ptr, DEFINEREF);
        }
    }
    else if (ptr->symflag & SYM_VALID) {    /* symbol defined and valid */
        if (pass == FIRST_PASS) {
            if (flag == DEF_NORMAL) {
                if ((ptr->symflag & SYM_WEAKDEF)) {
                    ptr->symflag &= ~SYM_WEAKDEF;
                }
                if (ptr->symvalue != val) {
                    if (ptr->changed) {
                        ptr->symflag |= SYM_DBLDEF;
                        symboldoubledef++;
                    }
                    if (flag == DEF_NORMAL) { // set by EQU, or ordinare label
                        ptr->changed++;
                        symbolchanged++;
                    }
                }
            }
//            else if (flag == DEF_WEAK) {
//                ptr->symflag |= SYM_WEAKDEF;
//            }
            ptr->symvalue = val;
            if (xreflist) { /* insert XREF */
                insertxref(ptr, DEFINEREF);
            }
        }
        else if (pass == LAST_PASS) {
            if (ptr->symflag & SYM_DBLDEF) {
                error(DOUBLEDEF, copyvarptrsav);
            }
        }
        else {
            if (ptr->symvalue != val) {
                if (flag == DEF_NORMAL) { // set by EQU, or ordinare label
                    ptr->changed++;
                    symbolchanged++;
                }
            }
            ptr->symvalue = val;
            ptr->symflag |= SYM_VALID;
            if (xreflist) { /* insert XREF */
                insertxref(ptr, DEFINEREF);
            }
        }
    }
    else {                      /* symbol defined, but not valid */
        if (flag == DEF_NORMAL) { // set by EQU, or ordinare label
            ptr->changed++;
            symbolchanged++;
        }
//        else if (flag == DEF_WEAK) {
//            ptr->symflag |= SYM_WEAKDEF;
//        }

        ptr->symvalue = val;
        ptr->symflag |= SYM_VALID;
        if (xreflist) { /* insert XREF */
            insertxref(ptr, DEFINEREF);
        }
    }

}

/*
 * setvar
 *
 * set current symbol
 *
 * param long val:  value
 * param int flag:  0=normal set, 1=set by SET directive
 *
 */
void setvar(long val, int flag)
{
    SymType *ptr;

    ptr = findsym(label);
//    printf("setvar   %s = 0x%04X\n", label, val);
    if (ptr) {
        if (ptr->symvalue != val || (ptr->symflag & SYM_VALID) == 0) {
            if (flag == DEF_NORMAL) { // set by EQU, or ordinare label
                ptr->changed++;
                symbolchanged++;
//                printf("setvar OLD %s = 0x%04X\n", label, ptr->symvalue);
            }
//            if (pass == MIDDLE_PASS) {
//                fprintf(hdl_scr, "Label '%s'   0x%04X -> 0x%04X\n",
//                        label, (uint) ptr->symvalue, (uint) val);
//            }
        }
        ptr->symvalue = val;
        ptr->symflag |= SYM_VALID;
        if (xreflist) { /* insert XREF */
            insertxref(ptr, DEFINEREF);
        }
    }
    else {
    }
}

/*
 * setvarsize
 *
 * set variable size
 *
 * param char *name: name of variable, current label if zero
 * param int val:  value
 *
 */
void setvarsize(char *name, int val)
{
    SymType *ptr;

    if (name == 0) {
        name = label;
    }

    ptr = findvar(name);
    if (ptr) {
        ptr->symsize = val;
    }
}

int readsize()
{
    SymType *ptr;
    int ptrsav;
    char symbol[64];

    ptrsav = txtptr;
    getvar(symbol);

    if (symbol[0] == 0) {
        return 0;
    }

    ptr = findsym(symbol);
    if (ptr) { /* symbol exist */
        if (ptr->symclass == SYMC_VAR) {
            if (xreflist) { /* insert XREF */
                insertxref(ptr, opref);
            }
            if (ptr->symflag & SYM_VALID) {  /* value is valid */
                return ptr->symsize;
            }
        }
        else if (ptr->symclass == SYMC_STRUCT) {
            return ptr->symsize;
        }
    }
    else {
    }

    if (pass == LAST_PASS) {
        error(SYMNOTDEF, ptrsav);
    }
    return 0;
}

long readvar()
{
    SymType *ptr;
    int ptrsav;

    ptrsav = txtptr;
    getvar(symbol);

    if (symbol[0] == 0) {
        return 0;
    }

    ptr = findvar(symbol);
    if (ptr) { /* symbol exist */
        if (xreflist) { /* insert XREF */
            insertxref(ptr, opref);
        }
        if (ptr->symflag & SYM_VALID) {  /* value is valid */
            return ptr->symvalue;
        }
    }
    else {
        ptr = findsym(symbol);
        if (ptr) {
            error(DOUBLEDEF, ptrsav);
        }
        else {
            ptr = insertsym(symbol);
            ptr->symclass = SYMC_VAR;
            if (xreflist) { /* insert XREF */
                insertxref(ptr, opref);
            }
        }
    }

/* the symbol is either not defined or has an invalid value */
    exprflag |= 2;           /* flag not defined symbol */
    if (pass == LAST_PASS) {
        error(SYMNOTDEF, ptrsav);
    }
    return 0;
}


/*
 *
 *
 *
 *  the '#' is used for generated labels, do not use in source code
 *
 */
int issymchar(int c)
{

    return isalnum(c) || c == '.' || c == '_' ||
        c == '$' || c == '@' || c == '?' || c == '#';
}

/*
 * get the variable from source text
 *
 *
 */
void getvar(char *name)
{
    int ch;
    int i;

    for (i = 0; issymchar(txt[txtptr]); txtptr++, i++) {
        ch = txt[txtptr];
        if (i < MAXSYMLEN) {
            *name++ = ch;
        }
    }
    *name = 0;

    if (i > symbmaxlen) {
        symbmaxlen = i;
    }
}


void printChangedSymbols()
{
    SymType *stp;

    for (stp = symtabh; stp; stp = stp->nxt) {
        if (stp->changed) {
            printf("%2d %-13s  %5d (0x%04X)\n",
                   stp->changed, stp->symbol,
                   (int)stp->symvalue, (uint)stp->symvalue);
        }
    }
}

void prntsymtab(FILE *hdl)
{
    SymType *stp;
    int i;
    int formlen;
    int symcnt;
    char tout[200];
    char tt[100];
    int maxlen;
    int len;
    int symperline;
    int ind;

    symcnt = symbolptr;
    fprintf(hdl, "\n symbol table:  (%d)\n", symcnt);

    maxlen = symbmaxlen;
    if (maxlen > 22) {
        maxlen = 22;
    }

    formlen = maxlen + 5;

    symperline = linelen / formlen;
    if (symperline == 0) {
        symperline = 1;
    }

    ind = 0;
    i = 0;
    for (stp = symtabh; stp; stp = stp->nxt) {
        if (stp->symclass == SYMC_VAR) {
            if (stp->symflag & SYM_VALID) {
                sprintf(tout+i, "%04lX %s",
                        stp->symvalue & 0xffffffff,
                        stp->symbol);
            }
            else {
                sprintf(tout+i, "**** %s",
                        stp->symbol);
            }

            len = strlen(tout+i);
            if (len > formlen) {
                strcpy(tout+i+formlen-2, "..");
                len = formlen;
            }

            ind++;
            if (ind < symperline) {
                while (len <= formlen) {
                    tout[i + len++] = ' ';
                }
                i += formlen + 1;
            }
            else {
                tout[i+len] = 0;
                fprintf(hdl, " %s\n", tout);
                i = 0;
                ind = 0;
            }
        }
    }

    if (i) {
        while (tout[i - 1] == ' ') {
            --i;
        }
        tout[i] = 0;
        fprintf(hdl, " %s\n", tout);
    }

}

void prntxreftab(FILE *hdl)
{
    SymType *stp;
    int i, formlen;
    int symcnt;
    XrefType *xptr;
    int xrchar;
    char tout[200];
    int maxlen;

    symcnt = symbolptr;
    fprintf(hdl, "\n cross reference table: (%d references) (%d symbols)\n",
            xrefcnt, symcnt);

    maxlen = symbmaxlen;
    if (maxlen > 22) {
        maxlen = 22;
    }

    formlen = maxlen + 5;

    for (stp = symtabh; stp; stp = stp->nxt) {
        if (stp->symclass == SYMC_VAR) {
            if (stp->symflag & SYM_VALID) {
                sprintf(tout, "%04lX %s",
                        stp->symvalue & 0xffffffff,
                        stp->symbol);
            }
            else {
                sprintf(tout, "**** %s",
                        stp->symbol);
            }

            len = strlen(tout);
            if (len > formlen) {
                strcpy(tout+formlen-2, "..");
                len = formlen;
            }

            while (len < formlen) {
                tout[i + len++] = ' ';
            }

            i = formlen;

            for (xptr = stp->symxref; xptr; xptr = xptr->xrefnxt) {
                switch (xptr->xreftype & 0xf) {
                default:
                    xrchar = '?';
                    break;
                case READREF:
                    xrchar = ' ';
                    break;
                case WRITEREF:
                    xrchar = 'W';
                    break;
                case MODIFYREF:
                    xrchar = 'M';
                    break;
                case DEFINEREF:
                    xrchar = '*';
                    break;
                }
                if (i == 0) {
                    while (i < formlen) {
                        tout[i++] = ' ';
                    }
                }
                sprintf(tout+i, " %05d%c", xptr->xrefline, xrchar);
                i += 7;

                if ((i + 7) > linelen) {
                    while (tout[i - 1] == ' ') {
                        --i;
                    }
                    tout[i] = 0;
                    fprintf(hdl, " %s\n", tout);
                    i = 0;
                }
            }
            if (i) {
                while (tout[i - 1] == ' ')
                    --i;

                tout[i] = 0;
                fprintf(hdl, " %s\n", tout);
                i = 0;
            }
        }
    }

    if (i) {
        while (tout[i - 1] == ' ')
            --i;

        tout[i] = 0;
        fprintf(hdl, " %s\n", tout);
    }

}

/**********************************************************************/
/*
 * generateLabel
 *
 * generate a unique label
 *
 *
 *
 */
int generateLabel(char *buf)
{

    labelCounter++;

    sprintf(buf, "L#%04d", labelCounter);

    return 0;
}

/*
 * dirgenlab
 *  Generate a label and push it on stack
 *
 *   .GENLAB
 *
 *
 */
int dirgenlab()
{
    char buf[32];

    if (labelstackind < LABELSTACKSIZE-1) {
        generateLabel(buf);
        labelstack[labelstackind] = strdup(buf);
        labelstackind++;
    }
    else {
        error(LBLSTACKERROR, 0);
    }

    return 0;
}

/*
 * dirpop
 *  pop top of stack
 *  
 *   .POP
 *
 */
int dirpop(int ptrsav)
{

    if (labelstackind >= 1) {
        labelstackind--;
        free(labelstack[labelstackind]);
        labelstack[labelstackind] = 0;
    }
    else {
        error(LBLSTACKERROR, ptrsav);
    }

    return 0;
}

/*
 * direxch
 *  exchange two values on label stack
 *
 *   .EXCH  [deep]
 *
 *
 */
int direxch(int ptrsav)
{
    char *p;
    int deep = 0;


    if (isdigit(txt[txtptr])) {
        deep = strtol(txt + txtptr, &p, 10);
        txtptr = p - txt;
    }

    deep++;
    if (labelstackind >= (deep+1)) {
        p = labelstack[labelstackind-deep];
        labelstack[labelstackind-deep] = labelstack[labelstackind-deep-1];
        labelstack[labelstackind-deep-1] = p;
    }
    else {
        error(LBLSTACKERROR, ptrsav);
    }

    return 0;
}

char *readlabelstack(int deep, int txtind)
{
    char *p;

    deep++;
    if (labelstackind >= deep) {
        p = labelstack[labelstackind-deep];
    }
    else {
        error(LBLSTACKERROR, txtind);
        p = "";
    }

    return p;
}

enum {
    SF_DEF,
    SF_FORW,
    SF_BACK
};

char currentLabel[32];      // holds label for the last \.
char nextLabel[32];         // holds label for a \+ until the \. is reached

/*
 * getShortFormLabel
 *  get a label in the short form \. , \+ or \-
 *
 *   \.  define a label
 *   \+  forward reference
 *   \-  backward reference
 *
 *
 *
 *
 */
char *getShortLabel(int type)
{
    char buf2[32];
    int ret = 0;
    char *p;

    if (type == SF_DEF) {
        if (nextLabel[0]) {
            strcpy(currentLabel, nextLabel);
            nextLabel[0] = 0;
        }
        else {
            generateLabel(currentLabel);
        }
        p = currentLabel;
    }
    else if (type == SF_FORW) {
        if (nextLabel[0] == 0) {
            generateLabel(nextLabel);
        }
        p = nextLabel;
    }
    else if (type == SF_BACK) {
        if (currentLabel[0] == 0) {     // this is an error case, signalling it
            generateLabel(currentLabel); // as undefined symbol
            ret = 1;
        }
        p = currentLabel;
    }

    return p;
}

void expandShortFormLabel(char *dest)
{
    int mind;
    char buf[64];
    int i;
    int val;
    char *p;
    int len;
    char *d = dest;
    char *s = txtout;

    strcpy(s, d);

    if (*s == '+' || *s == '-') {   // some markings of the lines
        *d++ = *s++;
    }

    if (*s == '\\' && s[1] == '.') { // first check for define a label
        if (isalnum(s[2])) {        // do not convert local labels in macro
            *d++ = *s++;            // just pass the backslash
        }
        else {          // convert short form label
            p = getShortLabel(SF_DEF);
            s += 2;

            // eat some spaces to make the listing more beautiful
            len = strlen(p) - 2;
            while (len > 0 && *s == ' ' && s[1] == ' ') {
                s++;
                len--;
            }
            strcpy(d, p);
            d += strlen(d);
        }
    }

    while (*s) {
        if (*s == '\\') {
            if (s[1] == '+') {  // forward ref
                p = getShortLabel(SF_FORW);
                s += 2;

                // eat some spaces to make the listing more beautiful
                len = strlen(p) - 2;
                while (len > 0 && *s == ' ' && s[1] == ' ') {
                    s++;
                    len--;
                }
                strcpy(d, p);
                d += strlen(d);
            }
            else if (s[1] == '-') {   // backward ref
                p = getShortLabel(SF_BACK);
                s += 2;

                // eat some spaces to make the listing more beautiful
                len = strlen(p) - 2;
                while (len > 0 && *s == ' ' && s[1] == ' ') {
                    s++;
                    len--;
                }
                strcpy(d, p);
                d += strlen(d);
            }
            else {
                *d++ = *s++;
                *d++ = *s++;
            }
        }
        else {
            *d++ = *s++;
        }
    }

    *d = 0;

}



/********************************************************************/
/*
 * Section
 *
 *
 *
 */
SECTION *makeSection(char *name, int type, int mode)
{
    SECTION *secp;

    secp = malloc(sizeof(SECTION));

    memset(secp, 0, sizeof(SECTION));

    secp->sectype = type;
    secp->progmode = mode;
    secp->secname = strdup(name);

    return secp;
}

SECTION *addSection(char *name, int type, int mode)
{
    SECTION *secp;

    secp = makeSection(name, type, mode);
    if (sectlh == 0) {
        sectlh = secp;
        sectle = secp;
    }
    else {
        sectle->nxt = secp;
        sectle = secp;
    }

}

void clearSections()
{
    SECTION *secp;

    for (secp = sectlh; secp; secp = secp->nxt) {
        secp->secprgcnt = 0;
    }
}

/********************************************************************/
/*
 * Macro
 *
 *
 *
 */


TRATXT *addTratxt(TRATXTHEAD *headp, char *abb, char *tra)
{
    TRATXT *tp;

    tp = malloc(sizeof(TRATXT));

    memset(tp, 0, sizeof(TRATXT));

    tp->abbr = strdup(abb);
    tp->trans = strdup(tra);

    if (headp->tlh == 0) {
        headp->tlh = tp;
        headp->tle = tp;
    }
    else {
        headp->tle->nxt = tp;
        headp->tle = tp;
    }

    return tp;
}

void freeTratxt(TRATXT *tp)
{

    free(tp->abbr);
    free(tp->trans);
    free(tp);
}

MACRO *findMacro(char *name)
{
    MACRO *mp;

    for (mp = maclh; mp; mp = mp->nxt) {
        if (!strcmp(mp->macname, name)) {
            return mp;
        }
    }

    return 0;
}

MACRO *insertMacro(char *name)
{
    MACRO *mp;

    mp = malloc(sizeof(MACRO));

    memset(mp, 0, sizeof(MACRO));

    mp->macname = strdup(name);

    if (maclh == 0) {
        maclh = mp;
        macle = mp;
    }
    else {
        macle->nxt = mp;
        macle = mp;
    }

    return mp;
}

void splitParamList(MACRO *mp, int flag);

/*
 * readMacroDef
 *  read from input and store macro
 *
 *  sym MACR [parnamelist]
 */
int readMacroDef()
{
    char *p;
    MACRO *mp;
    int i;
    int ret;
    int mnetyp;
    int ptrsav;
    int opref;
    int opcode;
    int optype;
    int comflag;

    if (pass != FIRST_PASS) {   // process macros during first pass
        mp = 0;
    }
    else if (lblflag) {
        mp = findMacro(label);  // macro name stored in var 'label'
        if (mp) {
            error(DOUBLEDEF, copyvarptrsav);
            mp = 0;     // no active macro, continue process input, but don't store
        }
        else {
            mp = insertMacro(label);
            mp->macfile = strdup(actfname);
        }

        while (txt[txtptr] == ' ') {
            txtptr++;
        }
        splitParamList(mp, 1);
    }
    else {
        error(SYMBMISS, 0);
        mp = 0;
    }

    if (pass == FIRST_PASS) {   // store input in text list
        storetxt(&srctxt, line, txt, actfname);
    }
    prntout();

    while (1) {
        ret = readasm();          // read from input file
        prntline = line;          /* print this line number */
        if (ret == EOF) {         // end of input
            error(MISSINGENDM, 0);
            break;
        }
        else if (ret == 2) {      // include file closed
            prntline = 0;         /* print no line number */
        }

        p = esp(txt+txtptr);

        if (iscomchar(*p)) {
            comflag = 1;
            mnetyp = 0;
        }
        else {
            comflag = 0;
            while (txt[txtptr] && txt[txtptr] != ' ') {
                txtptr++;
            }
            while (txt[txtptr] == ' ') {
                txtptr++;
            }
            ptrsav = txtptr;
            mnetyp = mnefind(&opcode, &optype, &opref, cputypemask, 0);
        }


//        printf("txt: %s\n", txt);
//        printf("mnetyp: %d, opcode: %d\n", mnetyp, opcode);

        if (mnetyp == 2 && opcode == AD_INCL) {         /* INCL(UDE) */
            if (pass == FIRST_PASS) {   // store input in text list
                storetxt(&srctxt, line, txt, actfname);
            }
            prntout();
            if (mp) {
                dirincl();
            }
        }
        else if (mnetyp == 3 && opcode == AD_MACRO) {   /* macro in macro */
            error(MACROINMACRO, ptrsav);
            break;
        }
        else if (mnetyp == 2 && opcode == AD_ENDM) {    /* end of macro */
            break;
        }
        else {
            if (pass == FIRST_PASS) {   // store input in text list
                storetxt(&srctxt, line, txt, actfname);
            }
            prntout();
            if (mp) {
                if (comflag == 0) {
                    storetxt(&(mp->mactxt), 0, txt, 0);
                }
                else {      // comment line
                    if ((macexpand & 1) == 1) {
                        storetxt(&(mp->mactxt), 0, txt, 0);
                    }
                }
            }
        }

    }

}

void freeParamList(MACRO *mp)
{
    int i;

    for (i = 0; i < mp->parcount; i++) {
        free(mp->parlist[i]);
    }

    free(mp->parlist);
    mp->parlist = 0;
    mp->parcount = 0;
}

void freeLabelList(MACRO *mp)
{
    TRATXT *tp;
    int i;

    while (mp->labels.tlh) {
        tp = mp->labels.tlh;
        mp->labels.tlh = tp->nxt;
        freeTratxt(tp);
    }
    mp->labels.tle = 0;
}


char *getMacroLabel(MACRO *mp, char *name)
{
    TRATXT *tp;
    int found = 0;
    char buf[32];

    for (tp = mp->labels.tlh; tp; tp = tp->nxt) {
        if (!strcmp(name, tp->abbr)) {
            return tp->trans;
        }
    }

    generateLabel(buf);
    addTratxt(&(mp->labels), name, buf);
    return mp->labels.tle->trans;
}

int findNamedPar(MACRO *mp, char *name)
{
    int i;

    for (i = 0; i < mp->parnamecount; i++) {
        if (!strcmp(mp->parnamelist[i], name)) {
            return i;
        }
    }

    return -1;
}

int expandMacroLine(MACRO *mp, char *dest, char*s)
{
    int mind;
    char buf[64];
    int i;
    int val;
    char *p;
    int len;
    char *d = dest;

    *d++ = '+';
    while (*s) {
        if (*s == '\\') {
            s++;
            if (*s == '{' || isdigit(*s)) {  // param found
                if (isdigit(*s)) {      // digit
                    buf[0] = *s++;
                    buf[1] = 0;
                    mind = strtol(buf, 0, 10);
                }
                else {                  // bracket
                    s++;
                    i = 0;
                    if (isdigit(*s)) {      // digit
                        while (isdigit(*s)) {
                            buf[i++] = *s++;
                        }
                        buf[i] = 0;
                        mind = strtol(buf, 0, 10);
                    }
                    else {                  // named param
                        while (isalnum(*s)) {
                            buf[i++] = *s++;
                        }
                        buf[i] = 0;

                        mind = findNamedPar(mp, buf);
                    }

                    if (*s == '}') {
                        s++;
                    }
                }
                if (mind >= 0 && mind < mp->parcount) { // replace parno with param text
                    p = mp->parlist[mind];
                    strcpy(d, p);
                    d += strlen(d);
                }
            }
            else if (*s == '.' || *s == ':') {   // read label
                if (*s == ':') {        // read from label stack
                    s++;
                    if (isdigit(*s)) {
                        buf[0] = *s;
                        buf[1] = 0;
                        val = *s++ - '0';
                    }
                    else {
                        buf[0] = 0;
                        val = 0;
                    }
                    p = readlabelstack(val, d-dest);
                }
                else {                  // local label
                    s++;
                    i = 0;
                    while (isalnum(*s)) {
                        buf[i++] = *s++;
                    }
                    buf[i] = 0;
                    p = getMacroLabel(mp, buf);
                }
                len = strlen(p) - strlen(buf) - 2;  // adjust to make the listing
                while (len > 0 && *s == ' ' && s[1] == ' ') { // more beautiful
                    s++;
                    len--;
                }
                strcpy(d, p);
                d += strlen(d);
            }
            else {
                *d++ = *s++;
            }
        }
        else {
            *d++ = *s++;
        }
    }

    *d = 0;
    
}

int expandMacro()
{
    MACRO *mp;
    int mnetyp;
    int opref;
    int opcode;
    int optype;
    char tt[200];

    if (macroLevel < 1) {
        return 0;
    }

    mp = macrotable[macroLevel-1];

    if (mp->txtlp == 0) {
        return 0;
    }

    expandMacroLine(mp, txt, mp->txtlp->txt);

    txtptr = 1;
    if (!iscomchar(txt[txtptr])) {
        while (txt[txtptr] && txt[txtptr] != ' ') {
            txtptr++;
        }
        while (txt[txtptr] == ' ') {
            txtptr++;
        }

        mnetyp = mnefind(&opcode, &optype, &opref, cputypemask, 0);


/*  EXIT - leave macro
 *   Problem: the nesting level for conditional assembler is
 *   leaved in a wrong state, simple fix exists.
 */
        if (mnetyp == 2 && opcode == AD_EXIT) {                /* EXIT */
            if (condskip == 0) {
                freeParamList(mp);
                freeLabelList(mp);
                macroLevel--;
                mp->txtlp = 0;
                condlevel = mp->condlevel;      // restore levels at EXIT
                skiplevel = mp->skiplevel;
                condskip = mp->condskip;
//                sprintf(tt, "! on EXIT: %s.  condlvl %d, skiplvl %d",
//                        mp->macname, condlevel, skiplevel);
//                storetxt(&srctxt, 0, tt, 0);
                return 1;
            }
        }
    }

    txtptr = 1;

    mp->txtlp = mp->txtlp->nxt;
    if (mp->txtlp == 0) {
        freeParamList(mp);
        freeLabelList(mp);
        macroLevel--;
        condlevel = mp->condlevel;      // restore levels
        skiplevel = mp->skiplevel;
        condskip = mp->condskip;
//        sprintf(tt, "! on ENDM: %s.  condlvl %d, skiplvl %d",
//                mp->macname, condlevel, skiplevel);
//        storetxt(&srctxt, 0, tt, 0);
    }

    return 1;
}

char *copyParam(char *d, char *p, int len)
{
    int parenlvl;

    len--;

    if (*p == '(') {
        parenlvl = 0;
        p++;
        while (*p && len > 0) {
            if (*p == '(') {
                *d++ = *p++;
                len--;
                parenlvl++;
            }
            else if (*p == ')') {
                if (parenlvl == 0) {
                    p++;
                    break;
                }
                else {
                    *d++ = *p++;
                    len--;
                    parenlvl--;
                }
            }
            else {
                *d++ = *p++;
                len--;
            }
        }
        *d = 0;
    }
    else {
        p = strdcpy(d, p, " ,");
    }

    return p;
}

/*
 * splitParamList
 *
 *
 * param MACRO *mp: pointer to macro
 * param int flag : 0 = translation, 1 = definition
 *
 */
void splitParamList(MACRO *mp, int flag)
{
    int parcount;
    char *parlist[20];  // no more than 20 params
    char param[128];
    char *p;
    int i;

    p = txt + txtptr;
    parcount = 0;

    while (*p) {
        p = copyParam(param, p, 128);
        parlist[parcount++] = strdup(param);
        if (*p != ',' || parcount >= 20) {
            break;
        }
        p++;
    }

    if (parcount > 0) {
        if (flag == 0) {
            mp->parlist = malloc(sizeof(char*) * parcount);
            //memset(mp->parlist, 0, sizeof(int) * parcount);
            for (i = 0; i < parcount; i++) {
                mp->parlist[i] = parlist[i];
            }
            mp->parcount = parcount;
        }
        else {
            mp->parnamelist = malloc(sizeof(char*) * parcount);
            for (i = 0; i < parcount; i++) {
                mp->parnamelist[i] = parlist[i];
            }
            mp->parnamecount = parcount;
        }
    }

    txtptr = p - txt;
}

/*
 * checkMacroDef
 * Check if current mnemonics is a macro.
 * If it is, prepare for macro expansion:
 * - read and split parameters
 * - Set txt pointer
 *
 */
int checkMacroDef()
{
    MACRO *mp;
    char name[64];
    int i;
    int p;
    char tt[200];

    i = 0;
    p = txtptr;
    while (txt[p] != ' ' && i < 62) {
        name[i++] = txt[p++];
    }
    name[i] = 0;

    mp = findMacro(name);
    if (mp == 0) {
        return 0;
    }

// macro found
    if (macroLevel >= MAXMACROLEVEL) {
        //error
        return 1;
    }

    if (pass != FIRST_PASS) {   // process macros during first pass
        return 1;
    }

    txtptr = p;
    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    splitParamList(mp, 0);

// prepare variables
    mp->txtlp = mp->mactxt.txtlh;
    mp->condlevel = condlevel;      // to solve problem with EXIT
    mp->skiplevel = skiplevel;
    mp->condskip = condskip;

    macrotable[macroLevel] = mp;

    i = macroLevel;
    if (i > 2) {
        i = 2;
    }
    mp->maccount[i]++;

    macroLevel++;

//    sprintf(tt, "! on entry: %s.  condlvl %d, skiplvl %d",
//            mp->macname, condlevel, skiplevel);
//    storetxt(&srctxt, 0, tt, 0);

    return 1;
}

void printmacros(FILE *hdl)
{
    MACRO *mp;
    TXTLINE *tlp;
    int i;

    fprintf(hdl,"List of macros:\n");
    fprintf(hdl,"\n");

    for (mp = maclh; mp; mp = mp->nxt) {
        fprintf(hdl,"---- macro: %s  ", mp->macname);
        for (i = 0; i < mp->parnamecount; i++) {
            if (i) {
                fprintf(hdl,", ");
            }
            fprintf(hdl,"%s", mp->parnamelist[i]);

        }

        fprintf(hdl," --------\n");

        for (tlp = mp->mactxt.txtlh; tlp; tlp = tlp->nxt) {
            fprintf(hdl,"%s\n", tlp->txt);
        }
    }
}

char *pdig(char *buf, int val)
{

    if (val == 0) {
        strcpy(buf, "");
    }
    else {
        sprintf(buf, "%d", val);
    }

    return buf;
}

void printmacstat(FILE *hdl)
{
    MACRO *mp;
    TXTLINE *tlp;
    int i;
    char buf1[8];
    char buf2[8];
    char buf3[8];

    fprintf(hdl,"Statistics of macros usage:\n");
    fprintf(hdl,"\n");

    fprintf(hdl,"macro           counts             defined in file\n");
    fprintf(hdl,"--------------  ----- ----- -----  -----------------------\n");

    for (mp = maclh; mp; mp = mp->nxt) {
        fprintf(hdl, "%-15s %5s %5s %5s  %s\n", mp->macname,
                pdig(buf1, mp->maccount[0]), pdig(buf2, mp->maccount[1]),
                pdig(buf3, mp->maccount[2]), mp->macfile);
    }
}

/*
 * Init variables for each pass
 *
 *
 *
 */
void initPassExp()
{

    clearSections();

    symbolchanged = 0;
    clearSymbols();

    macroLevel = 0;

    while (labelstackind > 0) {
        dirpop(0);
    }

}

/*
 * after Pass
 *  run after each pass(, or at least after the last)
 *  to detect some mismatching errors
 *
 *
 */
void afterPassExp()
{

    if (labelstackind > 0) {
        error(MISSINGPOP, 0);
    }



}


/********************************************************************/
/*
 * Struct
 *
 *
 *
 */

void appendToLst(SymType *top, SymType *sp1)
{
    SymType *sp;

    if (top->lst == 0) {
        top->lst = sp1;
    }
    else {
        for (sp = top->lst; sp->nxt2; sp = sp->nxt2) {
            if (sp == sp1) {
                return;
            }
        }
        sp->nxt2 = sp1;
    }
}

void addVarToStruct(char *name)
{
    SymType *sp;

    if (pass == FIRST_PASS) {
        if (structptr) {
            if (name == 0) {
                name = label;
            }

            sp = findvar(name);
            if (sp) {
                appendToLst(structptr, sp);
                
            }
        }
    }
}

void addVarToSymtab(char *name, SymType *sp, int prgcnt)
{
    char tt[200];
    int val = prgcnt;

    while (sp) {
        sprintf(tt, "%s.%s", name, sp->symbol);
        defineAndSetVar(tt, SYMC_VAR, val, DEF_NORMAL);
        if (sp->lst) {
            addVarToSymtab(tt, sp->lst, val);
        }
        val += sp->symsize;
        sp = sp->nxt2;
    }

}


/*
 * checkStructDef
 * Check if current mnemonics is a struct definition.
 * If it is, do more or less as it is a RMB.
 * - read the size of the struct
 * - Get the operand, if any (defaults to 1)
 * - then calculate the size = struct size times number
 * - Reserve space, this is like the RMB
 *
 */
int checkStructDef()
{
    SymType *sp;
    SymType *sp2;
    char name[64];
    int i;
    int p;
    char tt[200];
    int ptrsav;
    int val;

    // get name of struct
    i = 0;
    p = txtptr;
    while (txt[p] != ' ' && i < 62) {
        name[i++] = txt[p++];
    }
    name[i] = 0;

    sp = findstruct(name);
    if (sp == 0) {
        return 0;   // not found
    }

    // check operand
    txtptr = p;
    while (txt[txtptr] == ' ') {
        txtptr++;
    }

    val = 1;                                        // set default
    if (txt[txtptr] && !iscomchar(txt[txtptr])) {
        ptrsav = txtptr;
        val = expr();
    }

    val = val * sp->symsize;

    if (lblflag) {
        setvarsize(0, val);
        addVarToStruct(0);

        // concat and set additional names in sym-tab

        addVarToSymtab(label, sp->lst, prgcnt);

        sp2 = findvar(label);
        sp2->lst = sp->lst;
    }

    dta[0] = val >> 8;
    dta[1] = val;
    dflag = FDBPRT;         /* print data    */
    len = 2;
    prgcnt += val;
    if (prgmode == PRGTEXT) {
        if (bufflg) {
            prntbuff(1);
            bufflg = 0;
        }
        startadr = prgcnt;
    }

    return 1;
}

void dirstruct()
{
    SymType *sp;
    char name[64];

    sp = definevar(SYMC_STRUCT, DEF_NORMAL);
    structptr = sp;
    setvar(0, DEF_NORMAL);

}

void direndstruct()
{
    SymType *sp;
    char name[64];

    if (structptr) {
        structptr->symsize = structprgcnt;
        structptr->symsize = structprgcnt;
        structptr = 0;
    }
}

/* end */
