/****************************/
/*  Dump S19 file to output */
/*   including ascii        */
/****************************/

/*************************************************************************
*               R E V I S I O N   H I S T O R Y                       
*                         Release 0
*
* Rev #    Date        Modifications/Corrections for new Revision
* -----  ---------     ---------------------------------------------------
*  0.01    140511      New
*
*
*
*
*
*
*/
char *version = "0.01  2014-05-11";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define MAXBUFLEN   0x10000
#define INFILESMAX  32
#define S19LEN      32
#define ASCII_POS   77

enum {
    DATA_RECORD = 0,
    INIT_RECORD,
    END_RECORD
};


typedef unsigned int uint;
typedef unsigned char uchar;

typedef struct infile {
    char *name;
    int inaddr;
    int outaddr;

} INFILE;

int debugLevel = 0;

FILE *fe;
char *cmdbase;
char *cmd;

int rlength = S19LEN;

char outfile[250];
int pattern = 0xff;
int noS0 = 0;

INFILE infiles[INFILESMAX];
int infilescnt = 0;

int outputall = 0;
int outsize = MAXBUFLEN;
int outstart = 0;
uchar buf[MAXBUFLEN+128];
char mark[MAXBUFLEN+128];
char s0name[64];
char txtbuf[200];

int outbytecnt;

void init()
{
   int i;

   for (i = 0; i < outsize; i++) {
      buf[i] = pattern;
      mark[i] = 0;
   }

}

/*
 *      convert ascii string (of hex digits) to binary values
 */
int readhex(char *p, int sz, char **pp)
{
    int c;
    int val;

    val = 0;
    while (sz) {
        c = *p;

        if (c >= 'A' && c <= 'F') {
            c = c - 'A' + 10;
        }
        else if (c >= 'a' && c <= 'f') {
            c = c - 'a' + 10;
        }
        else if (c >= '0' && c <= '9') {
            c = c - '0';
        }
        else {
            break;
        }

        p++;
        val = (val * 16) + c;
        --sz;
    }

    if (pp) {
        *pp = p;
    }
    return val;
}

/*
 *
 * S0 0F 0000 6C69676874706F772E61736D 13
 * S1 23 F800 0F8E00FF860F970086FF970686FF970097044F9705D602C4F054D78286009708 FA
 * S9 03 0000 FC
 *
 * Return: type or -1 if error
 */
int splitline(uchar *buf, int *lenp, int *addrp, int *alen, char *p)
{
    int type;
    int addr;
    int addrlen;
    int count;
    int val;
    int cs;
    int ind;

// line header
    if (toupper(*p) != 'S') {
        fprintf(fe, "no S\n");
        return -1;
    }
    p++;

    if (*p == '0') {
        type = INIT_RECORD;
        addrlen = 2;
    }
    else if (*p == '1') {
        type = DATA_RECORD;
        addrlen = 2;
    }
    else if (*p == '2') {
        type = DATA_RECORD;
        addrlen = 3;
    }
    else if (*p == '3') {
        type = DATA_RECORD;
        addrlen = 4;
    }
    else if (*p == '7') {
        type = END_RECORD;
        addrlen = 4;
    }
    else if (*p == '8') {
        type = END_RECORD;
        addrlen = 3;
    }
    else if (*p == '9') {
        type = END_RECORD;
        addrlen = 2;
    }
    else {
        fprintf(fe, "S-record type\n");
        return -1;
    }
    p++;

    cs = 0;

// length
    if (!isxdigit(*p)) {
        fprintf(fe, "no length\n");
        return -1;
    }
    count = readhex(p, 2, &p);
    cs += count;

    count -= 3;
// addr
    if (!isxdigit(*p)) {
        fprintf(fe, "no addr\n");
        return -1;
    }
    addr = readhex(p, addrlen*2, &p);
    cs += ((addr >> 8) & 0xff) + (addr & 0xff);

//    fprintf(fe, "addr: %04X, len: %02X\n", addr, count);

// data
    ind = 0;
    while (ind < count) {
        if (!isxdigit(*p)) {
            fprintf(fe, "data missing\n");
            return -1;
        }
        val = readhex(p, 2, &p);
        cs += val;
        buf[ind] = val;
        ind++;
    }

// checksum
    cs = (~cs) & 0xff;

    if (!isxdigit(*p)) {
        fprintf(fe, "no CS\n");
        return -1;
    }
    val = readhex(p, 2, &p);
    if (val != cs) {
        fprintf(fe, "CS error\n");
        return -1;
    }

    *lenp = count;
    *addrp = addr;
    return type;
}

int ReadLine(FILE *ff, char *txtbuf, int buflen)
{
    int i, c;
    int rv;

    --buflen; // make space for endmark

    rv = 0;
    i = 0;
    while ((c = fgetc(ff)) != EOF && c != '\n') {
        if (c == 0x1a) {    // skip MSDOS END-OF-FILE char
        }
        else if (c == '\r') {    // skip CR
        }
        else if (i < buflen) {
            txtbuf[i++] = c;
        }
    }

    txtbuf[i] = 0;

    if (c == EOF) {
        return EOF;
    }
    return rv;
}

int convchar(int c)
{

    if (c < 0x20) {
        c = '.';
    }
    else if (c > 0x7e) {
        c = '.';
    }

    return c;
}

void readinput(INFILE *fp)
{
    FILE *fh;
    int eof;
    char *p;
    int len;
    int addr;
    int alen;
    int ind;
    int type;
    uchar dta[128];
    int line;
    int offset;
    int i;
    char tt[200];
    int txtlen;

    fh = fopen(fp->name, "r");
    if (fh == NULL) {
        fprintf(fe, "Infile not found: '%s'\n", fp->name);
        exit(1);
    }

    line = 0;
    eof = 0;
    while (eof != EOF) {
        eof = ReadLine(fh, txtbuf, 200);
        line++;

        if (txtbuf[0] == 0) {  // empty line
        }
        else {
            type = splitline(dta, &len, &addr, &alen, txtbuf);
            if (type < 0) {
                fprintf(fe, "%s: S-record error at line %d\n", fp->name, line);
//                exit(1);
            }
            else {
                txtlen = strlen(txtbuf);
                txtbuf[txtlen++] = ' ';
//                printf("len:%d  %d\n", txtlen, ASCII_POS);
                while (txtlen < ASCII_POS) {
                    txtbuf[txtlen++] = ' ';
                }
//                printf("len:%d  %d\n", txtlen, ASCII_POS);

                for (i = 0; i < len; i++) {
                    txtbuf[txtlen++] = convchar(dta[i]);
                }
                txtbuf[txtlen] = 0;
            }

            printf("%s\n", txtbuf);
        }
    }

    fclose(fh);
}

void usage(char *arg)
{
    int cmdlen = strlen(cmdbase);
    int i;

    printf("\n");
    printf("usage:  %s [-v] [-h] file\n", cmdbase);
    printf("\n");
    printf("   -v           print version\n");
    printf("   -h           print help\n");
//    printf("   -o outfile   output file\n");
    printf("   file         input files\n");
    printf("\n");

}

enum {
    NXT_OUTFILE = 1,
    NXT_SIZE,
    NXT_FILL,
    NXT_NAME,
    NXT_LENGTH,
};

int GetParameter(int argc, char *argv[])
{
    int i;
    int nxtarg;
    char *p;
    int val1 = -1;
    int val2 = -1;
    int mult;

    cmd = argv[0];
    cmdbase = strrchr(cmd, '/');
    cmdbase = cmdbase ? cmdbase+1 : cmd;

    nxtarg = 0;
    for (i = 1; i < argc; i++) {
        if (*argv[i] == '-') {
            nxtarg = 0;
            switch (argv[i][1]) {
            case 'v':   /* print version */
                printf("\n%s:   version %s\n\n", cmd, version);
                exit(0);

            default:
                fprintf(fe, "Unknown option: %s\n", argv[i]);
                usage(argv[i]);
                exit(2);

            case 'h':   /* help */
                usage(argv[i]);
                exit(0);

            case '-':
                break;

            case 'D':   /* deug */
                p = argv[i]+2;
                if (isxdigit(*p)) {
                    debugLevel = strtol(p, &p, 16);
                }
                else {
                    debugLevel = 1;
                }
                break;

            case 'o':   /* outfile */
                p = argv[i]+2;
                if (*p) {
                    strcpy(outfile, p);
                }
                else {
                    nxtarg = NXT_OUTFILE;
                }
                break;
            }
        }
        else {
            p = argv[i];
            switch (nxtarg) {
            case 0:
                if (infilescnt < INFILESMAX) {
                    infiles[infilescnt].name = p;
                    infiles[infilescnt].inaddr = val1;
                    infiles[infilescnt].outaddr = val2;
                    infilescnt++;
                    val1 = -1;
                    val2 = -1;
                }
                break;

            case NXT_OUTFILE:
                strcpy(outfile, p);
                nxtarg = 0;
                break;

            default:
                nxtarg = 0;
                break;
            }
        }
    }
    return 0;
}

int main(int argc, char *argv[])
{
    int i;

    fe = stderr;

    GetParameter(argc, argv);

    init();

    for (i = 0; i < infilescnt; i++) {
        readinput(infiles+i);
    }

    exit(0);
}

/* end */

